#
# $Source: /home/bhowes/BoilerPlate/RCS/BoilerPlate.py,v $
# $Author: bhowes $
# $Date: 1997/10/07 18:23:29 $
# $State: Exp $
#
#---------------------------------------------------------------------------
#
# Copyright (C) 1997 Motorola, Inc.
#
# Copyright in this software is owned by Motorla, Inc. Permission to
# use, copy and distribute this software royalty-free is hereby
# granted, provided that the above copyright notice appears in all
# copies and that both the copyright notice and this permission notice
# appear.
#
# THERE IS ABSOLUTELY NO WARRANTY FOR THIS SOFTWARE.
#
# The software is provided "as is" without warranty of any kind,
# either express or implied, including, but not limited to, the
# implied warranties of merchantability, fitness for a particular
# purpose, or non-infringement. This software could include technical
# inaccuracies or typographical errors. The author may make
# improvements and/or changes in this software at any time without
# notice to users of previous versions.
#
# --------------------------------------------------------------------------
#
# BoilerPlate -- collection of classes that provide simple yet powerful
# macro-type processing of Python text. Special tags in text delineate
# areas of blocks. A tag controls if a block is shown (`if' tags), or it
# initiates an iteration over a section of text using values found in a
# Python sequence or dictionary (`for' tags).
#
#---------------------------------------------------------------------------
# $Log: BoilerPlate.py,v $
# Revision 1.2  1997/10/07 18:23:29  bhowes
# Changed tag character from `@' to `#'.
# Added `lineMode' processing to BlockText.
# Added FileWatch class.
#
# Revision 1.1  1997/10/07 18:21:30  bhowes
# Initial revision
#

version = "$Id: BoilerPlate.py,v 1.2 1997/10/07 18:23:29 bhowes Exp bhowes $"

import copy, regex, string
from Sink import Sink

#
# RawText -- instances describe a range within a text string.
#
class RawText:

	#
	# __init__ -- remember text block and positions within it that we
	# own. If filter is true, then remove first and last character if
	# they are EOL's.
	#
	def __init__( self, text, start, end ):
		self.text = text[ start : end ]

	#
	# Cook -- apply the given dictionary to a range of text we own.
	#
	def Cook( self, sink, dict ):
		sink.Append( self.text % dict )

#
# Class Block -- fundamental class used to record blocks of unformatted
# text. Blocks nest within other blocks. See the derived classes IfBlock
# and ForBlock.
#
class Block:

	#
	# Distinguish ourselves from other Block types
	#
	kBlockKind = None

	#
	# Define regular expression used to locate start and end of blocks.
	# Does not detect `elif' and `else' blocks. First grouping is the
	# entire match. The second grouping is the kind of block that was
	# found. The third group contains any additional text following the
	# tag.
	#
	kTags = ( '!', 'char', 'end', 'for', 'if' )
	kTagsRe = '\(#\(%s\)\([^#]*\)#\)'
	tagre = regex.compile( kTagsRe % string.joinfields( kTags, '\|' ) )

	#
	# __init__ -- accept optional argument to control how we handle
	# end-of-line characters. If true, then we eat an EOL characters
	# between blocks.
	# 
	def __init__( self, lineMode = None ):
		self.lineMode = lineMode

	#
	# AddRaw -- assume ownership of a range of text if the range has any
	# text in it.
	#
	def AddRaw( self, text, startPos, endPos ):
		if self.lineMode:
			count = len( text )
			if startPos < count and text[ startPos ] == '\n':
				startPos = startPos + 1
			if endPos <= count and text[ endPos - 1 ] == '\n':
				endPos = endPos - 1
		if startPos < endPos:
			self.blocks.append( RawText( text, startPos, endPos ) )

	#
	# AddChild -- add the given child to our list of owned blocks, and
	# let the child eat text until it is done. Skips comment blocks.
	# Returns position of first unprocessed character.
	#
	def AddChild( self, child, text, startPos ):
		if child.kBlockKind != '!':
			self.blocks.append( child )
		return child.EatText( text, startPos )

	#
	# NewChild -- create a subblock based on the tag found in the text.
	# Returns position of first unprocessed character in text.
	#
	def NewChild( self, kind, text, startPos ):

		#
		# From a top-level, only `for', `if' and `char' blocks are
		# allowed
		#
		if kind == 'if':
			child = IfBlock( self.tagre.group( 3 ), self.lineMode )
		elif kind == 'for':
			child = ForBlock( self.tagre.group( 3 ), self.lineMode )
		elif kind == '!':
			child = CommentBlock( self.lineMode )
		elif kind == 'char':
			child = HatchChar()
		else:
			raise RuntimeError, 'invalid tag found:', kind

		return self.AddChild( child, text, startPos )

	#
	# IsClosure -- returns true if the statement is a block closure.
	# This is a method so that derived classes may override it.
	#
	def IsClosure( self, kind ):
		return kind == 'end'

	#
	# CloseBlock -- finish working with a block of text. Calculate the
	# position of the next character in the text string to process.
	#
	def CloseBlock( self, text, startPos, endPos ):

		#
		# Verify that the `end' statement belongs to us.
		#
		check = string.strip( self.tagre.group( 3 ) )
		if check != self.kBlockKind:
			raise RuntimeError, \
				  "expected 'end' to '%s' block -- got '%s'" % \
				  ( self.kBlockKind, check )

		#
		# Own everything up to our closure and return first free
		# character.
		#
		self.AddRaw( text, startPos, endPos )
		return endPos + len( self.tagre.group( 0 ) )

	#
	# EatText -- process the given text string until we run out of
	# characters or we have encountered a proper block closure. Returns
	# an index into the given text string of the first unprocessed
	# character.
	#
	def EatText( self, text, startPos ):

		#
		# Initialize list of blocks we control
		#
		self.blocks = []
		limit = len( text )
		index = -1
		while startPos < limit:

			#
			# Look for next block tag
			#
			index = self.tagre.search( text, startPos )
			if index < 0:
				break

			#
			# Found a tag.
			#
			else:
				kind = string.lower( self.tagre.group( 2 ) )

				#
				# Are we done? If so, calculate the last character we
				# processed to return to the caller.
				#
				if self.IsClosure( kind ):
					endPos = self.CloseBlock( text, startPos, index )
					break

				#
				# Nope -- must have a embedded block. Grab any
				# unprocessed text as ours and create a new child block.
				#
				else:
					self.AddRaw( text, startPos, index )
					startPos = index + len( self.tagre.group( 0 ) )
					startPos = self.NewChild( kind, text, startPos )

			index = -1

		#
		# Finished processing -- check for premature end of text.
		#
		if index == -1:

			#
			# If we are not a named block, just take everything as a
			# final text block
			#
			if self.kBlockKind == None:
				endPos = limit
				self.AddRaw( text, startPos, endPos )

			else:
				raise RuntimeError, "no 'end' to '%s' block" % \
					  self.kBlockKind

		return endPos

	#
	# Cook -- visit all of the blocks under our control and format each
	# using the given dictionary.
	#
	def Cook( self, sink, dict ):
		for each in self.blocks:
			each.Cook( sink, dict )

#
# Class HatchChar -- replaces #char# tag with `#' character.
#
class HatchChar( Block ):
	
	#
	# EatText -- override of Block method. Do nothing
	#
	def EatText( self, text, startPos ):
		return startPos

	#
	# Cook -- simply emit a `#' character
	#
	def Cook( self, sink, dict ):
		sink.Append( '#' )

#
# Class CommentBlock -- treats everything until its matching `end' tag
# as a comment.
#
class CommentBlock( Block ):

	kBlockKind = '!'

#
# Class IfBlock -- handles sequences of `if', `elif', and `else' text
# blocks. Each block except an `else' block must have a condition that
# will be evaluated during text formatting to see if the block's text is
# to be used.
#
class IfBlock( Block ):

	kBlockKind = 'if'

	#
	# Define regular expression to locate all block tags, including
	# `elif' and `else'.
	#
	kTags = Block.kTags + ( 'elif', 'else' )
	tagre = regex.compile( Block.kTagsRe % \
						   string.joinfields( kTags, '\|' ) )

	#
	# __init__ -- remember condition to test during formatting.
	#
	def __init__( self, condition, lineMode ):
		Block.__init__( self, lineMode )
		condition = string.strip( condition )
		if len( condition ) == 0:
			raise RuntimeError, "empty '%s' condition" % self.kind
		self.condition = condition

	#
	# AddAlternative -- add a new alternative block to our list.
	#
	def AddAlternative( self, child, text, startPos ):
		self.alternatives.append( child )
		return child.EatText( text, startPos )

	#
	# NewChild -- override of Block method. We handle `elif' and `else'
	# blocks here.
	#
	def NewChild( self, kind, text, startPos ):
		if kind == 'elif':
			child = ElifBlock( self.tagre.group( 3 ), self.lineMode )
			endPos = self.AddAlternative( child, text, startPos )

		#
		# There should not be any conditions following `else'
		#
		elif kind == 'else':
			extra = string.strip( self.tagre.group( 3 ) )
			if extra:
				raise RuntimeError, \
					  "illegal text after 'else' tag: %s" % \
					  self.tagre.group( 3 )

			child = ElseBlock( self.lineMode )
			endPos = self.AddAlternative( child, text, startPos )

		#
		# Must be something in our main block -- have not seen an `elif'
		# or `else' block yet.
		#
		else:
			endPos = Block.NewChild( self, kind, text, startPos )

		return endPos

	#
	# EatText -- override of Block method. Initializes a list of
	# alternative blocks to hold any `elif' or `else' blocks after any
	# text blocks.
	#
	def EatText( self, text, startPos ):
		self.alternatives = []
		return Block.EatText( self, text, startPos )

	#
	# Cook -- override of Block method. If our condition returns `true',
	# we format whatever text we have. Otherwise, we look for an
	# alternative clause (`elif' or `else' block) that has a true
	# condition.
	#
	def Cook( self, sink, dict ):
		if dict.Resolve( self.condition ):
			Block.Cook( self, sink, dict )
		else:

			#
			# Visit each alternative until a true value is returned.
			#
			for each in self.alternatives:
				value = each.Cook( sink, dict )
				if value:
					break

#
# Class ElseBlock -- handles text blocks starting with an `else'
# statement.
#
class ElseBlock( Block ):

	#
	# Really not an `if' block, but we do this so we will recognize `end
	# if' as a proper closure.
	# 
	kBlockKind = 'if'

	#
	# CloseBlock -- override of Block method. We don't move past the `#'
	# statement that closed us so that our parent `if' block will see
	# it.
	#
	def CloseBlock( self, text, startPos, endPos ):
		self.AddRaw( text, startPos, endPos )
		return endPos

#
# Class ElifBlock -- handles text blocks starting with an `elif'
# statement.
#
class ElifBlock( IfBlock ):

	#
	# CloseBlock -- override of Block method. We don't move past the `#'
	# statement that closed us so that our parent `if' block will see
	# it.
	#
	def CloseBlock( self, text, startPos, endPos ):
		self.AddRaw( text, startPos, endPos )
		return endPos

	#
	# IsClosure -- override of Block method. Treats another `elif' block
	# or an `else' block as an end to our block.
	#
	def IsClosure( self, kind ):
		return kind == 'elif' or kind == 'else' or kind == 'end'

	#
	# Cook -- override of IfBlock Cook method. Only formats its text if
	# its condition returns `true'. Returns result of condition
	# evaluation so that our parent `if' block can detect if we
	# formatted our text.
	#
	def Cook( self, sink, dict ):
		value = dict.Resolve( self.condition )
		if value:
			Block.Cook( self, sink, dict )

		return value

#
# Class ForBlock -- handles text blocks starting with a `for' statement.
# We accept sequences and dictionaries for iteration values.
#
class ForBlock( Block ):

	kBlockKind = 'for'

	#
	# __init__ -- install name and values to use for sequence/map
	# operations. Accept but do not require trailing `:' character
	#
	def __init__( self, contents, lineMode ):
		Block.__init__( self, lineMode )

		#
		# Expect data following the `for' tag formatted as <name> in
		# <data>, similar to Python's syntax.
		#
		contents = string.split( string.strip( contents ) )
		if len( contents ) < 3 or contents[ 1 ] != 'in':
			raise RuntimeError, "invalid 'for' block format: %s" % \
				  contents

		self.name = contents[ 0 ]
		contents = string.join( contents[ 2 : ] )
		if len( contents ) == 0:
			raise RuntimeError, "invalid 'for' block format: %s" % \
				  contents

		#
		# Remember data we will use for iteration.
		#
		self.data = contents
		return

	#
	# Cook -- apply the given formatter to the contents of the `for'
	# block.
	#
	def Cook( self, sink, formatter ):

		#
		# Obtain sequence/mapping to iterate over
		#
		data = formatter.Resolve( self.data )

		#
		# Create appropriate cooker based on type of data
		#
		varType = type( data )
		if varType in ( type( [] ), type( () ), type( '' ) ):

			#
			# Have a sequence -- create a cooker to format references to
			# sequence values.
			#
			cooker = ForSeqCooker( data )

		elif varType == type( {} ):

			#
			# Have a mapping -- create a cooker to format references to
			# dictionary values.
			#
			cooker = ForMapCooker( data )

		else:
			raise RuntimeError, "invalid argument for 'for' block"

		#
		# Let the cooker iterate over the elements of the sequence or
		# mapping and repeatedly cook our blocks. Install reference to
		# cooker with our name as the key.
		#
		formatter.AddValue( self.name, cooker )
		cooker.Cook( sink, formatter, self.blocks )

#
# Formatter -- handles resolution of %() format constructs within text
# blocks. Name resolution is handled thru Python's `eval' built-in
# function. This class provides a set of formattting functions that can
# be used within a %() format construct to change the data before it is
# given to Python.
#
# All functions except Null take just one argument -- the value to work
# on. The Null function takes an additional argument that is returned to
# Python if the value to be returned is None or the empty string.
#
class Formatter:

	#
	# Define the roman characters we support
	#
	kRomanSet = ( ( 'M', 1000 ), ( 'D', 500 ), ( 'C', 100 ), ( 'L', 50 ),
				  ( 'X', 10 ), ( 'V', 5 ), ( 'I', 1 ) )

	#
	# Define values that can have an alternative expression when
	# encountered. Used by __getitem__ method below.
	#
	kNullSet = ( None, '', 0 )

	#
	# __init__ -- remember instance and dictionary to use for evaluation
	# of formatting names and expressions.
	#
	def __init__( self, env ):
		self.env = env
		self.InitProcs()

	#
	# InitProcs -- install functions available for formatting values
	#
	def InitProcs( self ):
		dict = self.procs = {}
		dict[ 'Upper' ] = self.Upper
		dict[ 'Lower' ] = self.Lower
		dict[ 'Capitalize' ] = self.Capitalize
		dict[ 'Spacify' ] = self.Spacify
		dict[ 'Null' ] = self.Null
		dict[ 'Checked' ] = self.Null
		dict[ 'Selected' ] = self.Null
		dict[ 'Letter' ] = self.Letter
		dict[ 'Roman' ] = self.Roman
		dict[ 'HtmlEncode' ] = self.HtmlEncode
		dict[ 'UrlEncode' ] = self.UrlEncode
		return dict

	#
	# Upper -- returns text in uppercase
	#
	def Upper( self, value ):
		return string.upper( value )

	#
	# Lower -- returns text in lowercase
	#
	def Lower( self, value ):
		return string.lower( value )

	#
	# Capitalize -- returns text with first letter of all words in
	# uppercase and the rest in lowercase.
	#
	def Capitalize( self, value ):
		return string.join( map( lambda a: string.upper( a[ 0 ] ) + \
								 string.lower( a[ 1 : ] ),
								 string.split( value ) ) )

	#
	# Spacify -- convert all `_' characters into spaces
	#
	def Spacify( self, value ):
		return string.join( string.splitfields( value, '_' ) )

	#
	# Null -- use alternative value if main one is Python None.
	#
	def Null( self, value, altValue ):
		return value or altValue

	#
	# Checked -- if value is not false, returns 'CHECKED'
	#
	def Checked( self, value ):
		return ( value and ' CHECKED' ) or ''

	#
	# Selected -- if value is not false, returns 'SELECTED'
	#
	def Selected( self, value ):
		return ( value and ' SELECTED' ) or ''

	#
	# Letter -- returns letter that corresponds to the given number.
	#
	def Letter( self, value ):
		if value >= 1 and value <= '26':
			value = chr( ord( 'A' ) + value )
		else:
			value = '?'
		return value

	#
	# Roman -- converts the value into Roman numerals.
	#
	def Roman( self, value ):
		from regsub import sub
		import array

		acc = array.array( 'c' )
		for ( char, charVal ) in self.kRomanSet:
			while value >= charVal:
				value = value - charVal
				acc.append( char )

		#
		# Substitute acceptable contractions.
		#
		acc = sub( 'DCCCC', 'CM', acc.tostring() )
		acc = sub( 'CCCC', 'CD', acc )
		acc = sub( 'LXXXX', 'XC', acc )
		acc = sub( 'XXXX', 'XL', acc )
		acc = sub( 'VIIII', 'IX', acc )
		acc = sub( 'IIII', 'IV', acc )
		return acc

	#
	# HtmlEncode -- convert certain characters into HTML equivalents
	#
	def HtmlEncode( self, value ):
		jf = string.joinfields
		sf = string.splitfields
		value = jf( sf( value, '&' ), '&amp;' )
		value = jf( sf( value, '<' ), '&lt;' )
		value = jf( sf( value, '>' ), '&gt;' )
		value = jf( sf( value, '"' ), '&quot;' )
		return value

	#
	# UrlEncode -- convert certain characters into safe HTML variants.
	#
	def UrlEncode( self, value ):
		import urllib
		return urllib.quote( value )

	#
	# AddValue -- install a value under a particular key.
	#
	def AddValue( self, key, value ):
		self.procs[ key ] = value

	#
	# Resolve -- attempts to evaluate given key using the contents of
	# our formatting dictionary and our active environment.
	#
	def Resolve( self, key ):
		return eval( key, self.procs, self.env )

	#
	# __getitem__ -- provides rvalue indexing. Called by Python kernel
	# to resolve name values for `%' formats.
	#
	def __getitem__( self, key ):
		return self.Resolve( key )

#
# Class ForCooker -- performs the formatting of `for' blocks. Installs
# attributes in the formatting dictionary so that %() format constructs
# can reference the current index, counter, and value of the iterator,
# as well as statistical values calculated for numeric sequences.
#
class ForCooker:

	#
	# __init__ -- remembers name of iterator attribute and the data to
	# iterate over (must be a sequence or dictionary). Installs
	# ourselves into the given format dictionary under the name of the
	# iterator.
	#
	def __init__( self, data ):
		self.data = data

	#
	# SetStats -- attempt to calculate various statistical values from
	# the sequence given to us. If a non-numeric value is encountered,
	# the attempt is aborted, and incomplete statistics will result.
	# NOTE: sequence is sorted in-place.
	#
	def SetStats( self, values ):
		import math

		#
		# Create/init statistic attributes. These can be reference
		# within a %() construct with a name of tagName + `.' +
		# statName.
		#
		self.count = 0
		self.sum = 0
		self.mean = 0.0
		self.min = None
		self.max = None

		count = self.count = len( values )
		if count > 0:

			#
			# Ignore any type/value conflicts.
			#
			try:
				sum = 0;

				#
				# Calculate sum and sum of squares for use in other
				# statistics.
				#
				for each in values:

					#
					# If we have a string, try to make it a number.
					#
					if type( each ) == type( '' ):
						try:
							each = string.atoi( each )
						except ValueError:
							each = string.atof( each )

					#
					# Calculate values
					#
					sum = sum + each

				self.sum = sum
				self.mean = float( sum ) / count

			except ( TypeError, ValueError ):
				pass

			#
			# Sort the values so we can calculate the median value and
			# min/max values.
			#
			values.sort()
			self.min = values[ 0 ]
			self.max = values[ count - 1 ]

			#
			# Calculate the median element. Attempt to handle
			# non-numeric elements.
			#
			if count == 1:
				self.median = values[ 0 ]
			else:
				index = count / 2
				self.median = values[ index ]
				if index * 2 == count:
					try:
						self.median = ( values[ index - 1 ] +
										values[ index ] ) / 2
					except TypeError:
						pass

	def CookBlocks( self, sink, formatter, blocks ):
		for each in blocks:
			each.Cook( sink, formatter )

#
# ForSeqCooker -- provides resolution of sequencing variables available
# within a `for' block.
#
class ForSeqCooker( ForCooker ):

	#
	# __init__ -- install attributes for sequences.
	#
	def __init__( self, data ):
		ForCooker.__init__( self, data )

		#
		# Initialize iterator attributes available inside the blocks
		#
		self.counter = 0
		self.index = 0
		self.value = None

		#
		# Attempt to calculate statistics from the data. NOTE: we make a
		# copy of the sequence because SetStats sorts.
		#
		self.SetStats( list( copy.copy( data ) ) )

	#
	# Cook -- visit the given blocks as many times as there are elements
	# in the sequence we are using. Updates iterator attributes.
	#
	def Cook( self, sink, formatter, blocks ):
		for each in self.data:
			self.value = each
			self.counter = self.counter + 1
			self.CookBlocks( sink, formatter, blocks )
			self.index = self.index + 1

#
# ForMapCooker -- provides resolution of sequencing variables available
# within a `for' block.
#
class ForMapCooker( ForCooker ):

	#
	# __init__ -- install attributes for dictionaries.
	#
	def __init__( self, data ):
		ForCooker.__init__( self, data )

		#
		# Initialize iterator attributes available inside the blocks
		#
		self.counter = 0
		self.key = None
		self.value = None

		#
		# Attempt to calculate statistics from the values found in the
		# dictionary. Unlike ForSeqDict, we don't need to make a copy of
		# this.
		#
		self.SetStats( data.values() )

	#
	# Cook -- visit the given blocks as many times as there are elements
	# in the mapping we are using. Updates iterator attributes.
	#
	def Cook( self, sink, formatter, blocks ):
		data = self.data
		for key in data.keys():
			self.key = key
			self.value = data[ key ]
			self.counter = self.counter + 1
			self.CookBlocks( sink, formatter, blocks )

#
# Class BoilerPlate -- abstract base class that combines text block
# parsing and dictionary management for formatting.
#
class BoilerPlate( Block ):

	#
	# __init__ -- start with an empty dictionary of formatting values.
	#
	def __init__( self, lineMode = None ):
		Block.__init__( self, lineMode )
		self.dict = {}
		self.cooked = None
		self.formatter = None

	#
	# MakeFormatter -- create instance for text formatting
	#
	def MakeFormatter( self, dict ):
		return Formatter( dict )

	#
	# InstallDicts -- copy given dictionary values into our formatting
	# dictionary. *Internal method*
	#
	def InstallDicts( self, dict, kw ):
		tmp = self.dict
		if dict:
			for key in dict.keys():
				tmp[ key ] = dict[ key ]

		if kw:
			for key in kw.keys():
				tmp[ key ] = kw[ key ]

	#
	# Remeber -- copy key/values found in the given dictionary and
	# keywords to our formatting dictionary.
	#
	def Remember( self, dict = None, **kw ):
		self.InstallDicts( dict, kw )

	#
	# Reset -- forget any formatted (cooked) text
	#
	def Reset( self ):
		self.cooked = None

	#
	# Cook -- format our text by applying whatever dictionary entries we
	# have to our text blocks.
	#
	def Cook( self ):
		if self.cooked == None:
			formatter = self.formatter
			if formatter == None:
				formatter = self.formatter = \
							self.MakeFormatter( self.dict )
			sink = Sink()
			Block.Cook( self, sink, formatter )
			self.cooked = sink.Value()
		return self.cooked

	#
	# GetOutput -- apply contents of given dictionary and keywords,
	# format text, and return string result. *Internal Method*
	#
	def GetOutput( self, dict, kw ):
		if dict or kw:
			self.Reset()
			self.InstallDicts( dict, kw )
		return self.Cook()

	#
	# Value -- return result of applying the given dictionary and
	# keywords to our template value.
	#
	def Value( self, dict = None, **kw ):
		return self.GetOutput( dict, kw )

	#
	# __call__ -- enable Python call semantics for class instances.
	#
	def __call__( self, dict = None, **kw ):
		return self.GetOutput( dict, kw )

	#
	# __str__ -- enable Python string conversion for class instances.
	#
	def __str__( self ):
		return self.Cook()

	#
	# __repr__ -- return representation string to use for instances.
	# Defined because __str__ is defined above.
	#
	def __repr__( self ):
		return '<%s instance at %8.8X>' % \
			   ( self.__class__.__name__, id( self ) )

#
# Class String -- class that uses Python string as source for text.
#
class String( BoilerPlate ):

	#
	# __init__ -- process the given text into blocks
	#
	def __init__( self, text, dict = None, lineMode = None, **kw ):
		BoilerPlate.__init__( self, lineMode )
		self.InstallDicts( dict, kw )
		self.EatText( text, 0 )

#
# Class File -- class that uses a file as source for text.
#
class File( BoilerPlate ):

	#
	# __init__ -- remember name of file to use and load it.
	#
	def __init__( self, name, dict = None, lineMode = None, **kw ):
		BoilerPlate.__init__( self, lineMode )
		self.InstallDicts( dict, kw )
		self.name = name
		self.Reload()

	#
	# Reload -- cause instance to reload template from file and reparse
	# it. Text is not reformatted but will be when asked for.
	#
	def Reload( self ):

		#
		# Forget any cached formatted text.
		#
		self.Reset()

		#
		# Read file contents
		#
		fd = open( self.name )
		text = fd.read()
		fd.close()

		#
		# Rebuild list of blocks from text.
		#
		self.EatText( text, 0 )

#
# Class FileWatch -- derivative of File that detects when its text
# blocks are older than the template file they were derived from.
#
class FileWatch( File ):

	#
	# Reload -- override of File method. Sets the time when we parsed
	# the template.
	#
	def Reload( self ):
		self.timestamp = int( time.time() )
		File.Reload( self )

	#
	# Expired -- returns true if our text blocks are older than the
	# template file they were generated from.
	#
	def Expired( self ):
		stale = 0

		#
		# Ignore errors because file does not exist anymore.
		#
		try:
			stale = self.timestamp < \
					os.stat( self.name )[ stat.ST_MTIME ]
		except os.error:
			pass

		return stale

	#
	# Cook -- override of BoilerPlate method. Checks for newer template
	# data, and if necessary reprocesses it.
	#
	def Cook( self ):
		if self.Expired():
			self.Reload()
		return File.Cook( self )
