%{
/*                         MICRION SOFTWARE DEPT.
 *          Copyright (c) 1994 by Micrion Corp. All rights reserved.
 *
 * FILE NAME    : mvcl.y
 * AUTHOR       : R. Mello
 * DATE WRITTEN : 12/12/94
 * VERSION      : 1.00
 * USAGE        : grammar processor for the mvcl programming language
 * 
 *  This file contains a yacc grammar for the micrion ( M ) vacuum ( V )
 * command ( C ) language ( L ).
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <mvcl_int.h>

/* uncomment the following line to compile the main() in for the Mvcl
 * shell interpreter.
 */
/* #define MVCL_STAND_ALONE */

/* uncomment the following line to test unusual mvcl functions -- this is only
 * available if MVCL_STAND_ALONE is defined.
 */
/* #define TESTING_MVCL */

/* file equates. */
#define YYDEBUG 1

/* external reference to error func */
extern void  yyerror( char *s );

/* local globals */
static MvclProgramList *prog = (MvclProgramList *)NULL;
static MvclProgramNode *Node;
static char            *tmp;
static int             ret;
static MvclPrintItem   *plist = (MvclPrintItem *)NULL;
%}

%union
  {
   double   real;
   int	    integer;
   char	    *string;
  }

%token  PROC END
%token  IF THEN ELSE ENDIF
%token  FOR TO DOWNTO ENDFOR
%token  OPEN CLOSE READ SET STATUS WAIT TIMEOUT EXTEND RETRACT
%token  WAITFOR PUMPON PUMPOFF EXIT PERROR PRINT PRETURN
%token  V P HCIG CCIG BT TC NZ PT MFC
%token  NAME SEMICOLON COMPARISON ASSIGNMENT QSTRING IDENTIFIER INTARG REALARG
%token  SHELL_START STREAM_START STRING_START
%token	INTEGER REAL
%token  UNKNOWN
%token  HELP 

%type <string>  QSTRING
%type <integer> V P HCIG CCIG BT TC NZ PT MFC
%type <string>  IDENTIFIER
%type <integer> dev_class dev_id
%type <integer> COMPARISON
%type <integer> INTEGER
%type <real>    REAL
%type <string>  expr
 
%left '+' '-'
%left '*' '/'
%nonassoc UMINUS

%%
 /* Three grammars can be parsed here. One that will accept interactive
  * commands. Interpreting each command as it is entered. One that
  * will accept a stream input. Parse the stream into a list of program
  * nodes. Then pass the resultant list through the interpreter. Finally a
  * grammar that will accept input from a string.
  */
combined:		SHELL_START  mvcl_shell
	|		STREAM_START mvcl_stream
        |               STRING_START mvcl_string
	;

 /* ================= The stream grammar starts here ================== */

mvcl_stream:		proc_statement mvcl_statements end_program
	;

proc_statement:		PROC QSTRING SEMICOLON
	;

mvcl_statements: 	mvcl_stream_statements SEMICOLON
	|	 	mvcl_stream_statements SEMICOLON mvcl_statements
	;

mvcl_stream_statements: status_statement 
	|		read_statement
        |               mvcl_statement
        |               cond_statements
        |               loop_statements
	|		return_statement
        ;

status_statement:	IDENTIFIER ASSIGNMENT STATUS dev_class dev_id
                        { if( BuildMvclStatusNode( prog, $1, $4, $5 ) )
			    YYABORT;
			}
	;

read_statement:		IDENTIFIER ASSIGNMENT READ dev_class dev_id
			{ if( BuildMvclReadNode( prog, $1, $4, $5 ) )
			    YYABORT;
			}
	;

cond_statements:       if_statement
        |              if_else_statement
        ;

if_statement:          IF if_test mvcl_statements endif_statement
        ;

if_else_statement:     IF if_test mvcl_statements else_statement
                       mvcl_statements endif_statement
        ;

else_statement:        ELSE
                       {
                        if( BuildMvclElseNode( prog ) )
                          YYABORT;
                       }
        ;

endif_statement:       ENDIF
                       {
                        if( BuildMvclEndifNode( prog ) )
                          YYABORT;
		       }
        ;

if_test:               expr COMPARISON expr
                       {
                        if( BuildMvclIfNode( prog, $1, $2, $3 ) )
			  YYABORT;
                       }
        ;

loop_statements:       for_statement
        ;

for_statement:         FOR for_args mvcl_statements endfor_statement
        ;

for_args:              IDENTIFIER ASSIGNMENT expr TO expr
                       {
                        if( BuildMvclForNode( prog, $1, $3, MVCL_FOR_TO, $5 ) )
                          YYABORT;
		       }
        |              IDENTIFIER ASSIGNMENT expr DOWNTO expr
                       {
                        if( BuildMvclForNode(prog, $1, $3, MVCL_FOR_DOWNTO,$5))
                          YYABORT;
		       }
        ;

endfor_statement:      ENDFOR
                       {
                        if( BuildMvclEndForNode( prog ) )
                          YYABORT;
		       }
        ;

expr:                  IDENTIFIER  { $$ = $1;                               }
        |              REAL        { $$ = MvclCreateRealSymbol( prog, $1 ); }
        |              INTEGER     { $$ = MvclCreateIntSymbol( prog, $1 );  }
        |              QSTRING
                       { MvclParserError( "Incompatible Types\n" );
                         YYABORT;
                       }
        ;

return_statement:      PRETURN INTEGER
		       {
                        BuildMvclReturnNode( prog, $2 );
                       }

end_program:	       END SEMICOLON
	;

 /* ================= The string grammar starts here ================== */

mvcl_string:           mvcl_string_statements
        ;

mvcl_string_statements: /* empty */
        |               mvcl_shell_commands
        ;

 /* ================= The shell grammar starts here ================== */

mvcl_shell:		mvcl_shell_statements end_shell
	;

mvcl_shell_statements:  mvcl_shell_commands 
    	|		mvcl_shell_commands mvcl_shell_statements
	;

mvcl_shell_commands:    status_command
        |               read_command
        |               help_command
        |               mvcl_statement
        ;

status_command:         STATUS dev_class dev_id
                        { tmp = MvclGenerateVariable();
                          BuildMvclStatusNode( prog, tmp, $2, $3 );
			}
	;

read_command:		READ dev_class dev_id
                        { tmp = MvclGenerateVariable();
			  BuildMvclReadNode( prog, tmp, $2, $3 );
			}
	;

help_command:           HELP
                        { MvclShellHelp();  }

end_shell:		END
			{ exit( 0 ); }
	;

 /* ==================== Common grammar elements ===================== */

mvcl_statement:		open_statement
	|		close_statement
	|		timeout_statement
	|		set_statement
	|		pumpon_statement
	|		pumpoff_statement
	|		extend_statement
	|		retract_statement
	|		wait_statement
	|		waitfor_statement
        |               print_statement
        |               error
                        { if( MvclGetParserMode() == MVCL_STREAM )
                            YYABORT;
			}
	;

open_statement:		OPEN V dev_id
			{ if( BuildMvclOpenNode( prog, $3 ) )
                            YYABORT;
			}
	;

close_statement:	CLOSE V dev_id
			{ if( BuildMvclCloseNode( prog, $3 ) )
			    YYABORT;
                        }
	;


timeout_statement:	TIMEOUT REAL
			{ if( BuildMvclTimeoutNode( prog, $2 ) )
			    YYABORT;
			}
	;

set_statement:		SET dev_class dev_id REAL
			{ if( BuildMvclSetNode( prog, $2, $3, $4 ) )
			    YYABORT;
 			}
	;

pumpon_statement:	PUMPON P dev_id
			{ if( BuildMvclPumpOnNode( prog, $3 ) )
			    YYABORT;
			}
	;

pumpoff_statement:	PUMPOFF P dev_id
			{ if( BuildMvclPumpOffNode( prog, $3 ) )
			    YYABORT;
			}
	;
      
extend_statement:	EXTEND NZ dev_id
			{ if( BuildMvclExtendNode( prog, $3 ) )
			    YYABORT;
			}
	;

retract_statement:	RETRACT NZ dev_id
			{ if( BuildMvclRetractNode( prog, $3 ) )
			    YYABORT;
			}
	;

wait_statement:		WAIT REAL
			{ if( BuildMvclWaitNode( prog, $2 ) )
			    YYABORT;
			}
	;

waitfor_statement:	WAITFOR dev_class dev_id COMPARISON REAL
 			{ if( BuildMvclWaitForNode( prog, $2, $3, $4, $5 ))
			    YYABORT;
			}
	;

print_statement:        PRINT print_list
                        { ret = BuildMvclPrintNode( prog, plist );

                          /* init for another list. */
                          plist = (MvclPrintItem *)NULL;
                    
                          if( ret )
                            YYABORT;
                        }
        ;

print_list:             print_item
        |               print_item ',' print_list
        ;

print_item:             QSTRING
                        { MvclAddToPrintList( &plist, $1, MVCL_PRINT_STR );
                        }
        |               IDENTIFIER
                        { MvclAddToPrintList( &plist, $1, MVCL_PRINT_IDENT );
                        }
        |               REAL
                        { tmp = MvclCreateRealSymbol( prog, $1 );
                         MvclAddToPrintList( &plist, tmp, MVCL_PRINT_REAL );
                        }
        |               INTEGER
                        { tmp = MvclCreateIntSymbol( prog, $1 );
                         MvclAddToPrintList( &plist, tmp, MVCL_PRINT_INT );
                        }
        ;
        

dev_class:		V        { $$ = MVCL_VALVE_CLASS;	}
	|		P        { $$ = MVCL_PUMP_CLASS;    	}
	|		HCIG     { $$ = MVCL_HCIG_CLASS;	}
	|		CCIG     { $$ = MVCL_CCIG_CLASS;	}
	|		BT       { $$ = MVCL_BT_CLASS;		}
	|		TC       { $$ = MVCL_TC_CLASS;		}
	|		NZ       { $$ = MVCL_NZ_CLASS;		}
	|		PT       { $$ = MVCL_PT_CLASS;		}
	|		MFC      { $$ = MVCL_MFC_CLASS;		}
	;
     
dev_id:			/* empty */ { $$ = 0; 			}
	|		INTEGER  { $$ = $1;			}
	;
%%
/***************************************************************************/
int  MvclSetParserProgramList(
   MvclProgramList  *new_list )

  {
   /* MvclSetParserProgramList() -- This function will allow an outside source
    *     to set the current program list pointer to a list of thier own
    *     creation.
    *
    *     RETURNS: 1 if all went well
    *              0 if not.
    */

   prog = new_list;

   return( 1 );
  }
/***************************************************************************/
MvclProgramList *   MvclGetParserProgramList( void )

  {
   /* MvclGetParserProgramList() -- This function will return the pointer
    *     that points to the current mvcl program.
    *
    *     RETURNS: A pointer to the current program.
    */

   return( prog );
  }
/***************************************************************************/
#ifdef MVCL_STAND_ALONE
int main(
   int   argc,
   char  **argv )

  {
   /* main() -- O/S transfer point.
    *      Check for any command line arguments. By default the shell will
    *      run in interactive mode ( accepting commands ) if a file name is
    *      present on the command line then that file will be executed as a
    *      mvcl program.
    *
    *      RETURNS: 0 always.
    */
   int     ret;
   int     script_ret;
   int     ch;
   int     source_mode = 0;
   char    *source_fname;
   extern int   opterr;
   extern int   optind;
   extern char  *optarg;
   
   /* process the command line arguments -- let getopt report the errors */
   while( ( ch = getopt( argc, argv, "df:" ) ) != EOF )
     {
      switch( ch )
        {
	 case 'd':
           yydebug = 1;
           break;

         case 'f':
           source_mode = 1;
           source_fname = optarg;
           break;

         case '?':
         default:
           fprintf( stderr, "Usage: mvclshell [-d] [-f <file>]\n" );
	   exit( 1 );
           break;
        }           
     }

   /* Create an instance of an mvcl program. */
   prog = CreateMvclProgramList();

   /* Init mvcl data structures. */
   MvclQueryVacuumSystem( "../VacConfig.M" );

   if( source_mode == 1 )
     {
      /* make an attempt to parse the file named on the command line */
      fprintf( stderr, "Running The Program %s...\n", source_fname );

      ret = MvclRun( source_fname, &script_ret );

      if( ret == -1 )
        fprintf( stderr, "Parser Failed, Program NOT Interpreted...\n" );
      else if( ret == -2 )
        fprintf( stderr, "Program Failed...\n" );
      else if( ret == -3 )
        fprintf( stderr, "Error Opening File <%s>...\n", source_fname );
      else
        fprintf( stderr, "Done...\n" );
     }
   else
     MvclShell();

   /* Clear up memory before quitting ( just to be tidy ). */
   DestroyMvclProgramList( prog );

   return( 0 );
  }
#endif
/***************************************************************************/
/* finis */











