/*                         MICRION SOFTWARE DEPT.
 *          Copyright (c) 1995 by Micrion Corp. All rights reserved.
 *
 * FILE NAME    : mvcl_build.c
 * AUTHOR       : R. Mello
 * DATE WRITTEN : 3/25/95
 * VERSION      : 1.00
 * USAGE        : module of the mcvl parser
 * 
 * This source file will be responsible for building the program nodes
 * for the interpreter.
 */
#include <stdio.h>                          /* for NULL, fprintf(), ect...   */
#include <malloc.h>
#include <mvcl_int.h>

/* function prototypes. */
int   BuildMvclOpenNode( MvclProgramList *, int );
int   BuildMvclCloseNode( MvclProgramList *, int );
int   BuildMvclStatusNode( MvclProgramList *, char *, int, int );
int   BuildMvclTimeoutNode( MvclProgramList *, double );
int   BuildMvclPumpOnNode( MvclProgramList *, int );
int   BuildMvclPumpOffNode( MvclProgramList *, int );
int   BuildMvclExtendNode( MvclProgramList *, int );
int   BuildMvclRetractNode( MvclProgramList *, int );
int   BuildMvclSetNode( MvclProgramList *, int, int, double );
int   BuildMvclReadNode( MvclProgramList *, char *, int, int );
int   BuildMvclWaitNode( MvclProgramList *, double );
int   BuildMvclWaitForNode( MvclProgramList *, int, int, int, double );
int   BuildMvclIfNode( MvclProgramList *, char *, int, char * );
int   BuildMvclElseNode( MvclProgramList * );
int   BuildMvclEndifNode( MvclProgramList * );
int   BuildMvclForNode( MvclProgramList *, char *, char *, int, char * );
int   BuildMvclEndForNode( MvclProgramList * );
int   BuildMvclPrintNode( MvclProgramList *, MvclPrintItem * );
int   BuildMvclReturnNode( MvclProgramList *, int );

/* local functions */
static MvclProgramNode *  BuildMvclNullNode( MvclProgramList * );

/***************************************************************************/
int  BuildMvclOpenNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   int    valve )                          /* valve to open                */

  {
   /* BuildMvclOpenNode() -- This function will generate the required data
    *     structures for a open statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Validate the argument(s) first. If it's ok create the node. */
   if( MvclValidateValve( valve ) )
     {
      /* Allocate memory for the new program node. */
      new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

      if( new != (MvclProgramNode *)NULL )
        {
         /* now let's allocate memory for the statement. */
         new->def.OpenStatement =
         (MvclOpenStruct *)malloc( sizeof( MvclOpenStruct ) );

         if( new->def.OpenStatement != (MvclOpenStruct *)NULL )
           {
            /* We have allocated the memory now assign the specifics. */
            (new->def.OpenStatement)->valve = valve;

            /* basic housekeeping. */
            new->StatementType = MVCL_OPEN;
            new->next = (MvclProgramNode *)NULL;

            /* add the new node to the program list. */
	    AddMvclProgramNode( prog, new );
           }
         else
           {
            free( new );
            new = (MvclProgramNode *)NULL;
           }
        }
     }
   else
     MvclParserError( "Valve %d Is Not Valid.\n", valve );

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/***************************************************************************/
int  BuildMvclCloseNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   int    valve )

  {
   /* BuildMvclCloseNode() -- This function will generate the required data
    *     structures for a close statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Validate the argument(s) first. If it's ok create the node. */
   if( MvclValidateValve( valve ) )
     {
      /* Allocate memory for the new program node. */
      new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

      if( new != (MvclProgramNode *)NULL )
        {
         /* now let's allocate memory for the statement. */
         new->def.CloseStatement =
         (MvclCloseStruct *)malloc( sizeof( MvclCloseStruct ) );

         if( new->def.CloseStatement != (MvclCloseStruct *)NULL )
           {
            /* We have allocated the memory now assign the specifics. */
            (new->def.CloseStatement)->valve = valve;

            /* basic housekeeping. */
            new->StatementType = MVCL_CLOSE;
            new->next = (MvclProgramNode *)NULL;

            /* add the new node to the program list. */
	    AddMvclProgramNode( prog, new );
           }
         else
           {
            free( new );
            new = (MvclProgramNode *)NULL;
           }
        }
     }
   else
     MvclParserError( "Valve %d Is Not Valid.\n", valve );

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/***************************************************************************/
int  BuildMvclStatusNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   char   *identifier,                     /* ident to enter into sym tab  */
   int    device_class,			   /* class of device to check     */
   int    device_id )			   /* specifiec device id          */

  {
   /* BuildMvclStatusNode() -- This function will generate the required data
    *     structures for a status statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Validate the argument(s) first. If it's ok create the node. */
   if( MvclValidateDevice( device_class, device_id ) )
     {
      /* Allocate memory for the new program node. */
      new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

      if( new != (MvclProgramNode *)NULL )
        {
         /* now let's allocate memory for the statement. */
         new->def.StatusStatement =
         (MvclStatusStruct *)malloc( sizeof( MvclStatusStruct ) );

         if( new->def.StatusStatement != (MvclStatusStruct *)NULL )
           {
            /* We have allocated the memory now assign the specifics. */
            (new->def.StatusStatement)->dev_class = device_class;
            (new->def.StatusStatement)->dev_id = device_id;
            (new->def.StatusStatement)->SymAddress = (void *)NULL;

            /* Allocate s symbol table entry for this identifier. */
            if( !IsMvclSymbolInTable( prog->SymbolTable, identifier ) )
              {
               (new->def.StatusStatement)->SymAddress =
               AddMvclSymbolToTable( prog->SymbolTable, identifier,
                                     MVCL_SYM_INT, 0 );
	      }
            else
              MvclRunTimeError( "Symbol '%s' Has Been Previously Defined.\n",
                                identifier );

            /* basic housekeeping. */
            new->StatementType = MVCL_STATUS;
            new->next = (MvclProgramNode *)NULL;

            /* add the new node to the program list. */
	    AddMvclProgramNode( prog, new );
           }
         else
           {
            free( new );
            new = (MvclProgramNode *)NULL;
           }
        }
     }
   else
     MvclParserError( "%s Is Not Valid.\n",
                       MvclResolveDeviceName( device_class, device_id ) );

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/**************************************************************************/
int  BuildMvclTimeoutNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   double    timeout )                     /* in seconds                  */

  {
   /* BuildMvclTimeoutNode() -- This function will generate the required data
    *     structures for a timeout statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Allocate memory for the new program node. */
   new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

   if( new != (MvclProgramNode *)NULL )
     {
      /* now let's allocate memory for the statement. */
      new->def.TimeoutStatement =
      (MvclTimeoutStruct *)malloc( sizeof( MvclTimeoutStruct ) );

      if( new->def.TimeoutStatement != (MvclTimeoutStruct *)NULL )
        {
         /* We have allocated the memory now assign the specifics. */
         (new->def.TimeoutStatement)->timeout = timeout;

         /* basic housekeeping. */
         new->StatementType = MVCL_TIMEOUT;
         new->next = (MvclProgramNode *)NULL;

         /* add the new node to the program list. */
         AddMvclProgramNode( prog, new );
        }
      else
        {
         free( new );
         new = (MvclProgramNode *)NULL;
        }
     }

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/**************************************************************************/
int  BuildMvclPumpOnNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   int       pump )                        /* in seconds                  */

  {
   /* BuildMvclPumpOnNode() -- This function will generate the required data
    *     structures for a pump on statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Validate the argument(s) first. If it's ok create the node. */
   if( MvclValidateDevice( MVCL_PUMP_CLASS, pump ) )
     {
      /* Allocate memory for the new program node. */
      new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

      if( new != (MvclProgramNode *)NULL )
        {
         /* now let's allocate memory for the statement. */
         new->def.PumpOnStatement =
         (MvclPumpOnStruct *)malloc( sizeof( MvclPumpOnStruct ) );

         if( new->def.PumpOnStatement != (MvclPumpOnStruct *)NULL )
           {
            /* We have allocated the memory now assign the specifics. */
            (new->def.PumpOnStatement)->pump = pump;

            /* basic housekeeping. */
            new->StatementType = MVCL_PUMPON;
            new->next = (MvclProgramNode *)NULL;

            /* add the new node to the program list. */
            AddMvclProgramNode( prog, new );
           }
         else
           {
            free( new );
            new = (MvclProgramNode *)NULL;
           }
        }
     }
   else
     MvclParserError( "Pump %d Is Not Valid.\n", pump );

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/**************************************************************************/
int  BuildMvclPumpOffNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   int       pump )                        /* in seconds                  */

  {
   /* BuildMvclPumpOffNode() -- This function will generate the required data
    *     structures for a pump on statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Validate the argument(s) first. If it's ok create the node. */
   if( MvclValidateDevice( MVCL_PUMP_CLASS, pump ) )
     {
      /* Allocate memory for the new program node. */
      new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

      if( new != (MvclProgramNode *)NULL )
        {
         /* now let's allocate memory for the statement. */
         new->def.PumpOffStatement =
         (MvclPumpOffStruct *)malloc( sizeof( MvclPumpOffStruct ) );

         if( new->def.PumpOffStatement != (MvclPumpOffStruct *)NULL )
           {
            /* We have allocated the memory now assign the specifics. */
            (new->def.PumpOffStatement)->pump = pump;

            /* basic housekeeping. */
            new->StatementType = MVCL_PUMPOFF;
            new->next = (MvclProgramNode *)NULL;

            /* add the new node to the program list. */
            AddMvclProgramNode( prog, new );
           }
         else
           {
            free( new );
            new = (MvclProgramNode *)NULL;
           }
        }
     }
   else
     MvclParserError( "Pump %d Is Not Valid.\n", pump );

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/**************************************************************************/
int   BuildMvclExtendNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   int       nozzle )                      /* in seconds                  */

  {
   /* BuildMvclExtendNode() -- This function will generate the required data
    *     structures for a nozzle extend statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Validate the argument(s) first. If it's ok create the node. */
   if( MvclValidateDevice( MVCL_NZ_CLASS, nozzle ) )
     {
      /* Allocate memory for the new program node. */
      new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

      if( new != (MvclProgramNode *)NULL )
        {
         /* now let's allocate memory for the statement. */
         new->def.ExtendStatement =
         (MvclExtendStruct *)malloc( sizeof( MvclExtendStruct ) );

         if( new->def.ExtendStatement != (MvclExtendStruct *)NULL )
           {
            /* We have allocated the memory now assign the specifics. */
            (new->def.ExtendStatement)->nozzle = nozzle;

            /* basic housekeeping. */
            new->StatementType = MVCL_EXTEND;
            new->next = (MvclProgramNode *)NULL;

            /* add the new node to the program list. */
            AddMvclProgramNode( prog, new );
           }
         else
           {
            free( new );
            new = (MvclProgramNode *)NULL;
           }
        }
     }
   else
     MvclParserError( "Nozzle %d Is Not Valid.\n", nozzle );

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/**************************************************************************/
int   BuildMvclRetractNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   int       nozzle )                      /* in seconds                  */

  {
   /* BuildMvclRetractNode() -- This function will generate the required data
    *     structures for a nozzle Retract statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Validate the argument(s) first. If it's ok create the node. */
   if( MvclValidateDevice( MVCL_NZ_CLASS, nozzle ) )
     {
      /* Allocate memory for the new program node. */
      new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

      if( new != (MvclProgramNode *)NULL )
        {
         /* now let's allocate memory for the statement. */
         new->def.RetractStatement =
         (MvclRetractStruct *)malloc( sizeof( MvclRetractStruct ) );

         if( new->def.RetractStatement != (MvclRetractStruct *)NULL )
           {
            /* We have allocated the memory now assign the specifics. */
            (new->def.RetractStatement)->nozzle = nozzle;

            /* basic housekeeping. */
            new->StatementType = MVCL_RETRACT;
            new->next = (MvclProgramNode *)NULL;

            /* add the new node to the program list. */
            AddMvclProgramNode( prog, new );
           }
         else
           {
            free( new );
            new = (MvclProgramNode *)NULL;
           }
        }
     }
   else
     MvclParserError( "Nozzle %d Is Not Valid.\n", nozzle );

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/**************************************************************************/
int   BuildMvclSetNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   int       dev_class,                    /* class of device              */
   int       dev_id,                       /* specific device              */
   double    dev_value )                   /* value to set device to       */

  {
   /* BuildMvclSetNode() -- This function will generate the required data
    *     structures for a set statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Validate the argument(s) first. If it's ok create the node. */
   if( MvclValidateDevice( dev_class, dev_id ) )
     {
      /* Allocate memory for the new program node. */
      new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

      if( new != (MvclProgramNode *)NULL )
        {
         /* now let's allocate memory for the statement. */
         new->def.SetStatement =
         (MvclSetStruct *)malloc( sizeof( MvclSetStruct ) );

         if( new->def.SetStatement != (MvclSetStruct *)NULL )
           {
            /* We have allocated the memory now assign the specifics. */
            (new->def.SetStatement)->dev_class = dev_class;
            (new->def.SetStatement)->dev_id = dev_id;
            (new->def.SetStatement)->dev_value = dev_value;

            /* basic housekeeping. */
            new->StatementType = MVCL_SET;
            new->next = (MvclProgramNode *)NULL;

            /* add the new node to the program list. */
            AddMvclProgramNode( prog, new );
           }
         else
           {
            free( new );
            new = (MvclProgramNode *)NULL;
           }
        }
     }
   else
     MvclParserError( "%s Is Not Valid.\n",
                      MvclResolveDeviceName( dev_class, dev_id ) );

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/**************************************************************************/
int   BuildMvclReadNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   char      *identifier,                  /* ident to enter into sym tab  */
   int       dev_class,                    /* class of device              */
   int       dev_id )                      /* specific device              */

  {
   /* BuildMvclReadNode() -- This function will generate the required data
    *     structures for a read statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Validate the argument(s) first. If it's ok create the node. */
   if( MvclValidateDevice( dev_class, dev_id ) )
     {
      /* Allocate memory for the new program node. */
      new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

      if( new != (MvclProgramNode *)NULL )
        {
         /* now let's allocate memory for the statement. */
         new->def.SetStatement =
         (MvclSetStruct *)malloc( sizeof( MvclSetStruct ) );

         if( new->def.SetStatement != (MvclSetStruct *)NULL )
           {
            /* We have allocated the memory now assign the specifics. */
            (new->def.SetStatement)->dev_class = dev_class;
            (new->def.SetStatement)->dev_id = dev_id;

            /* Allocate a symbol table entry for this identifier. */
            if( !IsMvclSymbolInTable( prog->SymbolTable, identifier ) )
              {
               (new->def.ReadStatement)->SymAddress =
               AddMvclSymbolToTable( prog->SymbolTable, identifier,
                                     MVCL_SYM_DOUBLE, 0 );
	      }
            else
              MvclRunTimeError( "Symbol '%s' Has Been Previously Defined.\n",
                                identifier );

            /* basic housekeeping. */
            new->StatementType = MVCL_READ;
            new->next = (MvclProgramNode *)NULL;

            /* add the new node to the program list. */
            AddMvclProgramNode( prog, new );
           }
         else
           {
            free( new );
            new = (MvclProgramNode *)NULL;
           }
        }
     }
   else
     MvclParserError( "%s Is Not Valid.\n",
                      MvclResolveDeviceName( dev_class, dev_id ) );

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/**************************************************************************/
int   BuildMvclWaitNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   double           wait_time )            /* time to wait in seconds      */

  {
   /* BuildMvclWaitNode() -- This function will generate the required data
    *     structures for a wait statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Allocate memory for the new program node. */
   new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

   if( new != (MvclProgramNode *)NULL )
     {
      /* now let's allocate memory for the statement. */
      new->def.WaitStatement =
      (MvclWaitStruct *)malloc( sizeof( MvclWaitStruct ) );

      if( new->def.WaitStatement != (MvclWaitStruct *)NULL )
        {
         /* We have allocated the memory now assign the specifics. */
         (new->def.WaitStatement)->wait_time = wait_time;

         /* basic housekeeping. */
         new->StatementType = MVCL_WAIT;
         new->next = (MvclProgramNode *)NULL;

         /* add the new node to the program list. */
         AddMvclProgramNode( prog, new );
        }
      else
        {
         free( new );
         new = (MvclProgramNode *)NULL;
        }
     }
 
   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/****************************************************************************/
int  BuildMvclWaitForNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   int       dev_class,                    /* class of device              */
   int       dev_id,                       /* specific device              */
   int       comparison,                   /* id of compare operation      */
   double    value )                       /* value to compare against     */

  {
   /* BuildMvclWaitForNode() -- This function will generate the required data
    *     structures for a waitfor statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Validate the argument(s) first. If it's ok create the node. */
   if( MvclValidateDevice( dev_class, dev_id ) )
     {
      /* Allocate memory for the new program node. */
      new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

      if( new != (MvclProgramNode *)NULL )
        {
         /* now let's allocate memory for the statement. */
         new->def.WaitForStatement =
         (MvclWaitForStruct *)malloc( sizeof( MvclWaitForStruct ) );

         if( new->def.WaitForStatement != (MvclWaitForStruct *)NULL )
           {
            /* We have allocated the memory now assign the specifics. */
            (new->def.WaitForStatement)->dev_class = dev_class;
            (new->def.WaitForStatement)->dev_id = dev_id;
            (new->def.WaitForStatement)->compare = comparison;
            (new->def.WaitForStatement)->value = value;

            /* basic housekeeping. */
            new->StatementType = MVCL_WAIT_FOR;
            new->next = (MvclProgramNode *)NULL;

            /* add the new node to the program list. */
            AddMvclProgramNode( prog, new );
           }
         else
           {
            free( new );
            new = (MvclProgramNode *)NULL;
           }
        }
     }
   else
     MvclParserError( "Device %d Is Not Valid.\n", dev_id );

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/****************************************************************************/
int  BuildMvclIfNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   char      *left_ident,                  /* left of compare identifier   */
   int       compare,                      /* enumerated comparison        */
   char      *right_ident )                /* right of compare identifier  */

  {
   /* BuildMvclIfNode() -- This function will generate the required data
    *     structures for an if statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;
   MvclRunTimeStackEntry  *entry;

   /* Allocate memory for the new program node. */
   new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

   if( new != (MvclProgramNode *)NULL )
     {
      /* now let's allocate memory for the statement. */
      new->def.IfStatement =
      (MvclIfStruct *)malloc( sizeof( MvclIfStruct ) );

      if( new->def.IfStatement != (MvclIfStruct *)NULL )
        {
         /* We have allocated the memory now assign the specifics. */
         (new->def.IfStatement)->ident_left = left_ident;
         (new->def.IfStatement)->ident_right = right_ident;
         (new->def.IfStatement)->compare = compare;

         /* basic housekeeping. */
         new->StatementType = MVCL_IF;
         new->next = (MvclProgramNode *)NULL;

         /* Allocate memory for stack node. */
         entry =(MvclRunTimeStackEntry *)malloc(sizeof(MvclRunTimeStackEntry));

         if( entry != (MvclRunTimeStackEntry *)NULL )
           {
            /* Push this node pointer onto the stack */
            entry->ValueType = MVCL_SYM_ADDRESS;
            entry->Value.Address = new;
            MvclRunTimeStackPush( prog->RunTimeStack, entry );
	   }
         else
           MvclRunTimeError( "Memory Allocation Error\n" );

         /* add the new node to the program list. */
	 AddMvclProgramNode( prog, new );
        }
      else
        {
         free( new );
         new = (MvclProgramNode *)NULL;
        }
     }

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/***************************************************************************/
int   BuildMvclElseNode(
   MvclProgramList  *prog )                 /* Pointer to program list      */

  {
   /* BuildMvclElseNode() -- This function will generate the required data
    *     structures for an else statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;
   MvclProgramNode   *address;
   int               got_it = 0;
   MvclRunTimeStackEntry  *entry;
   MvclRunTimeStackEntry  *alf;

   /* Allocate memory for the new program node. */
   new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

   if( new != (MvclProgramNode *)NULL )
     {
      /* now let's allocate memory for the statement. */
      new->def.ElseStatement =
      (MvclElseStruct *)malloc( sizeof( MvclElseStruct ) );

      if( new->def.ElseStatement != (MvclElseStruct *)NULL )
        {
         /* basic housekeeping. */
         new->StatementType = MVCL_ELSE;
         new->next = (MvclProgramNode *)NULL;

         /* pop what should be the corresponding if node off of the stack. */
         entry = MvclRunTimeStackPop( prog->RunTimeStack );

         if( entry != (MvclRunTimeStackEntry *)NULL )
           {
            if( entry->ValueType == MVCL_SYM_ADDRESS )
              {
               /* see if it's an if statement. */
               address = entry->Value.Address;    

               if( address->StatementType == MVCL_IF )
                 {
                  /* Good. Now put the address of this node in the if
                   * statement. See the interpreter for more info.
                   */
                  (address->def.IfStatement)->FalseAddress = new;
                  got_it = 1;
                 } 
	      }

            /* Push back the if node address -- we may need it later. */
            MvclRunTimeStackPush( prog->RunTimeStack, entry );

            /* Allocate memory for stack node. */
            alf = (MvclRunTimeStackEntry *)malloc(sizeof(MvclRunTimeStackEntry));

            if( alf != (MvclRunTimeStackEntry *)NULL )
              {
               /* Push this node pointer onto the stack */
               alf->ValueType = MVCL_SYM_ADDRESS;
               alf->Value.Address = new;
               MvclRunTimeStackPush( prog->RunTimeStack, alf );
	      }
            else
              MvclRunTimeError( "Memory Allocation Error\n" );
           }

         /* check for errors. */
         if( !got_it )
           MvclRunTimeError( "Stack Sync Error In Else.\n" );

         /* add the new node to the program list. */
         AddMvclProgramNode( prog, new );
        }
     }

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/****************************************************************************/
int   BuildMvclEndifNode(
   MvclProgramList  *prog )                 /* Pointer to program list      */

  {
   /* BuildMvclEndifNode() -- This function will generate the required data
    *     structures for an endif statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;
   MvclProgramNode   *address;
   int               got_it = 0;
   MvclRunTimeStackEntry  *entry;

   /* Allocate memory for the new program node. */
   new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

   if( new != (MvclProgramNode *)NULL )
     {
      /* basic housekeeping. */
      new->StatementType = MVCL_ENDIF;
      new->next = (MvclProgramNode *)NULL;

      /* pop what can be either an 'if' node address or an else node address.*/
      entry = MvclRunTimeStackPop( prog->RunTimeStack );

      if( entry != (MvclRunTimeStackEntry *)NULL )
        {
         if( entry->ValueType == MVCL_SYM_ADDRESS )
           {
            /* see if it's an if statement. */
            address = entry->Value.Address;    

            if( address->StatementType == MVCL_IF )
              {
               /* Good. Now put the address of this node in the if statement.
                * See the interpreter for more info.
                */
               (address->def.IfStatement)->FalseAddress = new;               
               got_it = 1;
              } 
            else if( address->StatementType == MVCL_ELSE )
              {
               (address->def.ElseStatement)->EndifAddress = new;

               /* Now - pop the if off of the stack, just to clean up. */
               entry = MvclRunTimeStackPop( prog->RunTimeStack );
               got_it = 1;
	      }
           }
        }

      /* check for errors. */
      if( !got_it )
        MvclRunTimeError( "Stack Sync Error In Endif.\n" );

      /* add the new node to the program list. */
      AddMvclProgramNode( prog, new );
     }

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/****************************************************************************/
int  BuildMvclForNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   char      *identifier,                  /* identifier to use as control */
   char      *start_ident,                 /* initial value of identifier  */
   int       for_type,                     /* To = count up Downto down    */
   char      *target_ident )               /* target count                 */

  {
   /* BuildMvclForNode() -- This function will generate the required data
    *     structures for a for statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;
   MvclProgramNode   *null_node;
   MvclRunTimeStackEntry  *entry;

   /* Allocate memory for the new program node. */
   new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

   if( new != (MvclProgramNode *)NULL )
     {
      /* now let's allocate memory for the statement. */
      new->def.ForStatement =
      (MvclForStruct *)malloc( sizeof( MvclForStruct ) );

      if( new->def.ForStatement != (MvclForStruct *)NULL )
        {
         /* We have allocated the memory now assign the specifics. */
         (new->def.ForStatement)->control_ident = identifier;
         (new->def.ForStatement)->type = for_type;
         (new->def.ForStatement)->target_ident = target_ident;
         (new->def.ForStatement)->start_ident = start_ident;

         /* basic housekeeping. */
         new->StatementType = MVCL_FOR;
         new->next = (MvclProgramNode *)NULL;

         /* Allocate s symbol table entry for this identifier. */
         if( !IsMvclSymbolInTable( prog->SymbolTable, identifier ) )
           {
            AddMvclSymbolToTable( prog->SymbolTable, identifier,
                                     MVCL_SYM_INT, 0 );
           }
         else
           MvclRunTimeError( "Symbol '%s' Has Been Previously Defined.\n",
                             identifier );

         /* Allocate memory for stack node. */
         entry =(MvclRunTimeStackEntry *)malloc(sizeof(MvclRunTimeStackEntry));

         if( entry != (MvclRunTimeStackEntry *)NULL )
           {
            /* Push this node pointer onto the stack */
            entry->ValueType = MVCL_SYM_ADDRESS;
            entry->Value.Address = new;
            MvclRunTimeStackPush( prog->RunTimeStack, entry );
	   }
         else
           MvclRunTimeError( "Memory Allocation Error\n" );

         /* add the new node to the program list. */
	 AddMvclProgramNode( prog, new );

         null_node = BuildMvclNullNode( prog );

         /* Add a null node ( a place holder ) afer the loop control. */
	 AddMvclProgramNode( prog, null_node );         
        }
      else
        {
         free( new );
         new = (MvclProgramNode *)NULL;
        }
     }

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/***************************************************************************/
int  BuildMvclEndForNode(
   MvclProgramList  *prog )                 /* Pointer to program list      */


  {
   /* BuildMvclEndForNode() -- This function will generate the required data
    *     structures for a endfor statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;
   MvclProgramNode   *address;
   MvclProgramNode   *null_node;
   MvclRunTimeStackEntry  *entry;

   /* Allocate memory for the new program node. */
   new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

   if( new != (MvclProgramNode *)NULL )
     {
      /* now let's allocate memory for the statement. */
      new->def.EndForStatement =
      (MvclEndForStruct *)malloc( sizeof( MvclEndForStruct ) );

      if( new->def.EndForStatement != (MvclEndForStruct *)NULL )
        {
         /* basic housekeeping. */
         new->StatementType = MVCL_ENDFOR;
         new->next = (MvclProgramNode *)NULL;

         /* pop what should be a For node off of the stack. */
         entry = MvclRunTimeStackPop( prog->RunTimeStack );

         if( entry != (MvclRunTimeStackEntry *)NULL )
           {
            if( entry->ValueType == MVCL_SYM_ADDRESS )
             {
              /* see if it's an if statement. */
              address = entry->Value.Address;    

              if( address->StatementType == MVCL_FOR )
                {
                 /* We have allocated the memory now assign the specifics. */
                 (new->def.EndForStatement)->ForAddress = address->next;

                 /* make note of the control identifier. */
                 (new->def.EndForStatement)->control_ident = 
                 (address->def.ForStatement)->control_ident;

                 /* make note of the target identifier. */
                 (new->def.EndForStatement)->target_ident = 
                 (address->def.ForStatement)->target_ident;

                 /* make note of the type of count. */
                 (new->def.EndForStatement)->type = 
                 (address->def.ForStatement)->type;

                 null_node = BuildMvclNullNode( prog );

                 /* assign this address to the for in case the intial test
                  * fails the loop control test.
                  */
                 (address->def.ForStatement)->EndForAddress = null_node;
	        }
	     }
	   }
         else
           MvclRunTimeError( "Memory Allocation Error\n" );

         /* add the new node to the program list. */
	 AddMvclProgramNode( prog, new );

         /* Add a null node ( a place holder ) after the loop control. */
	 AddMvclProgramNode( prog, null_node );
        }
      else
        {
         free( new );
         new = (MvclProgramNode *)NULL;
        }
     }

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/***************************************************************************/
int   BuildMvclPrintNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   MvclPrintItem    *list )                /* list of arguments            */

  {
   /* BuildMvclPrintNode() -- This function will generate the required data
    *     structures for a print statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Allocate memory for the new program node. */
   new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

   if( new != (MvclProgramNode *)NULL )
     {
      /* now let's allocate memory for the statement. */
      new->def.PrintStatement =
      (MvclPrintStruct *)malloc( sizeof( MvclPrintStruct ) );

      if( new->def.PrintStatement != (MvclPrintStruct *)NULL )
        {
         /* We have allocated the memory now assign the specifics. */
         (new->def.PrintStatement)->list = list;

         /* basic housekeeping. */
         new->StatementType = MVCL_PRINT;
         new->next = (MvclProgramNode *)NULL;

         /* add the new node to the program list. */
         AddMvclProgramNode( prog, new );
        }
      else
        {
         free( new );
         new = (MvclProgramNode *)NULL;
        }
     }
 
   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/****************************************************************************/
int  BuildMvclReturnNode(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   int    return_val )                     /* in seconds                  */

  {
   /* BuildMvclReturnNode() -- This function will generate the required data
    *     structures for a return statement.
    *
    *     RETURNS: 0 for a successfull build
    *              1 if not
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Allocate memory for the new program node. */
   new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

   if( new != (MvclProgramNode *)NULL )
     {
      /* now let's allocate memory for the statement. */
      new->def.ReturnStatement =
      (MvclReturnStruct *)malloc( sizeof( MvclReturnStruct ) );

      if( new->def.ReturnStatement != (MvclReturnStruct *)NULL )
        {
         /* We have allocated the memory now assign the specifics. */
         (new->def.ReturnStatement)->return_value = return_val;

         /* basic housekeeping. */
         new->StatementType = MVCL_RETURN;
         new->next = (MvclProgramNode *)NULL;

         /* Initialize the return value to zero. The prod data structure
          * will hold the final return value.
          */
         prog->return_value = 0;

         /* add the new node to the program list. */
         AddMvclProgramNode( prog, new );
        }
      else
        {
         free( new );
         new = (MvclProgramNode *)NULL;
        }
     }

   return( new == (MvclProgramNode *)NULL ? 1 : 0 );
  }
/**************************************************************************/
static MvclProgramNode *  BuildMvclNullNode(
   MvclProgramList  *prog )                 /* Pointer to program list      */

  {
   /* BuildMvclNullNode() -- This function will create a "null" program node.
    *     This node serves no other purpose than to be a place holder for
    *     control or loop statements.
    *
    *     RETURNS: a pointer to the node added
    */
   MvclProgramNode   *new = (MvclProgramNode *)NULL;

   /* Allocate memory for the new program node. */
   new = (MvclProgramNode *)malloc( sizeof( MvclProgramNode ) );

   if( new != (MvclProgramNode *)NULL )
     {
      /* basic housekeeping. */
      new->StatementType = MVCL_NULL;
      new->next = (MvclProgramNode *)NULL;
     }

   return( new );
  }
/****************************************************************************/
/* finis */






















