/*                         MICRION SOFTWARE DEPT.
 *          Copyright (c) 1995 by Micrion Corp. All rights reserved.
 *
 * FILE NAME    : mvcl_interpret.c
 * AUTHOR       : R. Mello
 * DATE WRITTEN : 3/18/95
 * VERSION      : 1.00
 * USAGE        : module of the mcvl parser
 *
 * This source file contains the mvcl interpreter.
 */
#include <stdio.h>                          /* for NULL, fprintf(), ect...   */
#include <signal.h>                         /* for SIGALRM                   */
#include "mvcl.h"                           /* for statement ID's            */
#include <mvcl_int.h>
 
/* uncomment the following line for debugging. */
/* #define MVCL_DEBUG */

/* local globals. */
static int    OperationTimedOut = 0;
static PFV    OldAlarmHandler;

/* function prototypes. */
int  MvclInterpreter( MvclProgramList *, int * );

/* local function prototypes. */
static int  InterpretOpenStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretCloseStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretStatusStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretSetStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretReadStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretTimeoutStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretPumpOnStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretPumpOffStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretExtendStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretRetractStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretWaitForStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretWaitStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretIfStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretElseStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretEndifStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretForStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretEndForStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretNullStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretPrintStatement( MvclProgramList *, MvclProgramNode * );
static int  InterpretReturnStatement( MvclProgramList *, MvclProgramNode * );

static void StartWatchdog( double );
static int  StopWatchdog( void );
static void WatchdogAlarm( void );

/* Interpreter function dispatch table. */
static  PFI  InterpreterTable[]=
  {
   InterpretOpenStatement,
   InterpretCloseStatement,
   InterpretStatusStatement,
   InterpretSetStatement,
   InterpretReadStatement,
   InterpretTimeoutStatement,
   InterpretPumpOnStatement,
   InterpretPumpOffStatement,
   InterpretExtendStatement,
   InterpretRetractStatement,
   InterpretWaitForStatement,
   InterpretWaitStatement,
   InterpretIfStatement,
   InterpretElseStatement,
   InterpretEndifStatement,
   InterpretForStatement,
   InterpretEndForStatement,
   InterpretNullStatement,
   InterpretPrintStatement,
   InterpretReturnStatement
  };

/***************************************************************************/
int  MvclInterpreter(
   MvclProgramList   *program,              /* ptr to the program node list  */
   int               *script_return )

  {
   /* MvclInterpreter() -- This function will interpret a translated mvcl
    *     source file. The interpreter will simply traverse the list of
    *     program nodes executing the mvcl program.
    *
    *     RETURNS: 0 for success
    *              1 for failure
    */
   int       result;
   MvclProgramNode *current;

   /* Make sure we have a valid program before we continue... */
   if( program != (MvclProgramList *)NULL )
     {
      #ifdef MVCL_DEBUG
      printf( "Beginning Interpretation...\n" );
      #endif

      /* Let's get started... */
      current = program->ip;

      /* Enter the interpreter loop. */
      while( current != (MvclProgramNode *)NULL )
        {
         #ifdef MVCL_DEBUG
         printf( "Executing A '%s' Node At Address %p \n",
                 MvclStatementTypeToName( current->StatementType ), current );
         #endif

         /* based on the statement type, dispatch the appropriate function. */
         if( current->StatementType != END )
           {
            result =
            (*InterpreterTable[current->StatementType])( program, current );

            /* We're done if we see a return statement. */
            if( current->StatementType == MVCL_RETURN )
              break;
           }
         else
           {
            result = 0;
            break;
           }

         /* test the return value, exit the loop if an error occured. */
         if( result != 0 )
           break;

         /* The next node is based on the instruction pointer. */
         current = program->ip;
        }
     }

   /* reset the program's instruction pointer to the start of the program. */
   program->ip = program->head;

   /* Set the return value. */
   *script_return = program->return_value;

   return( result );
  }
/***************************************************************************/
static int  InterpretOpenStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretOpenStatement() -- This function will interpret an mvcl open
    *     statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;

   /* If recording is turned on, write this command out the record file. */
   MvclRecordValve( MVCL_VALVE_OPEN, node->def.OpenStatement->valve );

   /* Attempt to open the valve. */
   printf( "Opening Valve %d... ", node->def.OpenStatement->valve );
   ret = MvclOpenValve( node->def.OpenStatement->valve );
   printf( "%s\n", ret ? "Failed." : "Ok." );
   
   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/***************************************************************************/
static int  InterpretCloseStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretCloseStatement() -- This function will interpret an mvcl close
    *     statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;

   /* If recording is turned on, write this command out the record file. */
   MvclRecordValve( MVCL_VALVE_CLOSE, node->def.CloseStatement->valve );

   printf( "Closing Valve %d... ", node->def.CloseStatement->valve );

   /* Attempt to close the valve. */
   ret = MvclCloseValve( node->def.CloseStatement->valve );

   printf( "%s\n", ret ? "Failed." : "Ok." );

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/***************************************************************************/
static int  InterpretStatusStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretStatusStatement() -- This function will interpret an mvcl status
    *     statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int             ret = 0;
   int             stat;
   char            *name;
   MvclSymbolType  sym;
   MvclSymbol      *addr;

   /* If recording is turned on, write this command out the record file. */
   MvclRecordStatus( node->def.StatusStatement->dev_class,
                     node->def.StatusStatement->dev_id );

   printf( "Checking Status Of %s... ",
           MvclResolveDeviceName( node->def.StatusStatement->dev_class,
                                  node->def.StatusStatement->dev_id ) );

   /* Check the device status. */
   stat = MvclStatus( node->def.StatusStatement->dev_class,
                     node->def.StatusStatement->dev_id );

   printf( "It's %s\n",
          MvclResolveDeviceStatus(node->def.StatusStatement->dev_class, stat));

   /* Get the symbol table reference that we saved during node building. */
   addr = (MvclSymbol *)node->def.StatusStatement->SymAddress;
   name = (char *)addr->SymName;

   /* Update this symbols entry in the symbol table. */
   sym.IntVal = stat;
   UpdateMvclSymbolInTable( prog->SymbolTable, name, MVCL_SYM_INT, &sym );

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/***************************************************************************/
static int  InterpretSetStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretSetStatement() -- This function will interpret an mvcl set
    *     statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;

   /* If recording is turned on, write this command out the record file. */
   MvclRecordSet( node->def.SetStatement->dev_class,
                  node->def.SetStatement->dev_id,
                  node->def.SetStatement->dev_value );

   printf( "Setting %s To %f... ",
           MvclResolveDeviceName( node->def.SetStatement->dev_class,
                                  node->def.SetStatement->dev_id ),
                                  node->def.SetStatement->dev_value );

   /* Set the device. */
   ret = MvclSetDevice( node->def.SetStatement->dev_class,
                        node->def.SetStatement->dev_id,
                        node->def.SetStatement->dev_value );

   printf( "%s\n", ret ? "Failed." : "Ok." );
   
   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/***************************************************************************/
static int  InterpretReadStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretReadStatement() -- This function will interpret an mvcl read
    *     statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int             ret = 0;
   double          value;
   char            *name;
   MvclSymbolType  sym;
   MvclSymbol      *addr;

   /* If recording is turned on, write this command out the record file. */
   MvclRecordRead( node->def.ReadStatement->dev_class,
                   node->def.ReadStatement->dev_id );

   printf( "Reading %s... ",
           MvclResolveDeviceName( node->def.ReadStatement->dev_class,
                                  node->def.ReadStatement->dev_id ) );

   /* Read into var. */
   value = MvclReadDevice( node->def.ReadStatement->dev_class,
                           node->def.ReadStatement->dev_id );

   printf( "%f\n", value );

   /* Get the symbol table reference that we saved during node building. */
   addr = (MvclSymbol *)node->def.ReadStatement->SymAddress;
   name = (char *)addr->SymName;

   /* Update this symbols entry in the symbol table. */
   sym.DoubleVal = value;
   UpdateMvclSymbolInTable( prog->SymbolTable, name, MVCL_SYM_DOUBLE, &sym );

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/***************************************************************************/
static int  InterpretTimeoutStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretTimeoutOpenStatement() -- This function will interpret an mvcl
    *     timeout statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;

   /* If recording is turned on, write this command out the record file. */
   MvclRecordTimeout( node->def.TimeoutStatement->timeout );

   printf( "Setting Timeout Of %f Seconds... Ok. \n",
            node->def.TimeoutStatement->timeout );

   /* keep a copy of the timeout value. */
   prog->timeout_val = node->def.TimeoutStatement->timeout;

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/***************************************************************************/
static int  InterpretPumpOnStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretPumpOnStatement() -- This function will interpret a mvcl pump
    *     on statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;

   /* If recording is turned on, write this command out the record file. */
   MvclRecordPump( MVCL_PUMP_START, node->def.PumpOnStatement->pump );

   printf( "Starting Pump %d... ", node->def.PumpOnStatement->pump );

   ret = MvclStartPump( node->def.PumpOnStatement->pump );

   printf( "%s\n", ret ? "Failed." : "Ok." );

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/****************************************************************************/
static int  InterpretPumpOffStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretPumpOffStatement() -- This function will interpret a mvcl pump
    *     off statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;

   /* If recording is turned on, write this command out the record file. */
   MvclRecordPump( MVCL_PUMP_STOP, node->def.PumpOffStatement->pump );

   printf( "Stopping Pump %d... ", node->def.PumpOffStatement->pump );

   ret = MvclStopPump( node->def.PumpOffStatement->pump );

   printf( "%s\n", ret ? "Failed." : "Ok." );

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/****************************************************************************/
static int  InterpretExtendStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretExtendStatement() -- This function will interpret a mvcl extend
    *     statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;

   /* If recording is turned on, write this command out the record file. */
   MvclRecordNozzle( MVCL_NOZZLE_EXTEND, node->def.ExtendStatement->nozzle );

   printf( "Extending Nozzle %d... ", node->def.ExtendStatement->nozzle );

   ret = MvclExtendNozzle( node->def.ExtendStatement->nozzle );

   printf( "%s\n", ret ? "Failed." : "Ok." );

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/****************************************************************************/
static int  InterpretRetractStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretRetractStatement() -- This function will interpret a mvcl 
    *     retract statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;

   /* If recording is turned on, write this command out the record file. */
   MvclRecordNozzle( MVCL_NOZZLE_RETRACT, node->def.RetractStatement->nozzle );

   printf( "Retracting Nozzle %d... ", node->def.RetractStatement->nozzle );

   ret = MvclRetractNozzle( node->def.RetractStatement->nozzle );

   printf( "%s\n", ret ? "Failed." : "Ok." );

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/****************************************************************************/
static int  InterpretWaitStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretWaitStatement() -- This function will interpret an mvcl wait
    *     statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;
   
   /* If recording is turned on, write this command out the record file. */
   MvclRecordWait( node->def.WaitStatement->wait_time );

   printf( "Waiting For %f Seconds... ", node->def.WaitStatement->wait_time );

   /* Sleep for the selected time. */
   sleep( (unsigned int)(node->def.WaitStatement->wait_time ) );

   printf( "Ok.\n" );

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/***************************************************************************/
static int  InterpretWaitForStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretWaitForStatement() -- This function will interpret an mvcl
    *     "waitfor" statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;
   int    compare;
   int    dev_class;
   int    dev_id;
   int    waiting;
   double target_value;
   double read_value;

   /* get short names for the stuff we are working with. */
   compare = node->def.WaitForStatement->compare;
   dev_class = node->def.WaitForStatement->dev_class;    
   dev_id = node->def.WaitForStatement->dev_id;    
   target_value = node->def.WaitForStatement->value;    

    /* If recording is turned on, write this command out the record file. */
   MvclRecordWaitFor( dev_class, dev_id, compare, target_value );

   /* Start up the watchdog timer. */
   StartWatchdog( prog->timeout_val );

   /* Set the loop control conditions. */
   waiting = 1;
   
   /* While waiting for a condition... */
   while( waiting && !OperationTimedOut )
     {
      printf( "Waiting For %s To Be %s %f... ",
               MvclResolveDeviceName( dev_class, dev_id ),
               MvclResolveComparison( compare ), target_value );

      /* read the device and get it's value. */
      read_value = MvclReadDevice( dev_class, dev_id );

      /* based on the comparison type... */
      switch( compare )
        {
         case MVCL_EQUAL:
           if( read_value == target_value )
             waiting = 0;  
           break;

         case MVCL_NOTEQUAL:
           if( read_value != target_value )
             waiting = 0;  
           break;

         case MVCL_LESSTHAN:
           if( read_value < target_value )
             waiting = 0;  
           break;

         case MVCL_GREATERTHAN:
           if( read_value > target_value )
             waiting = 0;  
           break;

         case MVCL_LESS_EQUAL:
           if( read_value <= target_value )
             waiting = 0;  
           break;

         case MVCL_GREATER_EQUAL:
           if( read_value >= target_value )
             waiting = 0;  
           break;
    
        }/* END SWITCH */

      printf( "It's At %f\n", read_value );
	     
     }/* END WHILE */

   /* Stop the timer. */
   ret = StopWatchdog();

   if( ret )
     MvclRunTimeError( "Timeout Expired While Waiting For %s\n",
                       MvclResolveDeviceName( dev_class, dev_id ) );

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/***************************************************************************/
static int  InterpretIfStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretIfStatement() -- This function will interpret an mvcl "if"
    *     statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;
   int    result = 0;
   int    compare;
   char   *left_ident;
   char   *right_ident;
   double left_value;
   double right_value;
   MvclSymbol   *sym;
   MvclRunTimeStackEntry *alf;

   /* If recording is turned on, write this command out the record file. */
   /* RM@ -- LATER MvclRecordIf( node->def.WaitStatement->wait_time ); */

   /* shorthand names for struct elements. */
   compare = node->def.IfStatement->compare;
   left_ident = node->def.IfStatement->ident_left;
   right_ident = node->def.IfStatement->ident_right;

   /* Get the left identifiers value from the symbol table. */
   sym = QueryMvclSymbolTable( prog->SymbolTable, left_ident );

   if( sym != (MvclSymbol *)NULL )
     {
      if( sym->SymType == MVCL_SYM_DOUBLE )
        left_value = sym->def.DoubleVal;
      else if( sym->SymType == MVCL_SYM_INT )
        left_value = (double)sym->def.IntVal;

      /* Get the left identifiers value from the symbol table. */
      sym = QueryMvclSymbolTable( prog->SymbolTable, right_ident );

      if( sym != (MvclSymbol *)NULL )
        {
         if( sym->SymType == MVCL_SYM_DOUBLE )
           right_value = sym->def.DoubleVal;
         else if( sym->SymType == MVCL_SYM_INT )
           right_value = (double)sym->def.IntVal;

         /* based on the comparison type... */
         switch( compare )
           {
            case MVCL_EQUAL:
              if( left_value == right_value )
                result = 1;  
              break;

            case MVCL_NOTEQUAL:
              if( left_value != right_value )
                result = 1;  
              break;
  
            case MVCL_LESSTHAN:
              if( left_value < right_value )
                result = 1;  
              break;

            case MVCL_GREATERTHAN:
              if( left_value > right_value )
                result = 1;  
              break;
  
            case MVCL_LESS_EQUAL:
              if( left_value <= right_value )
                result = 1;  
              break;

            case MVCL_GREATER_EQUAL:
              if( left_value >= right_value )
                result = 1;  
              break;
    
           }/* END SWITCH */
        }
     }

   /* set the instruction pointer to the next link. */
   if( result == 0 )
     prog->ip = node->def.IfStatement->FalseAddress;
   else
     prog->ip = node->next;

   /* push this if's result. */
   alf = (MvclRunTimeStackEntry *)malloc(sizeof(MvclRunTimeStackEntry));

   if( alf != (MvclRunTimeStackEntry *)NULL )
     {
      /* Push this node pointer onto the stack */
      alf->ValueType = MVCL_SYM_INT;
      alf->Value.IntVal = result;
      MvclRunTimeStackPush( prog->RunTimeStack, alf );
     }
   else
     MvclRunTimeError( "Memory Allocation Error\n" );
   
   return( ret );
  }
/***************************************************************************/
static int  InterpretElseStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretElseStatement() -- This function will interpret an mvcl "else"
    *     statement. This node in and of it's self does nothing but advance the
    *     instruction pointer to the next node. When constructing the program
    *     list the next node has not been parsed yet so the address of the
    *     "else" or "endif" is used until such a time ( now ) that access to
    *     the next program node is available.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;
   MvclRunTimeStackEntry  *entry;
   
   /* If recording is turned on, write this command out the record file. */
   /* RM@ -- LATER MvclRecordElse( node->def.WaitStatement->wait_time ); */

   /* pop if's result */
   entry = MvclRunTimeStackPop( prog->RunTimeStack );

   /* if it's true - push result back on - then jump to the endif node. */
   if( entry != (MvclRunTimeStackEntry *)NULL )
     {
      if( entry->ValueType == MVCL_SYM_INT )
        {
         if( entry->Value.IntVal == 0 )
           prog->ip = node->next;
         else
           prog->ip = node->def.ElseStatement->EndifAddress;
        }

      /* push this entry back onto the stack. */
      MvclRunTimeStackPush( prog->RunTimeStack, entry );
     }

   return( ret );
  }
/****************************************************************************/
static int  InterpretEndifStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretEndifStatement() -- This function will interpret an mvcl "endif"
    *     statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;
   MvclRunTimeStackEntry  *entry;
   
   /* If recording is turned on, write this command out the record file. */
   /* RM@ -- LATER MvclRecordEndif( node->def.WaitStatement->wait_time ); */

   /* pop if's result */
   entry = MvclRunTimeStackPop( prog->RunTimeStack );

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/****************************************************************************/
static int  InterpretForStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretForStatement() -- This function will interpret an mvcl "for"
    *     statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int     ret = 0;
   int     control_ident;
   int     target_ident;
   int     start_ident;
   char    *control_identifier;
   char    *target_identifier;
   char    *start_identifier;
   MvclSymbol   *sym;

   /* If recording is turned on, write this command out the record file. */
   /* RM@ -- LATER MvclRecordFor( node->def.For ); */

   /* Get shorter names. */
   control_identifier = node->def.ForStatement->control_ident;
   target_identifier = node->def.ForStatement->target_ident;
   start_identifier = node->def.ForStatement->start_ident;

   /* initialize the control variable with the start variable. */
   sym = QueryMvclSymbolTable( prog->SymbolTable, start_identifier );

   if( sym != (MvclSymbol *)NULL )
     {
      if( sym->SymType == MVCL_SYM_DOUBLE )
        start_ident = (int)sym->def.DoubleVal;
      else if( sym->SymType == MVCL_SYM_INT )
        start_ident = sym->def.IntVal;

      /* Get the current value of the control from the symbol table. */
      sym = QueryMvclSymbolTable( prog->SymbolTable, control_identifier );

      if( sym != (MvclSymbol *)NULL )
        {
         /* intialize the control identifier and change the symbol table. */
         control_ident = start_ident;

         if( sym->SymType == MVCL_SYM_DOUBLE )
           sym->def.DoubleVal = (double)control_ident;
         else if( sym->SymType == MVCL_SYM_INT )
           sym->def.IntVal = control_ident;

         /* Get the current value of the target from the symbol table. */
         sym = QueryMvclSymbolTable( prog->SymbolTable, target_identifier );
       
         if( sym != (MvclSymbol *)NULL )
           {
            if( sym->SymType == MVCL_SYM_DOUBLE )
              target_ident = (int)sym->def.DoubleVal;
            else if( sym->SymType == MVCL_SYM_INT )
              target_ident = sym->def.IntVal;

            /* Test the control against the target value. This is based
             * on direction.
             */
            if( node->def.ForStatement->type == MVCL_FOR_TO )
              {
               if( control_ident > target_ident )
                 prog->ip = node->def.ForStatement->EndForAddress;
               else
                 prog->ip = node->next;
    	      }
            else
              {
               if( control_ident < target_ident )
                 prog->ip = node->def.ForStatement->EndForAddress;
               else
                 prog->ip = node->next;
   	      }
           }
        }
     }
   
   return( ret );
  }
/****************************************************************************/
static int  InterpretEndForStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretEndForStatement() -- This function will interpret an mvcl
    *     "endfor" statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int     ret = 1;
   int     control_ident;
   char    *control_identifier;
   char    *target_identifier;
   int     target_ident;
   MvclSymbol   *sym;

   /* If recording is turned on, write this command out the record file. */
   /* RM@ -- LATER MvclRecordEndFor( node->def.For ); */

   /* Get shorter names. */
   control_identifier = node->def.EndForStatement->control_ident;
   target_identifier = node->def.EndForStatement->target_ident;

   /* Get the current value of the control from the symbol table. */
   sym = QueryMvclSymbolTable( prog->SymbolTable, control_identifier );

   if( sym != (MvclSymbol *)NULL )
     {
      /* Adjust count accordingly. */
      if( node->def.EndForStatement->type == MVCL_FOR_TO )
        { 
         if( sym->SymType == MVCL_SYM_DOUBLE )
           sym->def.DoubleVal = sym->def.DoubleVal + 1.0;
         else if( sym->SymType == MVCL_SYM_INT )
           sym->def.IntVal = sym->def.IntVal + 1;
        }
      else
        {
         if( sym->SymType == MVCL_SYM_DOUBLE )
           sym->def.DoubleVal = sym->def.DoubleVal - 1.0;
         else if( sym->SymType == MVCL_SYM_INT )
           sym->def.IntVal = sym->def.IntVal - 1;
        }

      /* reduce to one decision. */
      if( sym->SymType == MVCL_SYM_DOUBLE )
        control_ident =  (int)sym->def.DoubleVal;
      else if( sym->SymType == MVCL_SYM_INT )
        control_ident =  sym->def.IntVal;

      /* Get the current value of the control from the symbol table. */
      sym = QueryMvclSymbolTable( prog->SymbolTable, target_identifier );

      if( sym != (MvclSymbol *)NULL )
        {
         /* extract the target ident. */
         if( sym->SymType == MVCL_SYM_DOUBLE )
           target_ident =  (int)sym->def.DoubleVal;
         else if( sym->SymType == MVCL_SYM_INT )
           target_ident =  sym->def.IntVal;

         /* Check the count and branch accordingly. */
         if( node->def.EndForStatement->type == MVCL_FOR_TO )
           {
            if( control_ident < target_ident )
              prog->ip = node->def.EndForStatement->ForAddress;
            else
              prog->ip = node->next;
           }
         else
           {
            if( control_ident > target_ident )
              prog->ip = node->def.EndForStatement->ForAddress;
            else
              prog->ip = node->next;
           }

         /* If we got here no errors. */
         ret = 0;
        }
     }

   return( ret );
  }
/****************************************************************************/
static int  InterpretNullStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretNullStatement() -- This function will interpret an mvcl "null"
    *     statement. This statement is only a place holder that just advances 
    *     to the next node. Sounds useless huh? 
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/****************************************************************************/
static int  InterpretPrintStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretPrintStatement() -- This function will interpret an mvcl print
    *     statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;
   MvclSymbol   *sym;
   MvclPrintItem *current;

   /* alias the list. */
   current = node->def.PrintStatement->list;
   while( current != (MvclPrintItem *)NULL )
     {   
      /* Switch based on the type of print. */
      switch( current->type )
        {
         case MVCL_PRINT_STR:
           fprintf( stdout, "%s", current->string );
           break;

         case MVCL_PRINT_INT:    /* NO BREAK */
         case MVCL_PRINT_REAL:
         case MVCL_PRINT_IDENT:
            /* Allocate a symbol table entry for this identifier. */
            if( IsMvclSymbolInTable( prog->SymbolTable, current->string ) )
              {
               sym = QueryMvclSymbolTable( prog->SymbolTable, current->string);

               if( sym != (MvclSymbol *)NULL )
                 {
                  if( sym->SymType == MVCL_SYM_DOUBLE )
                    fprintf( stdout, "%f", sym->def.DoubleVal );
                  else if( sym->SymType == MVCL_SYM_INT )
                    fprintf( stdout, "%d", sym->def.IntVal );
                  else if( sym->SymType == MVCL_SYM_STRING )
                    fprintf( stdout, "%s", sym->def.StringVal );
                 }
	      }
            else
              MvclRunTimeError( "Symbol '%s' Not Found.\n", current->string );
            break;
         }

      /* Advance link. */
      current = current->next;
     }

   /* final newline. */
   fprintf( stdout, "\n" );

   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/***************************************************************************/
static int  InterpretReturnStatement(
   MvclProgramList  *prog,                 /* just for the support structs  */
   MvclProgramNode  *node )                /* pointer to a program node     */

  {
   /* InterpretReturnStatement() -- This function will interpret an mvcl return
    *     statement.
    *
    *     RETURNS: 0 for success.
    *              1 for failure.
    */
   int    ret = 0;

   /* Save off the return value. */
   prog->return_value = node->def.ReturnStatement->return_value;
   
   /* set the instruction pointer to the next link. */
   prog->ip = node->next;

   return( ret );
  }
/***************************************************************************/
static void  StartWatchdog(
   double    timeout )                      /* timeout interval in seconds   */

  {
   /* StartWatchdog() -- This function will start the watchdog timer for 
    *     hardware control functions. If after the desired time the function
    *     does not return the watchdog alarm function will be invoked to
    *     process the error.
    *
    *     RETURNS: nothing.
    */

   /* Clear the timeout error flag. */
   OperationTimedOut = 0;
   OldAlarmHandler = signal( SIGALRM, (PFV)WatchdogAlarm );

   /* Set the alarm time. */
   alarm( (unsigned int)timeout );

   return;
  }
/****************************************************************************/
static int  StopWatchdog( void )

  {
   /* StopWatchdog() -- This function will stop the watchdog timer.
    *
    *     RETURNS: the value of the timeout flag.
    */

   /* Use the default signal */
   signal( SIGALRM, OldAlarmHandler );
   alarm( 0 );

   return( OperationTimedOut );
  }
/****************************************************************************/
static void  WatchdogAlarm( void )

  {
   /* WatchdogAlarm() -- This is the function that gets called when a timeout
    *     occurs.
    *
    *     RETURNS: nothing.
    */

   /* Set the timed out flag. */
   OperationTimedOut = 1;

   return;
  }
/****************************************************************************/
/* finis */
