/*                         MICRION SOFTWARE DEPT.
 *          Copyright (c) 1995 by Micrion Corp. All rights reserved.
 *
 * FILE NAME    : mvcl_list.c
 * AUTHOR       : R. Mello
 * DATE WRITTEN : 3/18/95
 * VERSION      : 1.00
 * USAGE        : module of the mcvl parser
 *
 * This source file contains code to manage the mvcl program list.
 */
#include <stdio.h>                          /* for NULL, fprintf(), ect...   */
#include <malloc.h>                         /* for malloc() and free()       */
#include <mvcl_int.h>

/* function prototypes */
MvclProgramList *  CreateMvclProgramList( void );
void    AddMvclProgramNode( MvclProgramList *, MvclProgramNode * );
void    DestroyMvclProgramList( MvclProgramList * );
void    DeleteMvclProgramList( MvclProgramList * );

/* local function prototypes */
static void  DestroyMvclProgramNode( MvclProgramNode * );

/***************************************************************************/
MvclProgramList *  CreateMvclProgramList( void )

  {
   /* CreateMvclProgramList() -- This function will create an instance of a
    *     mvcl program list. This will include creating any subordinate 
    *     structures as well.
    *
    *     RETURNS: A pointer to a new mvcl program list.
    *              NULL if this function was upset in any way.
    */
   MvclProgramList   *new;

   /* Allocate the memory for the list. */
   new = (MvclProgramList *)malloc( sizeof( MvclProgramList ) );

   if( new != (MvclProgramList *)NULL )
     {
      new->head = (MvclProgramNode *)NULL;
      new->tail = (MvclProgramNode *)NULL;
      new->ip = (MvclProgramNode *)NULL;

      /* create instances of the symbol table and the run time stack */
      new->SymbolTable = CreateMvclSymbolTable();

      if( new->SymbolTable != (MvclSymbolTable *)NULL )
        {
         new->RunTimeStack = CreateMvclRunTimeStack();

         if( new->RunTimeStack == (MvclRunTimeStack *)NULL )
	   {
            free( new->SymbolTable );
	    free( new );
	    new = (MvclProgramList *)NULL;
           }
        }
      else
	{
         free( new );
         new = (MvclProgramList *)NULL;
	}
     }

   return( new );
  }
/***************************************************************************/
void  AddMvclProgramNode(
   MvclProgramList    *MvclProgram,        /* pointer to the program list  */
   MvclProgramNode    *Node )              /* node to append to the list   */

  {
   /* AddMvclProgramNode() -- This function will add ( append actually ) a 
    *     new program node to the ever increasing list of program nodes.
    *
    *     RETURNS: nothing.
    */

   /* Only add valid nodes... */
   if( Node != (MvclProgramNode *)NULL )
     {
      /* Make sure the new node's next pointer is null */
      Node->next = (MvclProgramNode *)NULL;

      /* If it's the first node... */
      if( MvclProgram->head == (MvclProgramNode *)NULL )
        {
         MvclProgram->head = Node;
         MvclProgram->tail = Node;
         MvclProgram->ip = Node;
        }
      else
        {
         MvclProgram->tail->next = Node;
         MvclProgram->tail = MvclProgram->tail->next;
        }
     }

   return;
  }
/***************************************************************************/
void  DestroyMvclProgramList(
   MvclProgramList  *MvclProgram )         /* pointer to a program list     */

  {
   /* DestroyMvclProgramList() -- This function will free up all of the memory
    *     associated with the mvcl program passed in.
    *
    *     RETURNS: nothing.
    */

   /* first just delete all of the list memory. Leaving the main structure
    * in place.
    */
   DeleteMvclProgramList( MvclProgram );

   /* After all of the nodes have been destroyed free up the symbol
    * table and the stack.
    */
   DestroyMvclSymbolTable( MvclProgram->SymbolTable );
   DestroyMvclRunTimeStack( MvclProgram->RunTimeStack );

   /* then finally the main data structure. */
   free( MvclProgram );

   return;
  }
/***************************************************************************/
void  DeleteMvclProgramList(
   MvclProgramList  *MvclProgram )         /* pointer to a program list     */

  {
   /* DeleteMvclProgramList() -- This function will free up the memory
    *     contained within the program list. It will not mess with the
    *     main list data structure only the list it maintains.
    *
    *     RETURNS: nothing.
    */
   MvclProgramNode   *current;
   MvclProgramNode   *previous;

   /* init loop control stuff. */
   current = MvclProgram->head;
   previous = (MvclProgramNode *)NULL;

   /* Until all of the list nodes are deleted. */
   while( current != (MvclProgramNode *)NULL )
     {
      /* Adjust links. */
      previous = current;
      current = current->next;

      /* Destroy the node pointed to by previous. */
      DestroyMvclProgramNode( previous );
     }

   /* Set the list pointers to NULL */
   MvclProgram->head = (MvclProgramNode *)NULL;
   MvclProgram->tail = (MvclProgramNode *)NULL;

   return;
  }
/***************************************************************************/
static void  DestroyMvclProgramNode(
   MvclProgramNode   *Node )               /* node to delete                */

  {
   /* DestroyMvclProgramNode() -- This function will free up the memory for
    *     a program node. This is broken out in case more complex data types
    *     are required as the language grows.
    *
    *     RETURNS: nothing.
    */

   if( Node != (MvclProgramNode *)NULL )
     {
      free( &(Node->def) );
      free( Node );
     }

   return;
  }
/***************************************************************************/
/* finis */
