/*                         MICRION SOFTWARE DEPT.
 *          Copyright (c) 1995 by Micrion Corp. All rights reserved.
 *
 * FILE NAME    : mvcl_rec.c
 * AUTHOR       : R. Mello
 * DATE WRITTEN : 3/15/95
 * VERSION      : 1.00
 * USAGE        : module of the mcvl parser
 *
 * This source file contains the functions that handle the mvcl recording.
 */
#include <stdio.h>                          /* for NULL, fprintf(), ect...   */
#include <mvcl_int.h>

/* equates... */
#define  NO    0
#define  YES   1

/* locals... */
static FILE   *RecordFile = (FILE *)NULL;
static int    RecordCommands = YES;

/* prototypes... */
void  MvclOpenRecordFile( char *, char * );
void  MvclCloseRecordFile( void );
void  MvclSuspendRecording( void );
void  MvclResumeRecording( void );
void  MvclRecordProc( char * );
void  MvclRecordValve( int, int );
void  MvclRecordPump( int, int );
void  MvclRecordStatus( int, int );
void  MvclRecordNozzle( int, int );
void  MvclRecordRead( int, int );
void  MvclRecordSet( int, int, double );
void  MvclRecordEnd( void );
void  MvclRecordComment( char * );
void  MvclRecordTimeout( double );
void  MvclRecordWait( double );
void  MvclRecordWaitFor( int, int, int, double );

/***************************************************************************/
void  MvclOpenRecordFile(
   char    *fname,                        /* full path of the file to open  */
   char    *proc_name )                   /* name of this procedure         */

  {
   /* MvclOpenRecordFile() -- This function will open the requested file for
    *     writting mvcl commands into. This function will not do any checking
    *     other than verifing the file handle. It's the responsibility of the
    *     user interface to check for any overwrites.
    *
    *     RETURNS: nothing.
    */

   /* Make an attempt to open the file */
   RecordFile = fopen( fname, "w+" );

   if( RecordFile != (FILE *)NULL )
     {
      MvclRecordComment( "# This file was automatically generated by mvcl." );
      MvclRecordProc( proc_name );
     }

   return;
  }
/***************************************************************************/
void  MvclCloseRecordFile( void )

  {
   /* MvclCloseRecordFile() -- This function will terminate the recording of
    *     of mvcl commands into the current file. 
    *
    *     RETURNS: nothing
    */

   if( RecordFile != (FILE *)NULL )
     {
      MvclRecordEnd();

      /* Now close the file. */
      fclose( RecordFile );

      /* Then reset the control vars. */
      RecordFile = (FILE *)NULL;
      RecordCommands = YES;
     }

   return;
  }
/***************************************************************************/
void  MvclSuspendRecording( void )

  {
   /* MvclSuspendRecording() -- This function will temporarally stop the
    *     recording of mvcl statements. Only a corresponding Resume or
    *     file close will swap the state.
    *
    *     RETURNS: nothing.
    */

   RecordCommands = NO;
   return;
  }
/***************************************************************************/
void  MvclResumeRecording( void )

  {
   /* MvclResumeRecording() -- This function will, once again, allow the
    *     recording of mvcl statements.
    *
    *     RETURNS: nothing.
    */

   RecordCommands = YES;
   return;
  }
/***************************************************************************/
void  MvclRecordProc(
   char    *ProcName )                     /* name of procedure            */

  {
   /* MvclRecordProc() -- This function will record a mvcl procedure name.
    *
    *    RETURNS: nothing.
    */

   /* only record into a valid file. */
   if( RecordFile != (FILE *)NULL )
     {
      /* If recording commands. */
      if( RecordCommands == YES )
        fprintf( RecordFile, "PROC \"%s\";\n", ProcName );
     }

   return;
  }
/***************************************************************************/
void  MvclRecordValve(
   int     state,                          /* OPEN | CLOSE                 */
   int     valve )                         /* number of the valve to open  */

  {
   /* MvclRecordValve() -- This function will record a request to either open
    *     or close a valve.
    *
    *     RETURNS: nothing
    */

   /* only record into a valid file. */
   if( RecordFile != (FILE *)NULL )
     {
      /* If recording commands. */
      if( RecordCommands == YES )
        {
         if( state == MVCL_VALVE_OPEN )
           fprintf( RecordFile, "open v%d;\n", valve );
         else
           fprintf( RecordFile, "close v%d;\n", valve );
        }
     }

   return;
  }
/***************************************************************************/
void  MvclRecordPump(
   int     state,                          /* ON | OFF                     */
   int     pump )                          /* number of the pump           */

  {
   /* MvclRecordPump() -- This function will record a request to either start
    *     or stop a pump.
    *
    *     RETURNS: nothing
    */

   /* only record into a valid file. */
   if( RecordFile != (FILE *)NULL )
     {
      /* If recording commands. */
      if( RecordCommands == YES )
        {
         if( state == MVCL_PUMP_START )
           fprintf( RecordFile, "pumpon p%d;\n", pump );
         else
           fprintf( RecordFile, "pumpoff p%d;\n", pump );
        }
     }

   return;
  }
/***************************************************************************/
void  MvclRecordStatus(
   int    dev_class,                        /* class of device to get status */
   int    dev_id )                          /* specific device               */

  {
   /* MvclRecordStatus() -- This function will record the request for a device
    *     status.
    *
    *     RETURNS: nothing.
    */

   /* only record into a valid file. */
   if( RecordFile != (FILE *)NULL )
     {
      /* If recording commands. */
      if( RecordCommands == YES )
        {
         fprintf( RecordFile, "%s = status %s;\n",
		              MvclGenerateVariable(), 
                              MvclResolveDeviceName( dev_class, dev_id ) );
        }
     }

   return;
  }
/****************************************************************************/
void  MvclRecordNozzle(
   int     state,                          /* EXTENDED | RETRACTED         */
   int     nozzle )                        /* number of the nozzle         */

  {
   /* MvclRecordNozzle() -- This function will record a request to either
    *     extend of retract a nozzle ( funnel ).
    *
    *     RETURNS: nothing
    */

   /* only record into a valid file. */
   if( RecordFile != (FILE *)NULL )
     {
      /* If recording commands. */
      if( RecordCommands == YES )
        {
         if( state == MVCL_NOZZLE_EXTEND )
           fprintf( RecordFile, "extend nz%d;\n", nozzle );
         else
           fprintf( RecordFile, "retract nz%d;\n", nozzle );
        }
     }

   return;
  }
/***************************************************************************/
void  MvclRecordRead(
   int     class,                           /* Class of device to read      */
   int     device )                         /* DEVICE                       */

  {
   /* MvclRecordRead() -- This function will record a request to read a device.
    *
    *     RETURNS: nothing
    */

   /* only record into a valid file. */
   if( RecordFile != (FILE *)NULL )
     {
      /* If recording commands. */
      if( RecordCommands == YES )
        fprintf( RecordFile, "%s = read %s;\n",
                 MvclGenerateVariable(), 
                 MvclResolveDeviceName( class, device ));
     }

   return;
  }
/***************************************************************************/
void  MvclRecordSet(
   int     class,                           /* Class of device to read      */
   int     device,                          /* DEVICE                       */
   double  value )                          /* Value to write to device.    */

  {
   /* MvclRecordSet() -- This function will record a request to read a device.
    *
    *     RETURNS: nothing
    */

   /* only record into a valid file. */
   if( RecordFile != (FILE *)NULL )
     {
      /* If recording commands. */
      if( RecordCommands == YES )
        {
         fprintf( RecordFile, "set %s %f;\n",
                 MvclResolveDeviceName(class, device), value );
        }
     }

   return;
  }
/***************************************************************************/
void  MvclRecordEnd( void )

  {
   /* MvclRecordEnd() -- This function will record a procedure end.
    *
    *     RETURNS: nothing
    */

   /* only record into a valid file. */
   if( RecordFile != (FILE *)NULL )
     {
      /* If recording commands. */
      if( RecordCommands == YES )
        fprintf( RecordFile, "END;\n" );
     }

   return;
  }
/***************************************************************************/
void  MvclRecordComment(
   char     *comment )

  {
   /* MvclRecordComment() -- This function will record a mvcl comment.
    *
    *     RETURNS: nothing
    */

   /* only record into a valid file. */
   if( RecordFile != (FILE *)NULL )
     {
      /* If recording commands. */
      if( RecordCommands == YES )
        fprintf( RecordFile, "#%s\n", comment );
     }

   return;
  }
/***************************************************************************/
void  MvclRecordTimeout(
   double  timeout )                        /* timeout value in seconds     */

  {
   /* MvclRecordTimeout() -- This function will record a mvc timeout.
    *
    *     RETURNS: nothing.
    */

   /* only record into a valid file. */
   if( RecordFile != (FILE *)NULL )
     {
      /* If recording commands. */
      if( RecordCommands == YES )
        fprintf( RecordFile, "timeout %f;\n", timeout );
     }

   return;
  }
/****************************************************************************/
void  MvclRecordWait(
   double  seconds )                        /* timeout value in seconds     */

  {
   /* MvclRecordWait() -- This function will record a mvcl wait.
    *
    *     RETURNS: nothing.
    */

   /* only record into a valid file. */
   if( RecordFile != (FILE *)NULL )
     {
      /* If recording commands. */
      if( RecordCommands == YES )
        fprintf( RecordFile, "wait %f;\n", seconds );
     }

   return;
  }
/****************************************************************************/
void  MvclRecordWaitFor(
   int    dev_class,                        /* device class to record        */
   int    dev_id,                           /* specific device               */
   int    comparison,                       /* enumerated comparison         */
   double dev_value )                       /* target value                  */

  {
   /* MvclRecordWaitFor() -- This function will record a mvcl waitfor.
    *
    *     RETURNS: nothing.
    */

   /* only record into a valid file. */
   if( RecordFile != (FILE *)NULL )
     {
      /* If recording commands. */
      if( RecordCommands == YES )
        {
         fprintf( RecordFile, "waitfor %s %s %f;\n",
                 MvclResolveDeviceName( dev_class, dev_id ),
                 MvclResolveComparison( comparison ),
                 dev_value );
        }
     }

   return;
  }
/****************************************************************************/
/* finis */







