/*                         MICRION SOFTWARE DEPT.
 *          Copyright (c) 1994 by Micrion Corp. All rights reserved.
 *
 * FILE NAME    : mvcl_shl.c
 * AUTHOR       : R. Mello
 * DATE WRITTEN : 4/23/95
 * VERSION      : 1.00
 * USAGE        : module of the mvcl programming language
 *
 *  This file contains code to provide an innteractive shell for executing
 * mvcl commands.
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <mvcl_int.h>

/* function prototypes. */
int   MvclRun( char *, int * );
int   MvclShell( void );
void  MvclShellPrompt( void );
void  ExecMvclCommand( MvclProgramList * );
int   ExecMvclCommandString( char * );
MvclProgramList *  MvclCompile( char * );

/***************************************************************************/
int  MvclRun(
   char    *fname,                         /* full path name of the mvcl file*/
   int     *script_return )

  {
   /* MvclRun() -- This function will first translate then execute the mvcl
    *     program contained in fname.
    *
    *     RETURNS:  0 if all went well.
    *              -1 if there was a parser error.
    *              -2 if there was a run time error.
    *              -3 if the file name could not be opened.
    */
   int   Rval = -3;
   int   ret;
   FILE  *fp;
   MvclProgramList  *prog = (MvclProgramList *)NULL;
 
   /* Create an instance of an mvcl program. */
   prog = CreateMvclProgramList();

   if( prog != (MvclProgramList *)NULL )
     {
      /* Communicate the new program list pointer to the parser code. */
      if( MvclSetParserProgramList( prog ) )
        {
         /* Setup a mode for our parser so it will know what variation of the
          * grammar we are interested in.
          */
         MvclSetParserMode( MVCL_STREAM );

         /* Set the input source to stream. */
         MvclSetInputSource( MVCL_INPUT_STREAM );

         /* Make an attempt to open the file name for read only. */
         fp = fopen( fname, "r" );

         if( fp != (FILE *)NULL )
           {
            /* Assign the lexers file handle to the one we want. */
            MvclSetSourceStream( fp );

            /* Now attempt to parse the file into mvcl program nodes. */
            if( !Mvclparse() )
              {
               /* The file has parsed Ok. At this point we have a linked list
                * of Mvcl program nodes just waiting to do our bidding.
                */
               if( !MvclInterpreter( MvclGetParserProgramList(), &ret ) )
                 Rval = 0;
               else
                 Rval = -2;
              }

            /* close out program file. */
            fclose( fp );
           }
        }

      /* Return the return_value to the caller. */
      *script_return = ret;

      /* Clear up the program list. */
      DestroyMvclProgramList( prog );
     }

   return( Rval );
  }
/***************************************************************************/
int  MvclShell( void )

  {
   /* MvclShell() -- This function will run a sub-set of mvcl commands from
    *     an interactive shell.
    *
    *     RETURNS: 0 if all went well.
    *              1 if things ran amouk.
    */
   int   Rval = 1;
   MvclProgramList  *prog = (MvclProgramList *)NULL;
 
   /* Create an instance of an mvcl program. */
   prog = CreateMvclProgramList();

   if( prog != (MvclProgramList *)NULL )
     {
      /* Communicate the new program list pointer to the parser code. */
      if( MvclSetParserProgramList( prog ) )
        {
         /* Setup a que for our parser so it will know what variation of the
          * grammar we are interested in.
          */
         MvclSetParserMode( MVCL_SHELL );

         /* Set the input source to stream. */
         MvclSetInputSource( MVCL_INPUT_STREAM );
 
         /* Assign the lexers file handle to the one we want. */
         MvclSetSourceStream( stdin );

         /* print out the first prompt. Subsequent new lines will match and
          * produce any remaining prompts.
          */
         MvclShellPrompt();

         /* Now attempt to parse the user input. */
         Mvclparse();

         /* Things went well. */
         Rval = 0;
        }

      /* Clear up the program list. */
      DestroyMvclProgramList( prog );
     }

   return( Rval );
  }
/***************************************************************************/
void  MvclShellPrompt( void )

  {
   /* MvclShellPrompt() -- This function will put up the interactive shells
    *     prompt.
    *
    *     RETURNS: nothing.
    */

   printf( "[ MvclShell > " );
   return;
  }
/***************************************************************************/
void  ExecMvclCommand(
   MvclProgramList   *command )

  {
   /* ExecMvclCommand() -- This function will execute the statements 
    *     in the command specified.
    *
    *     RETURNS: nothing... Void's do that!
    */
   int   ret;

   /* Execute the command. */
   MvclInterpreter( command, &ret );

   /* Clear the program memory of any notion of command. */
   DeleteMvclProgramList( command );

   return;
  }
/***************************************************************************/
int  ExecMvclCommandString( 
   char    *command_string ) 
 
  {
   /* ExecMvclCommandString() -- This function will execute an Mvcl command
    *     from an input string.
    *
    *     RETURNS: nothing.
    */
   int  ret = 0;

   /* Set the parser mode to shell */
   MvclSetParserMode( MVCL_STRING );

   /* Set the input source to string */
   MvclSetInputSource( MVCL_INPUT_STRING );

   /* Transfer the string over to the lexer. */
   MvclSetInputString( command_string );

   /* Call the parser. */
   ret = Mvclparse();

   return( ret );
  }
/***************************************************************************/
MvclProgramList *  MvclCompile(
   char    *fname )                        /* full path name of the mvcl file*/

  {
   /* MvclCompile() -- This function will compile the mvcl script into a set of
    *     program nodes.
    *
    *     RETURNS: the program node list pointer.
    *              or NULL for an error. 
    */
   MvclProgramList  *prog = (MvclProgramList *)NULL;
   FILE  *fp;

   printf( "Compiling Mvcl Progam <%s>\n", fname );
 
   /* Create an instance of an mvcl program. */
   prog = CreateMvclProgramList();

   if( prog != (MvclProgramList *)NULL )
     {
      /* Communicate the new program list pointer to the parser code. */
      if( MvclSetParserProgramList( prog ) )
        {
         /* Setup a mode for our parser so it will know what variation of the
          * grammar we are interested in.
          */
         MvclSetParserMode( MVCL_STREAM );

         /* Set the input source to stream. */
         MvclSetInputSource( MVCL_INPUT_STREAM );

         /* Make an attempt to open the file name for read only. */
         fp = fopen( fname, "r" );

         if( fp != (FILE *)NULL )
           {
            /* Assign the lexers file handle to the one we want. */
            MvclSetSourceStream( fp );

            /* Now attempt to parse the file into mvcl program nodes. */
            if( Mvclparse() )
              {
               /* The parse faild. Delete any existing program nodes. */
               DestroyMvclProgramList( prog );
               prog = (MvclProgramList *)NULL;
              }

            /* close out program file. */
            fclose( fp );
           }
        }
     }

   return( prog );
  }
/****************************************************************************/
/* finis */









