/*                         MICRION SOFTWARE DEPT.
 *          Copyright (c) 1995 by Micrion Corp. All rights reserved.
 *
 * FILE NAME    : mvcl_stk.c
 * AUTHOR       : R. Mello
 * DATE WRITTEN : 4/2/95
 * VERSION      : 1.00
 * USAGE        : module of the mcvl parser
 *
 * This source file contains the code to manipulate the run time stack.
 */
#include <stdio.h>                          /* for NULL, fprintf(), ect...   */
#include <malloc.h>                         /* for malloc() and free()       */
#include <mvcl_int.h>

/* function prototypes. */
MvclRunTimeStack *   CreateMvclRunTimeStack( void );
int        MvclRunTimeStackPush( MvclRunTimeStack *, MvclRunTimeStackEntry * );
MvclRunTimeStackEntry *  MvclRunTimeStackPop( MvclRunTimeStack * );
int                  IsMvclRunTimeStackEmpty( MvclRunTimeStack * );
int                  IsMvclRunTimeStackFull( MvclRunTimeStack * );
void                 DestroyMvclRunTimeStack( MvclRunTimeStack * );
MvclRunTimeStackEntry *  CreateMvclRunTimeStackEntry( void );
void                 DestroyMvclRunTimeStackEntry( MvclRunTimeStackEntry * );

/***************************************************************************/
MvclRunTimeStack *  CreateMvclRunTimeStack( void )

  {
   /* CreateMvclRunTimeStack() -- This function will create an instance of a
    *     mvcl run time stack.
    *
    *     RETURNS: a pointer to the  stack.
    *              NULL for a memory allocation error.
    */
   MvclRunTimeStack  *new;

   /* Allocate memory for the new stack. */
   new = (MvclRunTimeStack *)malloc( sizeof( MvclRunTimeStack ) );

   if( new != (MvclRunTimeStack *) NULL )
     {
      new->StackSize = 0;
      new->top = (MvclRunTimeStackEntry *)NULL;
     }

   return( new );
  }
/***************************************************************************/
int  MvclRunTimeStackPush(
   MvclRunTimeStack      *stack,               /* stack to work on.        */
   MvclRunTimeStackEntry *stack_entry )        /* info to push onto stack. */

  {
   /* MvclRunTimeStackPush() -- This function will push the stack entry onto
    *     the stack passed.
    *
    *     RETURNS:  0 if all went well
    *              -1 if the max stack limit has been reached.
    */
   int   Rval = -1;

   if( stack->StackSize < MVCL_MAX_LOOP_NEST )
     {
      if( stack->top == (MvclRunTimeStackEntry *)NULL )
        {
         stack_entry->next = (MvclRunTimeStackEntry *)NULL;
         stack->top = stack_entry;
        }
      else
        {
         stack_entry->next = stack->top;
         stack->top = stack_entry;
        }

      /* Bump the count of elements in the stack. */
      stack->StackSize++;

      /* Success. */
      Rval = 0;
     }
   else
     MvclRunTimeError( "Stack overflow." );

   return( Rval );
  }
/***************************************************************************/
MvclRunTimeStackEntry *  MvclRunTimeStackPop(
   MvclRunTimeStack      *stack )              /* stack to work on.        */

  {
   /* MvclRunTimeStackPop() -- This function will pop a stack entry off of
    *     the stack passed.
    *
    *     RETURNS:  A pointer to the data structure that was on the top.
    *              NULL if there was nothing to pop off.
    */
   MvclRunTimeStackEntry  *RetEntry = (MvclRunTimeStackEntry *)NULL;

   /* Make sure that the stack has something to pop. */
   if( stack->StackSize > 0 )
     {
      RetEntry = stack->top;
      stack->top = stack->top->next;

      /* Decrement the count of elements in the stack */
      stack->StackSize--;
     }
   else
     MvclRunTimeError( "Stack underflow." );

   return( RetEntry );
  }
/***************************************************************************/
int  IsMvclRunTimeStackEmpty(
   MvclRunTimeStack      *stack )              /* stack to work on.        */

  {
   /* IsMvclRunTimeStackEmpty() -- This function will determine if the stack
    *     is empty.
    *
    *     RETURNS: 0 if the stack is not empty.
    *              1 if it is.
    */
   int    IsEmpty;

   /* Operate only on valid stacks. */
   if( stack != (MvclRunTimeStack *)NULL )
     IsEmpty = ( stack->StackSize == 0 ) ? 1 : 0;

   return( IsEmpty );
  }
/***************************************************************************/
int  IsMvclRunTimeStackFull(
   MvclRunTimeStack      *stack )              /* stack to work on.       */

  {
   /* IsMvclRunTimeStackFull() -- This function will determine if the stack
    *     is full.
    *
    *     RETURNS: 0 if the stack is not empty.
    *              1 if it is.
    */
   int    IsFull;

   /* Operate only on valid stacks. */
   if( stack != (MvclRunTimeStack *)NULL )
     IsFull = ( stack->StackSize == MVCL_MAX_LOOP_NEST ) ? 1 : 0;

   return( IsFull );
  }
/***************************************************************************/
void  DestroyMvclRunTimeStack(
   MvclRunTimeStack      *stack )              /* stack to work on.       */

  {
   /* DestroyMvclRunTimeStack() -- This function will destroy a mvcl stack.
    *
    *     RETURNS: nothing.
    */
   MvclRunTimeStackEntry   *current;
   MvclRunTimeStackEntry   *previous;

   /* for all of the nodes remaining in the stack... */
   current = stack->top;
   while( current != (MvclRunTimeStackEntry *)NULL )
     {
      previous = current;
      current = current->next;

      free( previous );
     }

   /* now delete the stack node. */
   free( stack );

   return;
  }
/***************************************************************************/
MvclRunTimeStackEntry *  CreateMvclRunTimeStackEntry( void )

  {
   /* CreateMvclRunTimeStackEntry() -- This function will create an instance
    *     of a stack entry.
    *
    *     RETURNS: A pointer to the new structure.
    *              NULL for a memory allocation error.
    */
   MvclRunTimeStackEntry  *new;

   /* Allocate memory for the new stack. */
   new = (MvclRunTimeStackEntry *)malloc( sizeof( MvclRunTimeStackEntry ) );

   if( new != (MvclRunTimeStackEntry *)NULL )
     {
      new->next = (MvclRunTimeStackEntry *)NULL;
     }

   return( new );
  }
/***************************************************************************/
void  DestroyMvclRunTimeStackEntry(
   MvclRunTimeStackEntry *stack_entry )

  {
   /* DestroyMvclRunTimeStackEntry() -- This function will destroy an instance
    *     of a stack entry.
    *
    *     RETURNS: nothing.
    */

   if( stack_entry != (MvclRunTimeStackEntry *)NULL )
     free( stack_entry );

   return;
  }
/***************************************************************************/
/* finis */
