/*                         MICRION SOFTWARE DEPT.
 *          Copyright (c) 1995 by Micrion Corp. All rights reserved.
 *
 * FILE NAME    : mvcl_sym.c
 * AUTHOR       : R. Mello
 * DATE WRITTEN : 4/1/95
 * VERSION      : 1.00
 * USAGE        : module of the mcvl parser
 *
 * This source file contains the code for the symbol table. Mvcl scripts
 * will not have a lot of symbols so a linear linked list will suffice. If
 * it turns out that we need to improve the search method, fine.
 */
#include <stdio.h>                          /* for NULL, fprintf(), ect...   */
#include <stdlib.h>                         /* for malloc() and free()       */
#include <string.h>                         /* for strcmp() and strcpy()     */
#include <assert.h>                         /* for assertions                */
#include <mvcl_int.h>

/* function prototypes. */
MvclSymbolTable * CreateMvclSymbolTable( void );
int    IsMvclSymbolInTable( MvclSymbolTable *, char * );
void * AddMvclSymbolToTable( MvclSymbolTable *, char *, int, MvclSymbolType *);
int    UpdateMvclSymbolInTable(MvclSymbolTable *, char *,int,MvclSymbolType *);
MvclSymbol * QueryMvclSymbolTable( MvclSymbolTable *, char * );
void   DestroyMvclSymbolTable( MvclSymbolTable * );

/***************************************************************************/
MvclSymbolTable *  CreateMvclSymbolTable( void )

  {
   /* CreateMvclSymbolTable() -- This function will create an instance of
    *     a mvcl symbol table.
    *
    *     RETURNS: a pointer to the symbol table.
    *              or NULL for a memory allocation error.
    */
   MvclSymbolTable  *new;

   /* Allocate memory for the new table. */
   new = (MvclSymbolTable *)malloc( sizeof( MvclSymbolTable ) );

   if( new != (MvclSymbolTable *) NULL )
     {
      new->SymCount = 0;
      new->head = (MvclSymbol *)NULL;
      new->tail = (MvclSymbol *)NULL;
     }

   return( new );
  }
/***************************************************************************/
int  IsMvclSymbolInTable(
   MvclSymbolTable  *SymbolTable,          /* symbol table to act on      */
   char             *name )                /* Name to search table for.   */

  {
   /* IsMvclSymbolInTable() -- This function will determine if the passed
    *     symbol has already been entered into the table.
    *
    *     RETURNS: 0 if it is not.
    *              1 if it is.
    */
   MvclSymbol  *current;
   int         found;

   /* assign loop control stuff. */
   current = SymbolTable->head;
   found = 0;

   /* Search the list. */
   while( current != (MvclSymbol *)NULL && !found )
     {
      if( !strcmp( name, current->SymName ) )
        found = 1;

      current = current->next;
     }

   return( found );
  }
/***************************************************************************/
void *  AddMvclSymbolToTable(
   MvclSymbolTable  *SymbolTable,          /* symbol table to act on       */
   char             *name,                 /* pointer to symbol name       */
   int              type,                  /* type is int double or string */
   MvclSymbolType   *value )               /* value union                  */

  {
   /* AddMvclSymbolToTable() -- This function will add the symbol to the
    *     table.
    *
    *     RETURNS:  0 on success.
    *              -1 on failure.
    */
   MvclSymbol   *new;

   /* Allocate the memory required for the new node. */
   new = (MvclSymbol *)malloc( sizeof( MvclSymbol ) );

   if( new != (MvclSymbol *)NULL )
     {
      /* Copy the name and the type into the new symbol table entry. */
      new->SymName = strdup( name );
      assert( new->SymName );

      /* Let's determine the type of symbol that we're about to add. */
      new->SymType = type;
      switch( new->SymType )
        {
         case MVCL_SYM_INT:
           /* cast and store an int. */
           new->def.IntVal = value->IntVal;
           break;

         case MVCL_SYM_DOUBLE:
           /* cast and store a double. */
           new->def.DoubleVal = value->DoubleVal;
           break;

         case MVCL_SYM_STRING:
           /* cast and store a string. */
           new->def.StringVal = strdup( value->StringVal );
           assert( new->def.StringVal );
           break;
        }

      /* The symbol table entry is complete. Add it to the table. */
      new->next = (MvclSymbol *)NULL;

      /* If it's the first node... */
      if( SymbolTable->head == (MvclSymbol *)NULL )
        {
         SymbolTable->head = new;
         SymbolTable->tail = new;
        }
      else
        {
         SymbolTable->tail->next = new;
         SymbolTable->tail = SymbolTable->tail->next;
        }

      #ifdef MVCL_DEBUG
      printf( "Added the symbol '%s' to the symbol table.\n", name );
      #endif
     }

   return( (void *)new );
  }
/***************************************************************************/
int  UpdateMvclSymbolInTable(
   MvclSymbolTable  *SymbolTable,          /* symbol table to act on       */
   char             *name,                 /* pointer to symbol name       */
   int              type,                  /* type is int double or string */
   MvclSymbolType   *value )               /* value union                  */

  {
   /* UpdateMvclSymbolInTable() -- This function will search the symbol table
    *     for an identifier named "name". Then update it's value.
    *
    *     RETURNS:  0 on success
    *              -1 on failure.
    */
   MvclSymbol  *current;
   int         found;

   /* assign loop control stuff. */
   current = SymbolTable->head;
   found = 0;

   /* Search the list. */
   while( current != (MvclSymbol *)NULL && !found )
     {
      if( !strcmp( name, current->SymName ) )
        {
         /* Found it! */
         found = 1;

         /* Let's determine the type of symbol that we're about to update. */
         current->SymType = type;
         switch( current->SymType )
           {
            case MVCL_SYM_INT:
              /* cast and store an int. */
              current->def.IntVal = value->IntVal;

              #ifdef MVCL_DEBUG
              printf( "Updating integer symbol '%s' to %d.\n",
                       name, value->IntVal );
              #endif
              break;

            case MVCL_SYM_DOUBLE:
              /* cast and store a double. */
              current->def.DoubleVal = value->DoubleVal;

              #ifdef MVCL_DEBUG
              printf( "Updating real symbol '%s' to %f.\n",
                       name, value->DoubleVal);
              #endif
              break;

            case MVCL_SYM_STRING:
              /* cast and store a string. */
              current->def.StringVal = strdup( value->StringVal );
              assert( current->def.StringVal );

              #ifdef MVCL_DEBUG
              printf( "Updating string symbol '%s' to %s.\n",
                      name, value->StringVal );
              #endif
              break;
           }
        }

      current = current->next;
     }

   return( found == 0 ? 0 : -1 );
  }
/****************************************************************************/
MvclSymbol *  QueryMvclSymbolTable(
   MvclSymbolTable  *SymbolTable,           /* symbol table to act upon     */
   char             *name )                 /* name of symbol to query      */

  {
   /* QueryMvclSymbolTable() -- This function will look in the symbol table 
    *     for the node containing the symbol "name".
    *
    *     RETURNS: a pointer to the node that contains the symbol.
    *              NULL for an error.
    */
   MvclSymbol  *current;
   int         found;

   /* assign loop control stuff. */
   current = SymbolTable->head;
   found = 0;

   /* Search the list. */
   while( current != (MvclSymbol *)NULL && !found )
     {
      if( !strcmp( name, current->SymName ) )
        {
         found = 1;
         continue;
        }

      /* Advance link. */
      current = current->next;
     }

   return( current );
  }
/****************************************************************************/
void  DestroyMvclSymbolTable(
   MvclSymbolTable  *SymbolTable )         /* symbol table to act on       */

  {
   /* DestroyMvclSymbolTable() -- This function will destroy an mvcl symbol
    *     table.
    *
    *     RETURNS: nothing.
    */
   MvclSymbol  *current;
   MvclSymbol  *previous;

   /* assign loop control stuff. */
   current = SymbolTable->head;

   /* Search the list. */
   while( current != (MvclSymbol *)NULL )
     {
      previous = current;
      current = current->next;

      /* First free up the name entry. */
      if( previous->SymName != (char *)NULL )
        free( previous->SymName );

      /* If this is a string type symbol then free that as well. */
      if( previous->SymType == MVCL_SYM_STRING )
        {
         if( previous->def.StringVal != (char *)NULL )
           free( previous->def.StringVal );
        }

      /* Now free up the node memory. */
      free( previous );
     }

   return;
  }
/***************************************************************************/
/* finis */
