/*                         MICRION SOFTWARE DEPT.
 *          Copyright (c) 1995 by Micrion Corp. All rights reserved.
 *
 * FILE NAME    : mvcl_utl.c
 * AUTHOR       : R. Mello
 * DATE WRITTEN : 7/6/95
 * VERSION      : 1.00
 * USAGE        : module of the mcvl parser
 *
 * This source file contains some utility functions.
 */
#include <stdio.h>                          /* for NULL, fprintf(), ect...   */
#include <string.h>
#include <mvcl_int.h>

/* function prototypes. */
char *    MvclResolveDeviceName( int, int );
char *    MvclGenerateVariable( void );
char *    MvclResolveComparison( int );
int       MvclLinSearchForMatch( int, int *, int );
char *    MvclResolveDeviceStatus( int, int );
char *    MvclStripQuotesOffString( char * );
int       MvclConvertComparisonToId( char * );
char *    MvclCreateRealSymbol( MvclProgramList *, double );
char *    MvclCreateIntSymbol( MvclProgramList *, int );
char *    MvclStatementTypeToName( int );
void      MvclAddToPrintList( MvclPrintItem **, char *, int );

/****************************************************************************/
char *   MvclResolveDeviceName(
   int   dev_class,                         /* class of the dev to resolve  */
   int   dev_id )                           /* id of the specific device    */

  {
   /* MvclResolveDeviceName() -- This function will take a device class and
    *     device id specification and return the device name.
    *
    *     This function returns a pointer to a static buffer. DO NOT attempt
    *     to free the memory returned from this function.
    *
    *     RETURNS: the name of the device or 
    *              NULL for an unknown device
    */
   static char   Abuffer[64];

   /* Clear Abuffer. */
   Abuffer[0] = '\0';

   /* Based on the device class. */
   switch( dev_class )
     {
      case MVCL_VALVE_CLASS:
        sprintf( Abuffer, "v%d", dev_id );                
        break;

      case MVCL_PUMP_CLASS:
        sprintf( Abuffer, "p%d", dev_id );                
        break;

      case MVCL_HCIG_CLASS:
        sprintf( Abuffer, "hcig%d", dev_id );                
        break;

      case MVCL_CCIG_CLASS:
        sprintf( Abuffer, "ccig%d", dev_id );                
        break;
  
      case MVCL_BT_CLASS:
        sprintf( Abuffer, "bt%d", dev_id );                
        break;

      case MVCL_TC_CLASS:
        sprintf( Abuffer, "tc%d", dev_id );                
        break;

      case MVCL_NZ_CLASS:
        sprintf( Abuffer, "nz%d", dev_id );                
        break;

      case MVCL_PT_CLASS:
        sprintf( Abuffer, "pt%d", dev_id );                
        break;

      case MVCL_MFC_CLASS:
        sprintf( Abuffer, "mfc%d", dev_id );                
        break;
     }

   return( Abuffer );
  }
/****************************************************************************/
char *  MvclGenerateVariable( void )

  {
   /* MvclGenerateVariable( void ) -- This function will create a new variable
    *     each time it is called. 
    *
    *     RETURNS: a pointer to the new variable name
    *              NULL for an error.
    */
   char    new_var[32];
   static int  index;

   sprintf( new_var, "var%d", index++ );

   return( strdup( new_var ) );
  }
/****************************************************************************/
char *    MvclResolveComparison(
   int    comparison_id )                   /* internal rep of comparison    */

  {
   /* MvclResolveComparison() -- This function will convert mvcl's internal
    *     representation of a comparison to the languages token.
    *
    *     This function returns a pointer to a static buffer. DO NOT attempt
    *     to free the memory returned from this function.
    *
    *     RETURNS: a pointer to the comparison string.
    *              NUL for an error.
    */
   static char   Abuffer[64];

   /* Clear Abuffer. */
   Abuffer[0] = '\0';

   /* switch based on the internal comparison type. */
   switch( comparison_id )
     {
      case MVCL_EQUAL:
        sprintf( Abuffer, "==" );
        break;

      case MVCL_NOTEQUAL:
        sprintf( Abuffer, "!=" );
        break;

      case MVCL_LESSTHAN:
        sprintf( Abuffer, "<" );
        break;

      case MVCL_GREATERTHAN:
        sprintf( Abuffer, ">" );
        break;

      case MVCL_LESS_EQUAL:
        sprintf( Abuffer, "<=" );
        break;

      case MVCL_GREATER_EQUAL:
        sprintf( Abuffer, ">=" );
        break;
     }
 
   return( Abuffer );
  }
/****************************************************************************/
int  MvclLinSearchForMatch(
   int     SearchValue,                     /* value to find in the array    */
   int     *TargetArray,                    /* array to search in            */
   int     TargetSize )                     /* size of the array             */
  
  {
   /* MvclLinSearchForMatch() -- This function will search the array passed
    *     for the search value using a simple linear search through the array.
    *
    *     RETURNS: the array index of the found element.
    *              -1 if the search value was not found
    */
   int  index = -1;
   int  i;

   /* For all array elements... */
   for( i = 0; i < TargetSize; i++ )
     {
      if( TargetArray[i] == SearchValue )
        {
         index = i;
         break;
        }
     }

   return( index );
  }
/****************************************************************************/
char *  MvclResolveDeviceStatus(
   int    dev_class,                        /* class of device              */
   int    status )                          /* it's status                  */

  {
   /* MvclResolveDeviceStatus() -- This function will take the device class and
    *     the status returned and come out with a string that states the 
    *     devices status.
    *
    *     This function returns a pointer to a static buffer. DO NOT attempt
    *     to free the memory returned from this function.
    *
    *     RETURNS: the name of the device or 
    *              NULL for an unknown device
    */
   static char   Abuffer[64];

   /* Clear Abuffer. */
   Abuffer[0] = '\0';

   /* generally if the status is -1 then the state is unknown. */
   if( status == -1 )
     sprintf( Abuffer, "%s", "State Unknown" );                
   else
     {
      /* Based on the device class. */
      switch( dev_class )
        {
         case MVCL_VALVE_CLASS:
           sprintf( Abuffer, "%s", status ? "Open" : "Closed" );
           break;

         case MVCL_PUMP_CLASS:
           sprintf( Abuffer, "%s", status ? "Running" : "Off" ); 
           break;

         case MVCL_HCIG_CLASS:
           sprintf( Abuffer, "%s", status ? "On" : "Off" );                
           break;

         case MVCL_CCIG_CLASS:
           sprintf( Abuffer, "%s", status ? "On" : "Off" );                
           break;
  
         case MVCL_BT_CLASS:
           sprintf( Abuffer, "%s", status ? "Up" : "Down" );                
           break;

         case MVCL_TC_CLASS:
           sprintf( Abuffer, "%s", status ? "Up" : "Down" );                
           break;

         case MVCL_NZ_CLASS:
           sprintf( Abuffer, "%s", status ? "Extended" : "Retracted" );        
           break;

         case MVCL_PT_CLASS:
           sprintf( Abuffer, "%s", status ? "Up" : "Down" );                
           break;

         case MVCL_MFC_CLASS:
           sprintf( Abuffer, "%s", status ? "Up" : "Down" );                
           break;
        }
     }

   return( Abuffer );
  }
/****************************************************************************/
char * MvclStripQuotesOffString(
   char   *Qtext )                                   /* matched pattern     */

  {
   /* MvclStripQuotesOffString() -- This function will remove the two quotes
    *     from around the quoted string.
    *
    *     RETURNS: the un-quoted ( is that a word? ) string.
    */
   int      len;
   char     *str;

   /* get the length of the string */
   len = strlen( Qtext );

   /* allocate memory for the string without the double quotes. Then add the
    * null term space in.
    */
   str = (char *)malloc( len - 1 );

   if( str != (char *)NULL )
     {
      /* Null terminate the string so we can use the standard C library stuff*/
      *str = '\0';

      /* remove the leading quote be copying the string over it. */
      strncpy( str, Qtext + 1, len - 2 );

      /* remove the trailing quote by replacing that position with a
       * null term
       */
      str[ len - 2 ] = '\0';
     }

   return( str );
  }
/****************************************************************************/
int  MvclConvertComparisonToId(
   char    *compare_string )               /* compare token string          */

  {
   /* MvclConvertComparisonToId() -- This function will convert the comparison
    *     token string to an internal representation.
    *
    *     RETURNS: the new type.
    */
   int   rep;

   /* convert... */
   if( !strcmp( compare_string, "==" ) )
     rep = MVCL_EQUAL;
   else if( !strcmp( compare_string, "!=" ) )
     rep = MVCL_NOTEQUAL;
   else if( !strcmp( compare_string, "<" ) )
     rep = MVCL_LESSTHAN;
   else if( !strcmp( compare_string, ">" ) )
     rep = MVCL_GREATERTHAN;
   else if( !strcmp( compare_string, "<=" ) )
     rep = MVCL_LESS_EQUAL;
   else if( !strcmp( compare_string, ">=" ) )
     rep = MVCL_GREATER_EQUAL;
   else
     rep = MVCL_COMP_UNKNOWN;

   return( rep );
  }
/****************************************************************************/
char *  MvclCreateRealSymbol(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   double     value )                      /* symbol's initial value       */

  {
   /* MvclCreateRealSymbol() -- This function will create a symbol table entry
    *     by first creating a new variable name and assigning the value
    *     passed to it.
    *
    *     RETURNS: a pointer to the new variable name that must be freed.
    */
   char   *name;
   MvclSymbolType  sym;

   /* Generate a new variable name. */
   name = MvclGenerateVariable();

   if( name != (char *)NULL )
     {
      /* Add this entry into the symbol table. */
      sym.DoubleVal = value;
      AddMvclSymbolToTable( prog->SymbolTable, name, MVCL_SYM_DOUBLE, &sym );
     }

   return( name );
  }
/****************************************************************************/
char *  MvclCreateIntSymbol(
   MvclProgramList  *prog,                 /* Pointer to program list      */
   int        value )                      /* symbol's initial value       */

  {
   /* MvclCreateIntSymbol() -- This function will create a symbol table entry
    *     by first creating a new variable name and assigning the value
    *     passed to it.
    *
    *     RETURNS: a pointer to the new variable name that must be freed.
    */
   char       *name;
   MvclSymbolType sym;

   /* Generate a new variable name. */
   name = MvclGenerateVariable();

   if( name != (char *)NULL )
     {
      /* Add this entry into the symbol table. */
      sym.IntVal = value;
      AddMvclSymbolToTable( prog->SymbolTable, name, MVCL_SYM_INT, &sym );
     }

   return( name );
  }
/****************************************************************************/
char *  MvclStatementTypeToName(
   int     type )

  {
   /* MvclStatementTypeToName() -- This function will convert a statement
    *     type to the name of the statement. This is real handy for debugging.
    *
    *     This function returns a pointer to a static buffer. DO NOT attempt
    *     to free the memory returned from this function.
    *
    *     RETURNS: the name of the device or 
    *              NULL for an unknown device
    */
   static char   Abuffer[64];

   /* Clear Abuffer. */
   Abuffer[0] = '\0';

   /* Based on the device class. */
   switch( type )
     {
      case MVCL_OPEN:
        sprintf( Abuffer, "%s", "Open" );
        break;

      case MVCL_CLOSE:
        sprintf( Abuffer, "%s", "Close" );
        break;

      case MVCL_STATUS:
        sprintf( Abuffer, "%s", "Status" );
        break;

      case MVCL_SET:
        sprintf( Abuffer, "%s", "Set" );
        break;

      case MVCL_READ:
        sprintf( Abuffer, "%s", "Read" );
        break;

      case MVCL_TIMEOUT:
        sprintf( Abuffer, "%s", "Timeout" );
        break;

      case MVCL_PUMPON:
        sprintf( Abuffer, "%s", "PumpOn" );
        break;

      case MVCL_PUMPOFF:
        sprintf( Abuffer, "%s", "PumpOff" );
        break;

      case MVCL_EXTEND:
        sprintf( Abuffer, "%s", "Extend" );
        break;

      case MVCL_RETRACT:
        sprintf( Abuffer, "%s", "Retract" );
        break;

      case MVCL_WAIT_FOR:
        sprintf( Abuffer, "%s", "WaitFor" );
        break;

      case MVCL_WAIT:
        sprintf( Abuffer, "%s", "Wait" );
        break;

      case MVCL_IF:
        sprintf( Abuffer, "%s", "If" );
        break;

      case MVCL_ELSE:
        sprintf( Abuffer, "%s", "Else" );
        break;

      case MVCL_ENDIF:
        sprintf( Abuffer, "%s", "Endif" );
        break;

      case MVCL_FOR:
        sprintf( Abuffer, "%s", "For" );
        break;

      case MVCL_ENDFOR:
        sprintf( Abuffer, "%s", "EndFor" );
        break;

      case MVCL_NULL:
        sprintf( Abuffer, "%s", "Null" );
        break;

      case MVCL_PRINT:
        sprintf( Abuffer, "%s", "Print" );
        break;

     }

   return( Abuffer );
  }
/****************************************************************************/
void  MvclAddToPrintList(
   MvclPrintItem  **list,                   /* pointer to the list head      */
   char           *pstring,                 /* basic type of print           */
   int            type )                    /* type of node                  */

  {
   /* MvclAddToPrintList() -- This function will append this print statement
    *     node to the list of print arguments.
    *
    *     RETURNS: nothing.
    */
   MvclPrintItem   *current;
   MvclPrintItem   *previous = (MvclPrintItem *)NULL;
   MvclPrintItem   *new;

   /* Create an instance of a print list node. */
   new = (MvclPrintItem *)malloc( sizeof( MvclPrintItem ) );

   if( new != (MvclPrintItem *)NULL )
     {
      /* Make sure the new node's next pointer is null. */
      new->next = (MvclPrintItem *)NULL;

      /* Store the data. */
      new->string = pstring;
      new->type = type;

      /* Find the end of the list... Flex builds the list backwards. So we
       * have to do one of two things. Allow the list to be created backwards
       * then order it when it's done. Or, search for the end of the list at
       * each addition. I chose method two. 
       */
      current = *list;
      while( current != (MvclPrintItem *)NULL )
	{
         previous = current;
         current = current->next;
        }

      /* If it's the first node... */
      if( previous == (MvclPrintItem *)NULL )
        *list = new;
      else
        previous->next = new;
     }

   return;
  }
/****************************************************************************/
void  MvclShellHelp( void )

  {
   /* MvclShellHelp() -- This function will display a help message when the
    *     "help" command is issued.
    *
    *     RETURNS: nothing.
    */

   fprintf( stderr, "This is the default help message.\n" );
   fprintf( stderr, "To install help replace this call to MvclShellHelp()\n" );
   fprintf( stderr, "with a local copy in the application. \n" );

   return;
  }
/****************************************************************************/
/* finis */
















