/*                         MICRION SOFTWARE DEPT.
 *          Copyright (c) 1995 by Micrion Corp. All rights reserved.
 *
 * FILE NAME    : mvcl_vld.c
 * AUTHOR       : R. Mello
 * DATE WRITTEN : 5/16/95
 * VERSION      : 1.00
 * USAGE        : module of the mvcl interpreter
 *
 * This source file contains code to validate a users request to operate
 * on a hardware specific device.
 */
#include <stdio.h>                          /* for NULL, fprintf(), ect...   */
#include <mvcl_int.h>

/* file equates */
#define MAX_VALVES    30
#define MAX_PUMPS     10
#define MAX_TCS       10
#define MAX_BTS       10
#define MAX_NOZZLES   10
#define MAX_PTS       10
#define MAX_MFCS      10
#define INVALID       0
#define VALID         1
#define NOT_FOUND    -1

/* uncomment the following line to run parser with hardware control */
/* #define HARDWARE_AVAIL */

/* Create a structure that will represent the vacuum system. --
 * This structure will get initialized the first time any mvcl command
 * gets executed for the current process. It will describe the vacuum
 * system in such a way it will allow validation of the hardware requests
 * that mvcl will be making.
 */
struct  VacSys_tag
  {
   char   MachineType[32];
   int    ValveList[MAX_VALVES];
   int    PumpList[MAX_PUMPS];
   int    TCList[MAX_TCS];
   int    BTList[MAX_BTS];
   int    NozzleList[MAX_NOZZLES];
   int    PTList[MAX_PTS];
   int    MFCList[MAX_MFCS];
   int    HcigPresent;
   int    CcigPresent;
  };

/* vacuum system model. */
static struct VacSys_tag VacSys;

/* function prototypes. */
void  MvclQueryVacuumSystem( char * );
int   MvclValidateDevice( int, int );
int   MvclValidateValve( int );
int   MvclValidatePump( int );
int   MvclValidateTC( int );
int   MvclValidateBT( int );
int   MvclValidateNozzle( int );
int   MvclValidatePT( int );
int   MvclValidateMFC( int );
int   MvclValidateHCIG( int );
int   MvclValidateCCIG( int );

/***************************************************************************/
void  MvclQueryVacuumSystem( 
   char    *fname )

  {
   /* MvclQueryVacuumSystem() -- This function wil be called when the procedure
    *      is first started. It will be reponsible for determining the current
    *      system this program is running on.
    *
    *      RETURNS: nothing.
    */
   static int  Initialized = 0;
   int         *tmp_array;
   int         i;
   int         size;
   FILE        *fp;

   /* Invalidate everything. */
   if( !Initialized )
     {
      VacSys.MachineType[0] = '\0';

      for( i = 0; i < MAX_VALVES; i++ )
        VacSys.ValveList[i] = INVALID;

      for( i = 0; i < MAX_PUMPS; i++ )
        VacSys.PumpList[i] = INVALID;

      for( i = 0; i < MAX_TCS; i++ )
        VacSys.TCList[i] = INVALID;

      for( i = 0; i < MAX_BTS; i++ )
        VacSys.BTList[i] = INVALID;

      for( i = 0; i < MAX_NOZZLES; i++ )
        VacSys.NozzleList[i] = INVALID;

      for( i = 0; i < MAX_PTS; i++ )
        VacSys.PTList[i] = INVALID;

      for( i = 0; i < MAX_MFCS; i++ )
        VacSys.MFCList[i] = INVALID;

      VacSys.HcigPresent = INVALID;
      VacSys.CcigPresent = INVALID;

      fp = fopen( fname, "r" );

      if( fp != (FILE *)NULL )
        {
         tmp_array = MvclReadConfigOptions( fp, "ValveList", &size );

         if( tmp_array != (int *)NULL )
           {
            for( i = 0; i < size; i++ )
              VacSys.ValveList[i] = tmp_array[i];

            free( tmp_array );
           }

         tmp_array = MvclReadConfigOptions( fp, "PumpList", &size );

         if( tmp_array != (int *)NULL )
           {
            for( i = 0; i < size; i++ )
              VacSys.PumpList[i] = tmp_array[i];

            free( tmp_array );
           }

         tmp_array = MvclReadConfigOptions( fp, "TCList", &size );

         if( tmp_array != (int *)NULL )
           {
            for( i = 0; i < size; i++ )
              VacSys.TCList[i] = tmp_array[i];

            free( tmp_array );
           }

         tmp_array = MvclReadConfigOptions( fp, "BTList", &size );

         if( tmp_array != (int *)NULL )
           {
            for( i = 0; i < size; i++ )
              VacSys.BTList[i] = tmp_array[i];

            free( tmp_array );
           }

         tmp_array = MvclReadConfigOptions( fp, "PTList", &size );

         if( tmp_array != (int *)NULL )
           {
            for( i = 0; i < size; i++ )
              VacSys.PTList[i] = tmp_array[i];

            free( tmp_array );
           }

         tmp_array = MvclReadConfigOptions( fp, "NozzleList", &size );

         if( tmp_array != (int *)NULL )
           {
            for( i = 0; i < size; i++ )
              VacSys.NozzleList[i] = tmp_array[i];

            free( tmp_array );
           }

         tmp_array = MvclReadConfigOptions( fp, "MFCList", &size );

         if( tmp_array != (int *)NULL )
           {
            for( i = 0; i < size; i++ )
              VacSys.MFCList[i] = tmp_array[i];

            free( tmp_array );
           }

         tmp_array = MvclReadConfigOptions( fp, "HcigPresent", &size );

         if( tmp_array != (int *)NULL )
           {
            for( i = 0; i < size; i++ )
              VacSys.HcigPresent = tmp_array[i];

            free( tmp_array );
           }

         tmp_array = MvclReadConfigOptions( fp, "CcigPresent", &size );

         if( tmp_array != (int *)NULL )
           {
            for( i = 0; i < size; i++ )
              VacSys.CcigPresent = tmp_array[i];

            free( tmp_array );
           }

         fclose( fp );
        }
 
      /* we're initialized... */
      Initialized = 1;
     }

   return;
  }
/***************************************************************************/
int   MvclValidateDevice(
   int    dev_class,                        /* Class of device to validate  */
   int    dev_id )                          /* Special device to check      */

  {
   /* MvclValidateDevice() -- This function will test to make sure that the
    *     device is part of the current system.
    *
    *     RETURNS: 0 if the device is NOT present.
    *              1 if it is.
    */
   int   there = 0;

   /* Based on the class of device... */
   switch( dev_class )
     {
      case MVCL_VALVE_CLASS:
        there = MvclValidateValve( dev_id );
        break;

      case MVCL_PUMP_CLASS:
        there = MvclValidatePump( dev_id );
        break;

      case MVCL_TC_CLASS:
        there = MvclValidateTC( dev_id );
        break;

      case MVCL_BT_CLASS:
        there = MvclValidateBT( dev_id );
        break;

      case MVCL_NZ_CLASS:
        there = MvclValidateNozzle( dev_id );
        break;

      case MVCL_HCIG_CLASS:
        there = MvclValidateHCIG( dev_id );
        break;

      case MVCL_CCIG_CLASS:
        there = MvclValidateCCIG( dev_id );
        break;

      case MVCL_PT_CLASS:
        there = MvclValidatePT( dev_id );
        break;

      case MVCL_MFC_CLASS:
        there = MvclValidateMFC( dev_id );
        break;
     }

   return( there );
  }
/***************************************************************************/
int  MvclValidateValve(
   int    valve )                          /* id of valve to check         */

  {
   /* MvclValidateValve() -- This function will check to see if the requested
    *     valve argument is in fact one in the system. It will look in the
    *     VacSys data structure maintained by this module for the valve.
    *
    *     RETURNS: 0 if the valve is NOT present.
    *              1 if it is.
    */
   int     there = 0;

   if( MvclLinSearchForMatch( valve, VacSys.ValveList, MAX_VALVES ) != -1 )
     there = 1;

   return( there );
  }
/***************************************************************************/
int   MvclValidatePump(
   int    pump )                           /* id of pump to check          */

  {
   /* MvclValidatePump() -- This function will check to see if the requested
    *     pump argument is in fact one in the system. It will look in the
    *     VacSys data structure maintained by this module for the pump.
    *
    *     RETURNS: 0 if the pump is NOT present.
    *              1 if it is.
    */
   int     there = 0;

   if( MvclLinSearchForMatch( pump, VacSys.PumpList, MAX_PUMPS ) != -1 )
     there = 1;

   return( there );
  }
/***************************************************************************/
int   MvclValidateTC(
   int   tc )                              /* id of TC to validate          */

  {
   /* MvclValidateTC() -- This function will check to see if the requested
    *     TC gauge is in fact one in the system. It will look in the
    *     VacSys data structure maintained by this module for the gauge.
    *
    *     RETURNS: 0 if the TC is NOT present.
    *              1 if it is.
    */
   int     there = 0;

   if( MvclLinSearchForMatch( tc, VacSys.TCList, MAX_TCS ) != -1 )
     there = 1;

   return( there );
  }
/***************************************************************************/
int   MvclValidateBT(
   int   bt )                              /* bt of TC to validate          */

  {
   /* MvclValidateBT() -- This function will check to see if the requested
    *     BT gauge is in fact one in the system. It will look in the
    *     VacSys data structure maintained by this module for the gauge.
    *
    *     RETURNS: 0 if the BT is NOT present.
    *              1 if it is.
    */
   int     there = 0;

   if( MvclLinSearchForMatch( bt, VacSys.BTList, MAX_BTS ) != -1 )
     there = 1;

   return( there );
  }
/***************************************************************************/
int   MvclValidateNozzle(
   int   nozzle )                              /* id of Nozzle to validate  */

  {
   /* MvclValidateNozzle() -- This function will check to see if the nozzle
    *     is, in fact, in the system. It will look in the VacSys data
    *     structure maintained by this module for the gauge.
    *
    *     RETURNS: 0 if the nozzle is NOT present.
    *              1 if it is.
    */
   int     there = 0;

   if( MvclLinSearchForMatch( nozzle, VacSys.NozzleList, MAX_NOZZLES ) != -1 )
     there = 1;

   return( there );
  }
/***************************************************************************/
int   MvclValidatePT(
   int   pt )                                 /* id of Nozzle to validate  */

  {
   /* MvclValidatePT() -- This function will check to see if the pressure
    *     transducer is, in fact, in the system. It will look in the VacSys
    *      data structure maintained by this module for the data.
    *
    *     RETURNS: 0 if the pt is NOT present.
    *              1 if it is.
    */
   int     there = 0;

   if( MvclLinSearchForMatch( pt, VacSys.PTList, MAX_PTS ) != -1 )
     there = 1;

   return( there );
  }
/***************************************************************************/
int   MvclValidateMFC(
   int   mfc )                                 /* id of Nozzle to validate  */

  {
   /* MvclValidateMFC() -- This function will check to see if the mass flow
    *     controller is, in fact, in the system. It will look in the VacSys
    *     data structure maintained by this module for the data.
    *
    *     RETURNS: 0 if the mfc is NOT present.
    *              1 if it is.
    */
   int     there = 0;

   if( MvclLinSearchForMatch( mfc, VacSys.MFCList, MAX_MFCS ) != -1 )
     there = 1;

   return( there );
  }
/***************************************************************************/
int   MvclValidateHCIG(
   int   hcig )                              /* id of HCIG to validate     */

  {
   /* MvclValidateHCIG() -- This function will check to see if the HCIG
    *     is, in fact, in the system. It will look in the VacSys data
    *     structure maintained by this module for the gauge.
    *
    *     RETURNS: 0 if the HCIG is NOT present.
    *              1 if it is.
    */

   return( VacSys.HcigPresent );
  }
/***************************************************************************/
int   MvclValidateCCIG(
   int   ccig )                              /* id of CCIG to validate     */

  {
   /* MvclValidateCCIG() -- This function will check to see if the CCIG
    *     is, in fact, in the system. It will look in the VacSys data
    *     structure maintained by this module for the gauge.
    *
    *     RETURNS: 0 if the CCIG is NOT present.
    *              1 if it is.
    */

   return( VacSys.CcigPresent );
  }
/***************************************************************************/
/* finis */

