/*
 * Invoke.java --
 *
 *	This class implements the common routines used by both
 *	NativeTcl and Jacl to access the Java Reflection API.
 *
 * Copyright (c) 1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL
 * WARRANTIES.
 *
 * SCCS: @(#) Invoke.java 1.22 97/10/21 11:45:47
 *
 */

package tcl.lang;

import java.lang.reflect.*;
import java.util.*;
import java.beans.*;

/*
 * This class implements the common routines used by both NativeTcl
 * and Jacl to access the Java Reflection API. It does the following:
 *
 *	- Match Java method ID's with the "signatures" used by Tcl.
 *	- Convert arguments passed by Tcl into the appropriate types
 *	  accepted by the reflection API.
 *
 */

class Invoke {

/*
 * This variable is set to true if we are doing internal regression
 * tests for Tcl inside Sun. It should always be false inside a
 * distribution version of Tcl.
 *
 * If this variable is true, we can access package protected members
 * of the tcl.lang package using the java:: commands.
 */

static boolean tclTest = false;

/*
 * Stores the lists of accessible methods of classes that have been
 * accessed by the reflection API.
 */

private static Hashtable accMethTable = new Hashtable();

/*
 * We need to use empty array Object[0] a lot. We keep a static copy
 * and re-use it to avoid garbage collection.
 */

static private Object EMPTY_ARGS[] = new Object[0];


/*
 *----------------------------------------------------------------------
 *
 * callConstructor --
 *
 *	Call the specified constructor.
 *
 * Results:
 *	When successful, the object created by the constructor.
 *
 * Side effects:
 *	The constructor can cause arbitrary side effects.
 *
 *----------------------------------------------------------------------
 */

static Object
callConstructor(
    Object sig[],			// Signature of the constructor.
    Object args[])			// Argument list.
throws
    ClassNotFoundException,		// Exceptions thrown by various parts
    NoSuchMethodException,		//   of the Java reflection API.
    AmbiguousSignatureException,
    IllegalAccessException,
    IllegalArgumentException, 
    InstantiationException,
    InvocationTargetException
{
    String clsName = sig[0].toString();

    /* 
     * Temporary:  error is thrown if the class whose constructor is being
     * invoked is an array class.
     */

    if (clsName.endsWith("[]") || clsName.startsWith("[")) {
	throw new IllegalArgumentException(
	    "\n  Cannot create new instance of an array \""
            + clsName + "\":  not yet implemented."); 
    }

    Class cls = getClassByName(clsName);

    if (!(Modifier.isPublic(cls.getModifiers())) && !tclTest) {
	throw new IllegalAccessException(cls.toString());
    }

    Constructor constructor = (Constructor) matchSig(cls,
	    sig, args, true);
    return call(constructor, null, args, true);
}

/*
 *----------------------------------------------------------------------
 *
 * callMethod --
 *
 *	Call the specified instance method of the given object.
 *
 * Results:
 *	When successful, returns an array: Object result[2]. result[0]
 *	is the object returned by the method; result[1] is the return
 *	type of the method.
 *
 * Side effects:
 *	The method can cause arbitrary side effects.
 *
 *----------------------------------------------------------------------
 */

static Object[]
callMethod(
    Object object,			// The object whose method to invoke.
    Object sig[],			// The method signature.
    Object args[])			// Argument list.
throws
    ClassNotFoundException,		// Exceptions thrown by various parts
    NoSuchMethodException,		//   of the Java reflection API.
    AmbiguousSignatureException,
    IllegalAccessException,
    IllegalArgumentException, 
    InstantiationException,
    InvocationTargetException
{
    Method method = (Method)matchSig(object.getClass(), sig, args, false);

    Object value = call(method, object, args, false);
    Object result[] = new Object[2];
    result[0] = value;
    result[1] = method.getReturnType();

    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * callStaticMethod --
 *
 *	Call the specified static method method of the given class.
 *
 * Results:
 *	When successful, returns an array: Object result[2]. result[0]
 *	is the object returned by the method; result[1] is the return
 *	type of the method.
 *
 * Side effects:
 *	The method can cause arbitrary side effects.
 *
 *----------------------------------------------------------------------
 */

static Object[]
callStaticMethod(
    Class cls,				// The class whose static method to
					// invoke.
    Object sig[],			// The method signature.
    Object args[])			// Argument list.
throws
    NoSuchMethodException,		// If the method doesn't exist
					// or is not a static method.
    ClassNotFoundException,		// Exceptions thrown by various parts
    AmbiguousSignatureException,	//   of the Java reflection API.
    IllegalAccessException,
    IllegalArgumentException, 
    InstantiationException,
    InvocationTargetException
{
    if (!(Modifier.isPublic(cls.getModifiers())) && !tclTest) {
	throw new IllegalAccessException(cls.toString());
    }

    Method method = (Method)matchSig(cls, sig, args, false);
    if (!(Modifier.isStatic(method.getModifiers()))) {
	throw new NoSuchMethodException();
    }

    Object value = call(method, null, args, false);
    Object result[] = new Object[2];
    result[0] = value;
    result[1] = method.getReturnType();

    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * matchSig --
 *
 *	Try to get a constructor/method that matches with the
 *	signature and the parameters.
 *
 * Results:
 *	The method or constructor that matches with the
 *	signature and the parameters.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

private static Object
matchSig(
    Class cls,				// Class who constructor or method
					// to query.
    Object sig[],			// The constructor/method signature.
    Object args[],			// Argument list.
    boolean isConstructor)		// True if the signature represents
    					// a constructor; false if it
					// represents a method.
throws
    AmbiguousSignatureException,	// If the signature is ambiguous.
    ClassNotFoundException,		// If a class in the signature
					// cannot be found.
    NoSuchMethodException		// If the given method/constructor
					// doesn't exist.
{
    Object match = null;
    int numArgs = args.length;

    if ((sig.length > 1) || (sig.length == 1 && numArgs == 0)) {
	/*
	 * Two cases:
	 * -- (sig.length > 1): A signature has been given.
	 * -- (sig.length == 1 && numArgs == 0): A signature of no
	 *    parameters is implied.
	 */

	int sigNumArgs = sig.length - 1;
	Class paramTypes[] = new Class[sigNumArgs];

	for (int i=0; i<sigNumArgs; i++) {
	    /*
	     * (ToDo) We now restrict all parameter type specifier
	     * in the signature to be a string name. Is this OK??
	     */

	    String clsName = sig[i+1].toString();
	    paramTypes[i] = getClassByName(clsName);
	}

	if (isConstructor) {
	    match = cls.getConstructor(paramTypes);
	} else {
	    /*
	     * We can't use cls.getMethod() because it may return
	     * a unusable method ID (see comments in
	     * getAccessibleMethods).  We need to call
	     * getAccessibleMethods() and pick the public method
	     * whose name and parameter types matches with the
	     * signature.
	     */

	    Method methods[] = getAccessibleMethods(cls);
	    String methodName = sig[0].toString();
	    
	    for (int i=0; i<methods.length; i++) {
		if (!(methodName.equals(methods[i].getName()))) {
		    continue;
		}

		Class pt[] = methods[i].getParameterTypes();
		if (pt.length != paramTypes.length) {
		    continue;
		}

		boolean good = true;
		for (int j = 0; j < pt.length; j ++) {
		    if (!(pt[j].equals(paramTypes[j]))) {
			good = false;
			break;
		    }
		}
		if (good) {
		    match = methods[i];
		    break;
		}
	    }

	    if (match == null) {
		throw new NoSuchMethodException(methodName);
	    }
	}
    } else {
	/*
	 * No argument types has been specified. Use number of args to
	 * match with the functions. Throws execption if more than
	 * one function has the same number of args.
	 */

	Object funcs[];
	if (isConstructor) {
	    funcs = cls.getConstructors();
	} else {
	    funcs = getAccessibleMethods(cls);
	}
	Class paramTypes[];
	String methodName = sig[0].toString();
	for (int i=0; i<funcs.length; i++) {
	    if (isConstructor) {
		paramTypes = ((Constructor)funcs[i]).getParameterTypes();
	    } else {
		Method method = ((Method)funcs[i]);
		if (methodName.compareTo(method.getName()) != 0) {
		    continue;
		}
		paramTypes = method.getParameterTypes();
	    }

	    /*
	     * An method/constructor matches only if its name is an exact
	     * match and the number of args are the same.
	     *
	     * (ToDo): add more "guessing" rules depending on the forms
	     * types of the arguments.
	     */

	    if (paramTypes.length == numArgs) {
		if (match == null) {
		    match = funcs[i];
		} else {
		    throw new AmbiguousSignatureException(methodName);
		}
	    }
	}
	if (match == null) {
	    throw new NoSuchMethodException(methodName);
	}
    }

    return match;
}

/*
 *----------------------------------------------------------------------
 *
 * call --
 *
 *	Call the constructor, instance method, or static method with
 *	the given parameters. Check the parameter types and perform
 *	string->object conversion if necessary.
 *
 * Results:
 *	The object created by the constructor, or the return value
 *	of the method call.
 *
 * Side effects:
 *	The constructor/method call may have arbitrary side effects.
 *
 *----------------------------------------------------------------------
 */

private static Object
call(
    Object func,		// The Constructor or Method to call.
    Object obj,			// The object associated with an instace
				// method call. Should be null for
				// constructor calls and static method
				// calls.
    Object args[],		// Argument list.
    boolean isConstructor)	// True if the signature represents
    				// a constructor; false if it
				// represents a method.
throws
    IllegalAccessException, 	// Exceptions thrown by various parts
    IllegalArgumentException,	//   of the Java reflection API.
    InstantiationException,
    InvocationTargetException
{
    Class paramTypes[];
    Constructor cons = null;
    Method method = null;
    int i;

    if (isConstructor) {
	cons = (Constructor)func;
	paramTypes = cons.getParameterTypes();
    } else {
	method = (Method)func;
	paramTypes = method.getParameterTypes();
    }

    int numArgs = args.length;
    if (numArgs != paramTypes.length) {
	throw new IllegalArgumentException("Wrong # of arguments");
    }

    for (i=0; i<numArgs; i++) {
	if (args[i] instanceof String) {
	    String str = (String)args[i];
	    Class type = paramTypes[i];

	    /*
	     * To optimize performance, the following "if"
	     * statements are arranged according to (more or less)
	     * the frequency that a certain type is used.
	     */

	    if (type.equals(String.class)){
		args[i] = str;
	    } else if ((type == Integer.TYPE) || type == Integer.class) {
		args[i] = new Integer(str);
	    } else if ((type == Boolean.TYPE) || type == Boolean.class) {
		args[i] = new Boolean(str);
	    } else if ((type == Long.TYPE) || type == Long.class) {
		args[i] = new Long(str);
	    } else if ((type == Float.TYPE) || type == Float.class) {
		args[i] = new Float(str);
	    } else if ((type == Double.TYPE) || type == Double.class) {
		args[i] = new Double(str);
	    } else if ((type == Byte.TYPE) || type == Byte.class) {
		args[i] = new Byte(str);
	    } else if ((type == Short.TYPE) || type == Short.class) {
		args[i] = new Short(str);
	    } else if ((type == Character.TYPE)||type == Character.class) {
		if (str.length() == 1) {
		    args[i] = new Character(str.charAt(0));
		}
	    }
	}
    }

    if (isConstructor) {
	return cons.newInstance(args);
    } else {
	return method.invoke(obj, args);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * getField --
 *
 *	Returns the value of a field in the given object.
 *
 * Results:
 *	When successful, returns an array: Object result[2]. result[0]
 *	is the value of the field; result[1] is the type of the field.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static Object[]
getField(
    Class cls,			// The class which the object belongs to.
    Object obj,			// The object whose field to query.
    String field)		// The name of the field to query.
throws
    NoSuchFieldException,	// The field doesn't exist.
    IllegalAccessException	// The field is not public.
{
    /*
     * ToDo: should get a public field from the nearest public super class.
     */

    if (!(Modifier.isPublic(cls.getModifiers())) && !tclTest) {
	throw new IllegalAccessException(cls.toString());
    }

    Field f = cls.getField((String)field);

    Object value = f.get(obj);
    Object result[] = new Object[2];
    result[0] = value;
    result[1] = f.getType();

    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * setField --
 *
 *	Sets the value of a field in the given object.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When successful, the field is set to the given value.
 *
 *----------------------------------------------------------------------
 */

static void
setField(
    Class cls,			// The class which the object belongs to.
    Object obj,			// The object whose field to query.
    String field,		// The name of the field to query.
    Object value)		// New value for the field.
throws
    NoSuchFieldException,	// The field doesn't exist.
    IllegalAccessException	// The field is not public.
{
    /*
     * ToDo: should get a public field from the nearest public super class.
     */

    if (!(Modifier.isPublic(cls.getModifiers())) && !tclTest) {
	throw new IllegalAccessException(cls.toString());
    }

    Field f = cls.getField((String)field);
    Class type = f.getType();

    if ((value != null) && (value instanceof String)) {
	String str = (String) value;
	if (type.equals(String.class)) {
	    f.set(obj, str);
	} else if (type == Integer.TYPE) {
	    f.setInt(obj, Integer.parseInt(str));
	} else if (type == Boolean.TYPE) {
	    f.set(obj, new Boolean(str));
	} else if (type == Long.TYPE) {
	    f.setLong(obj, Long.parseLong(str));
	} else if (type == Float.TYPE) {
	    f.set(obj, new Float(str));
	} else if (type == Double.TYPE) {
	    f.set(obj, new Double(str));
	} else if (type == Byte.TYPE) {
	    f.setByte(obj, Byte.parseByte(str));
	} else if (type == Short.TYPE) {
	    f.setShort(obj, Short.parseShort(str));
	} else if ((type == Character.TYPE) && (str.length() == 1)) {
	    f.setChar(obj, str.charAt(0));
	} else {
	    f.set(obj, value);
	}
    } else {
	f.set(obj, value);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * getProperty --
 *
 *	Returns the value of a property in the given object.
 *
 * Results:
 *	When successful, returns an array: Object result[2]. result[0]
 *	is the value of the property; result[1] is the type of the
 *	property.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static Object[]
getProperty(
    Class cls,			// The class which the object belongs to.
    Object obj,			// The object whose property to query.
    String property)		// The name of the property to query.
throws
    NoSuchFieldException,	// The property doesn't exist.
    IllegalAccessException,	// The property is not public.
    InvocationTargetException,	// Error occurred when calling the GET method.
    IllegalArgumentException	// Wrong type of argument for the GET method.
{
    /*
     * ToDo: must call an accessible GET method.
     */

    handle_prop : {
	BeanInfo beanInfo;

	try {
	    beanInfo = Introspector.getBeanInfo(cls);
	} catch (IntrospectionException e) {
	    break handle_prop;
	}

	PropertyDescriptor desc[] = beanInfo.getPropertyDescriptors();

	if (desc == null) {
	    break handle_prop;
	}

	for (int i=0; i<desc.length; i++) {
	    if (desc[i].getName().equals(property)) {
		if (!(desc[i] instanceof IndexedPropertyDescriptor)) {
		    Method readMethod = desc[i].getReadMethod();
		    if (readMethod == null) {
			/*
			 * This is a write-only property. Can't read it.
			 */

			throw new IllegalAccessException();
		    }

		    Class params[] = readMethod.getParameterTypes();
		    if (params.length != 0) {
			/*
			 * Some error? readMethod should take no args!
			 */

			break handle_prop;
		    }

		    Object value = readMethod.invoke(obj, EMPTY_ARGS);
		    Object result[] = new Object[2];
		    result[0] = value;
		    result[1] = desc[i].getPropertyType();

		    return result;
		} else {
		    /*
		     * This is an indexed property. Donno how to handle
		     * yet.
		     */

		    throw new IllegalAccessException();
		}
	    }
	}
    }

    throw new NoSuchFieldException();
}

/*
 *----------------------------------------------------------------------
 *
 * setProperty --
 *
 *	Returns the value of a property in the given object.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When successful, the property will have the new value.
 *
 *----------------------------------------------------------------------
 */

static void
setProperty(
    Class cls,			// The class which the object belongs to.
    Object obj,			// The object whose property to query.
    String property,		// The name of the property to query.
    Object value)		// New value for the property.
throws
    NoSuchFieldException,	// The property doesn't exist.
    IllegalAccessException,	// The property is not public.
    InvocationTargetException,	// Error occurred when calling the SET method.
    IllegalArgumentException	// Wrong type of argument for the SET method.
{
    /*
     * ToDo: must call an accessible SET method.
     */

    handle_prop: {
	BeanInfo beanInfo;

	try {
	    beanInfo = Introspector.getBeanInfo(cls);
	} catch (IntrospectionException e) {
	    break handle_prop;
	}

	PropertyDescriptor desc[] = beanInfo.getPropertyDescriptors();
	if (desc == null) {
	    break handle_prop;
	}

	for (int i=0; i<desc.length; i++) {
	    if (desc[i].getName().equals(property)) {
		if (!(desc[i] instanceof IndexedPropertyDescriptor)) {
		    Method writeMethod = desc[i].getWriteMethod();
		    if (writeMethod == null) {
			/*
			 * Read-only property. Can't write to.
			 */

			throw new IllegalAccessException();
		    }

		    Class params[] = writeMethod.getParameterTypes();
		    if (params.length == 1) {
			/*
			 * Some error? readMethod should take one arg!
			 */

			break handle_prop;
		    } else {
			Object args[] = new Object[1];
			args[0] = value;
			writeMethod.invoke(obj, args);
			return;
		    }
		} else {
		    /*
		     * This is an indexed property. Donno how to handle
		     * yet.
		     */

		    throw new IllegalAccessException();
		}
	    }
	}
    }

    throw new NoSuchFieldException();
}

/*
 *----------------------------------------------------------------------
 *
 * getAccessibleMethods --
 *
 *	Returns the public method ID's of all the  methods of a class that
 *	can be invoked via the reflection API.
 *
 *	This function is needed because of the lameness of
 *	Class.getMethods() the the reflection API in
 *	general. Class.getMethods returns the method IDs of all the
 *	public methods declared in a class. However, if the class
 *	itself is not public, trying to use the method ID's to invoke
 *	methods will lead to IllegalAccessException. For example,
 *
 * 	Java code:
 *		public class Foo {
 *	    	    public void bar() {...}
 *	    	    public static Foo getFooImpl() {
 *			returns new FooImpl();
 *		    }
 *		}
 *		class FooImpl extends Foo {	// A package-protected class
 *	    	    public void bar();
 *		}
 *
 * 	Tcl code:
 *		set x [java::call Foo getFooImpl]
 *		$x bar
 *
 *	The "$x bar" call must use the Foo.bar method ID instead of
 *	FooImpl.bar.
 * 
 *	The solution is:
 *
 *	    If both of the following are true
 *
 *		1) cls is not public and 
 *		2) method.getDeclaringClass() is not a public class
 *
 *	    then search a public method ID of the same signature in
 *
 *		(1) the nearest public superclass of cls
 *		(2) all the public interfaces implemented by cls (or
 *	            its super classes.)
 * Results:
 *	An array of all the accessible method IDs of the given class.
 *
 * Side effects:
 *	The first time a class is queried, the array of the accessible
 *	method ID's is created and stored in a hash-table for future
 *	queries.
 *
 *----------------------------------------------------------------------
 */

private static Method[]
getAccessibleMethods(
    Class cls)				// The class to query.
{
    Method accMethods[];

    if (tclTest) {
	return cls.getDeclaredMethods();
    }

    if (Modifier.isPublic(cls.getModifiers())) {
	return cls.getMethods();
    }

    accMethods = (Method[])accMethTable.get(cls);
    if (accMethods != null) {
	return accMethods;
    }

    Vector vec = new Vector();
    Method methods[] = cls.getMethods();

    /*
     * Get the nearest public super class (npsCls).
     */

    Class npsCls = cls.getSuperclass();
    while ((npsCls != null) && 
	    !(Modifier.isPublic(npsCls.getModifiers()))) {
	npsCls = npsCls.getSuperclass();
    }

    int i;

    for (i=0; i<methods.length; i++) {
	/*
	 * Go through all the methods returned by Class.getMethods().
	 */

	Class declarCls = methods[i].getDeclaringClass();
	if (Modifier.isPublic(declarCls.getModifiers())) {
	    /*
	     * This method is declared in a public class. Good. Add it
	     * to the list of accessible methods.
	     */

	    vec.addElement(methods[i]);
	    continue;
	}

	/*
	 * We'll treat this method as accessible only if the method
	 * that it overrides (if any) comes from a public superclass
	 * or a public interface.
	 */

	Class params[] = methods[i].getParameterTypes();
	String methodName = methods[i].getName();

	try {
	    if (npsCls != null) {
		/*
		 * If the npsCls has such a public method, add it to
		 * the list. Otherwise a NoSuchMethodException will
		 * be raised and we'll go into the block1 below.
		 */

		vec.addElement(npsCls.getMethod(methodName, params));
		continue;
	    }
	} catch (NoSuchMethodException e) {;}

        block1: {
	    /*
	     * We go through all the public interfaces implemented
	     * by this class and its superclasses, and see if any
	     * of those interfaces declares the method as a public
	     * method.
	     */

	    for (Class c = cls; c != null; c=c.getSuperclass()) {
		Class intf[] = c.getInterfaces();

		for (int j=0; j<intf.length; j++) {
		    if (!Modifier.isPublic(intf[j].getModifiers())) {
			continue;
		    }

		    try {
			vec.addElement(intf[j].getMethod(
				methodName, params));
			break block1;
		    } catch (NoSuchMethodException e) {
			/*
			 * This interface doesn't support this method.
			 * Ignore the exception and go to the next
			 * interface.
			 */
		    }
		}
	    }
	}
    }

    accMethods = new Method[vec.size()];
    
    for (i=0; i<vec.size(); i++) {
	accMethods[i] = (Method)vec.elementAt(i);
    }

    accMethTable.put(cls, accMethods);
    return accMethods;
}

/*
 *----------------------------------------------------------------------
 *
 * getClassByName --
 *
 *	Returns Class object identified by the string name. We allow
 *	abbreviation of the java.lang.* class if there is no ambiguity:
 *	e.g., if there is no class whose fully qualified name is "String",
 *	then "String" means java.lang.String.
 *
 * Results:
 *	If successful, The Class object identified by the string name.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static Class
getClassByName(
    String clsNameArg)			// String name of the class.
throws
    ClassNotFoundException		// If the class cannot be found.
{
    String clsName = clsNameArg;
    int dimention;

    /*
     * If the string is of the for className[][]..., strip out the trailing []s
     * and record the dimention of the array.
     */

    String prefix = "";
    String suffix = "";
    for (dimention = 0; clsName.endsWith("[]"); dimention++) {
	int newLen = clsName.length() - 2;
	clsName = clsName.substring(0, newLen);
	prefix = prefix + "[";
    }

    if (clsName.indexOf('.') == -1) {
	if (dimention > 0) {
	    if (clsName.equals("int")) {
		return Class.forName(prefix + "I");
	    } else if (clsName.equals("boolean")) {
		return Class.forName(prefix + "Z");
	    } else if (clsName.equals("long")) {
		return Class.forName(prefix + "J");
	    } else if (clsName.equals("float")) {
		return Class.forName(prefix + "F");
	    } else if (clsName.equals("double")) {
		return Class.forName(prefix + "D");
	    } else if (clsName.equals("byte")) {
		return Class.forName(prefix + "B");
	    } else if (clsName.equals("short")) {
		return Class.forName(prefix + "S");
	    } else if (clsName.equals("char")) {
		return Class.forName(prefix + "C");
	    } else {
		prefix = prefix + "L<";
		suffix = ">;";
	    }
	} else {
	    if (clsName.equals("int")) {
		return Integer.TYPE;
	    } else if (clsName.equals("boolean")) {
		return Boolean.TYPE;
	    } else if (clsName.equals("long")) {
		return Long.TYPE;
	    } else if (clsName.equals("float")) {
		return Float.TYPE;
	    } else if (clsName.equals("double")) {
		return Double.TYPE;
	    } else if (clsName.equals("byte")) {
		return Byte.TYPE;
	    } else if (clsName.equals("short")) {
		return Short.TYPE;
	    } else if (clsName.equals("char")) {
		return Character.TYPE;
	    }
	}
	try {
	    return Class.forName(prefix + clsName + suffix);
	} catch (ClassNotFoundException e) {
	    try {
		return Class.forName(prefix + "java.lang." + clsName + suffix);
	    } catch (ClassNotFoundException e1) {
		throw new ClassNotFoundException(clsNameArg);
	    }
	}
    } else {
	if (dimention > 0) {
	    return Class.forName(prefix + "L<" + clsName + ">;");
	} else {
	    return Class.forName(clsName);
	}
    }
}

} // end Invoke

