/*
 * JavaFieldCmd.java --
 *
 *	Implements the built-in "java::field" command.
 *
 * Copyright (c) 1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL
 * WARRANTIES.
 *
 * SCCS: @(#) JavaFieldCmd.java 1.4 97/10/16 14:53:16
 *
 */

package tcl.lang;

/*
 * This class implements the built-in "java::field" command.
 */

class JavaFieldCmd implements Command {


/*----------------------------------------------------------------------
 *
 * cmdProc --
 *
 * 	This procedure is invoked to process the "java::field" Tcl
 * 	command.  See the user documentation for details on what they
 * 	do.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A standard Tcl result is stored in the interpreter.
 *
 *----------------------------------------------------------------------
 */

public void
cmdProc(
    Interp interp,			// Current interpreter.
    TclObject argv[])			// Argument list.
throws
    TclException			// A standard Tcl exception.
{
    boolean query = false;
    Class cls;
    Object obj;

    if (argv.length < 3) {
	throw new TclNumArgsException(interp, 1, argv,
		"objOrClass field ?value field value ...?");
    }

    if (argv.length == 3) {
	query = true;
    } else if ((argv.length % 2) != 0) {
	throw new TclNumArgsException(interp, 1, argv,
		"objOrClass field ?value field value ...?");
    }

    /*
     * In the java::field command, both class and object are allowed.
     */

    try {
	obj = ReflectObject.get(interp, argv[1]);
	cls = obj.getClass();
    } catch (TclException e) {
	try {
	    obj = null;
	    cls = Invoke.getClassByName(argv[1].toString());
	} catch (ClassNotFoundException e1) {
	    throw new TclException(interp, "unknown class or object \"" + 
		    argv[1]  + "\"");
	} catch (Exception e2) {
	    throw new ReflectException(interp, e2);
	}
    }

    if (query) {
	/*
	 * Query the field.
	 */

	Object sigList[] = getFieldSignature(interp, argv[2], cls);
	String fieldName = (String) sigList[0];
	Class fieldCls = (Class) sigList[1];

	try {
	    Object result[] = Invoke.getField(fieldCls, obj, fieldName);
	    ReflectObject.wrapJavaObject(interp, result);
	} catch (Exception e2) {
	    throw new ReflectException(interp, e2);
	}
    } else {
	/*
	 * Set one or more fields.
	 */

	for (int i=2; i<argv.length; i+=2) {
	    Object sigList[] = getFieldSignature(interp, argv[i], cls);
	    String fieldName = (String) sigList[0];
	    Class fieldCls = (Class) sigList[1];

	    Object jvalue = ReflectObject.wrapTclObject(interp,
		     argv[i+1]);

	    try {
		Invoke.setField(fieldCls, obj, fieldName, jvalue);
	    } catch (Exception e2) {
		throw new ReflectException(interp, e2);
	    }
	}
	interp.resetResult();
    }
}

/*
 *----------------------------------------------------------------------
 *
 * getFieldSignature --
 *
 *	Returns the fieldName and Class in a field signature.
 *
 * Results:
 *	When successful, an Object[2] array. Object[0] is the String
 *	name of the field. Object[1] is the Class which the field
 *	was declared.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

private Object[]
getFieldSignature(
    Interp interp,		// Current interpreter
    TclObject sigObj,		// The TclObject that holds a field signature.
    Class defaultCls)		// The default class to use if it's not
				// explicitly set in the signature.
throws
    TclException		// Standard Tcl exception.
{
    String fieldName;
    TclObject clsObj;
    Class cls;

    int len = TclList.getLength(interp, sigObj);
    if ((len < 1) || (len > 2)) {
	throw new TclException(interp, "bad field signature");
    }

    fieldName = TclList.index(interp, sigObj, 0).toString();
    clsObj = TclList.index(interp, sigObj, 1);

    if (clsObj != null) {
	try {
	    cls = Invoke.getClassByName(clsObj.toString());
	} catch (Exception e) {
	    throw new ReflectException(interp, e);
	}
    } else {
	cls = defaultCls;
    }
    
    Object sigList[] = new Object[2];
    sigList[0] = fieldName;
    sigList[1] = cls;

    return sigList;
}

} // end JavaFieldCmd

