/*
 * Interp.java
 *
 * Copyright (c) 1997 Cornell University.
 * Copyright (c) 1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL
 * WARRANTIES.
 * 
 * SCCS: @(#) Interp.java 1.12 97/10/20 17:21:00
 *
 */

package tcl.lang;

import java.util.*;
import java.io.*;
import java.net.*;

/*
 * The Tcl interpreter class.
 */
public class Interp {
    /*
     * This constructor is called when an Interp is being created to wrap
     * an already existing Tcl_Interp.
     */

    private Interp(long l)
    {
	interpPtr = l;
    }

    /*
     * Construct a new Tcl interpreter.
     */

    public Interp()
    {
	interpPtr = create();
    }

    /*
     * Disposes of the Tcl interpreter.  This method must be called to release
     * the resources that are used for the interpreter.
     */

    public void dispose()
    {
	doDispose(interpPtr);
	interpPtr = 0;
    }


/*
 *----------------------------------------------------------------------
 *
 * getWorkingDir --
 *
 *	Retrieve the current working directory for this interpreter.
 *
 * Results:
 *	Returns the File for the directory.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

File
getWorkingDir()
{
    return new File(Util.getCwd());
}

    /*-----------------------------------------------------------------
     *
     *	                     VARIABLES
     *
     *-----------------------------------------------------------------
     */

    /*
     * Sets a variable.
     */
    public final TclObject setVar(String name, TclObject value,
	    int flags) throws TclException
    {
	return doSetVar2(interpPtr, name, null, value,
		(flags | TCL.PARSE_PART1));
    }

    /*
     * Sets a variable. 
     */
    public final TclObject setVar(String name1, String name2,
	    TclObject value, int flags) throws TclException
    {
	return doSetVar2(interpPtr, name1, name2, value, flags);
    }

    /*
     * Gets a variable.
     */
    public final TclObject getVar(String name, int flags)
	    throws TclException
    {
	return doGetVar2(interpPtr, name, null, (flags | TCL.PARSE_PART1));
    }

    /*
     * Gets a variable.
     */
    public final TclObject getVar(String name1, String name2, int flags)
	    throws TclException
    {
	return doGetVar2(interpPtr, name1, name2, flags);
    }

    /*
     * Unsets a variable.
     */
    public final void unsetVar(String name, int flags)
	    throws TclException
    {
	unsetVar(name, null, (flags | TCL.PARSE_PART1));
    }

    /*
     * Unsets a variable. 
     */
    public final void unsetVar(String name1, String name2, int flags)
	    throws TclException
    {
	int result = doUnsetVar2(interpPtr, name1, name2, flags);
	if (result !=  TCL.OK) {
	    throw new TclException(null, null, result, -1);
	}
    }

/*
 *----------------------------------------------------------------------
 *
 * traceVar --
 *
 *	Add a trace to a variable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

public void
traceVar(
    String name,		// Name of variable;  may end with "(index)"
				// to signify an array reference.
    VarTrace trace,		// Object to notify when specified ops are
				// invoked upon varName.
    int flags)			// OR-ed collection of bits, including any
				// of TCL_TRACE_READS, TCL_TRACE_WRITES,
				// TCL_TRACE_UNSETS, TCL_GLOBAL_ONLY,
				// TCL_NAMESPACE_ONLY.
throws
    TclException
{
    traceVar(name, null, trace, (flags | TCL.PARSE_PART1));
}

/*
 *----------------------------------------------------------------------
 *
 * traceVar --
 *
 *	Add a trace to a variable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

public native void
traceVar(
    String part1,		// Name of scalar variable or array.
    String part2,		// Name of element within array;  null means
				// trace applies to scalar variable or array
				// as-a-whole.  
    VarTrace trace,		// Object to notify when specified ops are
				// invoked upon varName.
    int flags)			// OR-ed collection of bits, including any
				// of TCL_TRACE_READS, TCL_TRACE_WRITES,
				// TCL_TRACE_UNSETS, TCL_GLOBAL_ONLY,
				// TCL_NAMESPACE_ONLY and
				// TCL_PARSE_PART1.
throws
    TclException;		// If variable doesn't exist.

/*
 *----------------------------------------------------------------------
 *
 * untraceVar --
 *
 *	Remove a trace from a variable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

public void
untraceVar(
    String name,		// Name of variable;  may end with "(index)"
				// to signify an array reference.
    VarTrace trace,		// Object associated with trace.
    int flags)			// OR-ed collection of bits describing current
				// trace, including any of TCL_TRACE_READS,
				// TCL_TRACE_WRITES, TCL_TRACE_UNSETS,
				// TCL_GLOBAL_ONLY, TCL_NAMESPACE_ONLY and
				// TCL_PARSE_PART1. 
throws
    TclException
{
    untraceVar(name, null, trace, (flags | TCL.PARSE_PART1));
}

/*
 *----------------------------------------------------------------------
 *
 * untraceVar --
 *
 *	Remove a trace from a variable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

public native void
untraceVar(
    String part1,		// Name of scalar variable or array.
    String part2,		// Name of element within array;  null means
				// trace applies to scalar variable or array
				// as-a-whole.  
    VarTrace trace,		// Object associated with trace.
    int flags)			// OR-ed collection of bits describing current
				// trace, including any of TCL_TRACE_READS,
				// TCL_TRACE_WRITES, TCL_TRACE_UNSETS,
				// TCL_GLOBAL_ONLY, TCL_NAMESPACE_ONLY and
				// TCL_PARSE_PART1. 
throws
    TclException;

    /*
     * Create a new command in the interpreter.
     */
    public void createCommand(String name, Command cmd)
    {
	doCreateCommand(interpPtr, name, cmd);
    }

    /*
     * Delete a command in the interpreter by its name.
     *
     * Results:
     *     Returns 0 if the command was deleted successfully.
     *     Returns -1 if there didn't exist a command by that name.
     */
    public int deleteCommand(String name)
    {
	return doDeleteCommand(interpPtr, name);
    }


/*
 *----------------------------------------------------------------------
 *
 * commandComplete --
 *
 *	Tests if the String is a complete command.
 *
 * Results:
 *	Boolean value.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

public static native boolean
commandComplete(
    String cmd);	// Complete or partially complete command

//     /*
//      * Returns the command procedure of the given command.
//      *
//      * @param name String name of the command.
//      * @return the command procedure of the given command, or null if
//      *      the command doesn't exist
//      */
//     public Command getCommand(String name);

    /*-----------------------------------------------------------------
     *
     *	                     EVAL
     *
     *-----------------------------------------------------------------
     */

    /*
     * Queries the value of the result.
     *
     * @return the current result in the interpreter.
     */
    public final TclObject getResult()
    {
	return doGetResult(interpPtr);
    }

    /*
     * Sets the result to the given TclObject.
     *
     * @param r the new result in the interpreter.
     */
    public final void setResult(TclObject r)
    {
	doSetResult(interpPtr, r);
    }

    /*
     * Sets the result to a new TclObject whose value is equal to the
     * given string.
     */
    public final void setResult(String r)
    {
	doSetResult(interpPtr, TclString.newInstance(r));
    }

    /*
     * Sets the result to a new TclObject whose value is equal to the
     * given integer.
     */
    public final void setResult(int r)
    {
	doSetResult(interpPtr, TclInteger.newInstance(r));
    }

    /*
     * Sets the result to a new TclObject whose value is equal to the
     * given boolean.
     */
    public final void setResult(boolean r)
    {
	doSetResult(interpPtr, TclBoolean.newInstance(r));
    }

    /*
     * Sets the result to a new TclObject whose value is equal to the
     * given double.
     */
    public final void setResult(double r)
    {
	doSetResult(interpPtr, TclDouble.newInstance(r));
    }

    /*
     * Resets the result to an empty value.
     *
     */
    public final void resetResult()
    {
	doResetResult(interpPtr);
    }


/*
 *----------------------------------------------------------------------
 *
 * eval --
 *
 *	Execute a Tcl command in a string or Tcl Object.
 *
 * Results:
 *	The return value is void.  However, a standard Tcl Exception
 *	may be generated.  The interpreter's result object will contain
 *	the value of the evaluation but will persist only until the next 
 *	call to one of the eval functions.
 *
 * Side effects:
 *	The side effects will be determined by the exact Tcl code to be 
 *	evaluated.
 *
 *----------------------------------------------------------------------
 */

public void 
eval(
    String script,	// A script to evaluate.
    int flags)		// Flags, either 0 or TCL_GLOBAL_ONLY.
throws 
    TclException 	// A standard Tcl exception.
{
    int result;
    result = doEval(interpPtr, script, flags);
    if (result != TCL.OK) {
	throw new TclException(null, null, result);
    }
}

public void 
eval(
    String script)	// A script to evaluate.
throws 
    TclException 	// A standard Tcl exception.
{
    eval(script, 0);
}


public void 
eval(
    TclObject tobj,	// A Tcl object holding a script to evaluate.
    int flags)		// Flags, either 0 or TCL_GLOBAL_ONLY.
throws 
    TclException 	// A standard Tcl exception.
{
    eval(tobj.toString(), flags);
}

/*
 *----------------------------------------------------------------------
 *
 * setErrorCode --
 *
 *	These functions set the errorCode variable in the interpreter
 *	to the given value.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Sets the interpreter error state so the interpreter doesn't
 *	set errorCode to NONE after the current eval returns.
 *
 *----------------------------------------------------------------------
 */

public native void
setErrorCode(
    TclObject code);

public void
setErrorCode(
    String code)
{
    setErrorCode(TclString.newInstance(code));
}


/*
 *----------------------------------------------------------------------
 *
 * addErrorInfo --
 *
 *	This function adds the given string to the errorInfo variable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

public native void
addErrorInfo(
    String message);		// Message to add to errorInfo


    /*
     * Private methods that are implemented in native code.
     */

    private final native long create();

    private static final native void doCreateCommand(long interpPtr,
	    String name, Command cmd);
    private static final native int doDeleteCommand(long interpPtr,
	    String name);
    private static final native void doDispose(long interpPtr);
    private static final native int doEval(long interpPtr, String s,
	    int flags);
    private static final native TclObject doGetResult(long interpPtr);
    private static final native TclObject doGetVar2(long interpPtr,
	    String name1, String name2, int flags) throws TclException;
    private static final native void doResetResult(long interpPtr);
    private static final native void doSetResult(long interpPtr, TclObject v);
    private static final native TclObject doSetVar2(long interpPtr,
	    String name1, String name2, TclObject value, int flags)
	    throws TclException;
    private static final native int doUnsetVar2(long interpPtr,
	    String name1, String name2, int flags);

    /*
     * The interpPtr contains the C Tcl_Interp* used in native code.  This
     * field is declared with package visibility so that it can be passed
     * to native methods by other classes in this package.
     */

    long interpPtr;

    /*
     * Initialize the Interp class by loading the native methods.
     */

    static {
	System.loadLibrary("tclblend");
    }
}
