/* 
 * ReflectObject.java --
 *
 *	This file contains the native implementation of Java
 *	reflected objects.
 *
 * Copyright (c) 1997 by Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) ReflectObject.java 1.6 97/10/20 17:21:14
 */

package tcl.lang;

/*
 * This class implements the reflection object type in Tcl.  It is based
 * on the native "jref" implementation, so it extends CObject to
 * inherent the shadowing behavior.
 */

class ReflectObject extends CObject {

/*
 *----------------------------------------------------------------------
 *
 * ReflectObject --
 *
 *	Construct a new ReflectObject from the given Tcl_Obj reference.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Increments the reference count of the Tcl_Obj.
 *
 *----------------------------------------------------------------------
 */

ReflectObject(
    long objPtr)		// pointer to Tcl_Obj from C.
{
    super(objPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * setReflectObjectFromAny --
 *
 *	Called to convert an TclObject's internal rep to ReflectObject.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When successful, the internal representation of tobj is
 *	changed to ReflectObject, if it is not already so.
 *
 *----------------------------------------------------------------------
 */

static void
setReflectObjectFromAny(
    Interp interp,		// Current interpreter. Must be non-null.
    TclObject tobj)		// The object to convert.
throws
    TclException		// If the object's internal rep is not
				// already ReflectObject, and the string rep
				// is not the name of a java.lang.Object
				// registered in the given interpreter.
				// Error message is left inside interp.
{
    InternalRep rep = tobj.getInternalRep();

    /*
     * If the object is already a Tcl_Obj reference, then we only need to check
     * to see if it is a valid Object name in the current interpreter and copy
     * the reference into a new ReflectObject.  Otherwise we need to create a
     * new Tcl_Obj to hold the reference.
     */

    if (!(rep instanceof ReflectObject)) {
	ReflectObject o;

	if (rep instanceof CObject) {
	    o = new ReflectObject(((CObject) rep).objPtr);
	} else {
	    o = new ReflectObject(newCObject(tobj.toString()));
	}
	/*
	 * Validate the object handle before updating the internal rep.
	 */

	getObject(interp.interpPtr, o.objPtr);
	tobj.setInternalRep(o);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * newInstance --
 *
 *	Creates a new instance of a TclObject that wraps a java.lang.Object.
 *
 * Results:
 *	The newly created TclObject.
 *
 * Side effects:
 *	The java.lang.Object will be registered in the interpreter.
 *
 *----------------------------------------------------------------------
 */

public static TclObject
newInstance(
    Interp interp,		// Current interpreter.
    Object obj)			// java.lang.Object to wrap.
{
    /*
     * Creates a new ReflectObject around the given java.lang.Object
     * unless the object is null, in which case it just returns
     * the name of the null object.
     */

    if (obj != null) {
	return new TclObject(new ReflectObject(newObject(interp.interpPtr,
		obj)));
    } else {
	return TclString.newInstance("java0x0");
    }
}

/*
 *----------------------------------------------------------------------
 *
 * get --
 *
 *	Returns a java.lang.Object represented by tobj. tobj must have a
 *	ReflectObject internal rep, or its string rep must be one of the
 *	currently registered objects.
 *
 * Results:
 *	The Java object represented by tobj.
 *
 * Side effects:
 *	When successful, the internal representation of tobj is
 *	changed to ReflectObject, if it is not already so.
 *
 *----------------------------------------------------------------------
 */

public static Object
get(
    Interp interp,		// Current interpreter. Must be non-null.
    TclObject tobj)		// The TclObject to query.
throws
    TclException		// If the internal rep of tobj cannot
				// be converted to a ReflectObject. 
				// Error message is left inside interp.
{
    setReflectObjectFromAny(interp, tobj);
    ReflectObject rep = (ReflectObject)(tobj.getInternalRep());
    return getObject(interp.interpPtr, rep.objPtr);
}

/*
 *----------------------------------------------------------------------
 *
 *  wrapJavaObject --
 *
 *	Wraps the java.lang.Object into a Tcl object and set the
 *	interpreter's result with this Tcl object. Primitive data
 *	types are converted into primitive Tcl data types. Otherwise,
 *	a ReflectObject wrapper is created for the object so that it
 *	can be later accessed with the Reflection API.
 *
 *	This method is called to wrap objects returned by
 *	Invoke.callMethod(), Invoke.getField() and
 *	Invoke.getProperty().
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The TclObject converted from the tcl.lang.Object will be set as
 *	the result of the interpreter. Both null and void values are
 *	treated as an empty result.
 *
 *----------------------------------------------------------------------
 */

static void
wrapJavaObject(
    Interp interp,			// Current interpreter.
    Object result[])			// The java.lang.Object to set as the
					// result of the interpreter.
{
    Object obj = result[0];
    Class cls = (Class)result[1];

    if (cls.equals(Void.TYPE)) {
	interp.resetResult();
	return;
    }

    if (obj == null) {
	interp.setResult("java0x0");
    } else if (obj instanceof Integer) {
	interp.setResult(((Integer) obj).intValue());
    } else if (obj instanceof Long) {
	interp.setResult(((Long) obj).intValue());
    } else if (obj instanceof Short) {
	interp.setResult(((Short) obj).intValue());
    } else if (obj instanceof Byte) {
	interp.setResult(((Byte) obj).intValue());
    } else if (obj instanceof Double) {
	interp.setResult(((Double) obj).doubleValue());
    } else if (obj instanceof Float) {
	interp.setResult(((Float) obj).doubleValue());
    } else if (obj instanceof Boolean) {
	interp.setResult(((Boolean) obj).booleanValue());
    } else if (obj instanceof Character) {
	interp.setResult(((Character) obj).toString());
    } else if (obj instanceof String) {
	interp.setResult((String)obj);
    } else {
	interp.setResult(ReflectObject.newInstance(interp, obj));
    }
}

/*
 *----------------------------------------------------------------------
 *
 * wrapTclObject --
 *
 *	This method wraps a TclObject into a java.lang.Object suitable
 *	to pass as an argument to the Invoke.call*() methods. A ReflectObject
 *	is returned if the TclObject's internalRep is a RefectObject,
 *	of if its stringRep is one of the regiestred ReflectObject. Otherwise
 *	a Java String is returned.
 *
 * Results:
 * 	A java.lang.Object version of tobj to pass as an argument to the
 *	Invoke.call*() methods. 
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static Object
wrapTclObject(
    Interp interp,		// Current interpreter.
    TclObject tobj)		// The TclObject to convert into a Java Object.
{
    try {
	return get(interp, tobj);
    } catch (TclException e) {
	return tobj.toString();
    }
}

/*
 *----------------------------------------------------------------------
 *
 * getSignature --
 *
 *	Get the signature specifiec by a TclObject. A signature is an
 *	Object array. For constructor signatures, sig[0] is the fully
 *	qualified name of the class; for method signatures, sig[0] is
 *	the string name of the method. Subsequently, sig[1] is the
 *	type of the first argument, sig[2] is the type of the second
 *	argument, and so on.
 *
 * Results:
 *	The signature Object array; null if the signature is invalid.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static Object[]
getSignature(
    Interp interp,		// Current interpreter.
    TclObject sigObj)		// The TclObject to translate into a signature
throws
    TclException		// If signature is not a valid Tcl list.
{
    int sigLen = TclList.getLength(interp, sigObj);
    if (sigLen < 1) {
	return null;
    }

    Object sig[] = new Object[sigLen];

    for (int i = 0; i < sigLen; i ++) {
	sig[i] = TclList.index(interp, sigObj, i).toString();
    }

    return sig;
}

/*
 *----------------------------------------------------------------------
 *
 * getObject --
 *
 *	Retrieve the Object associated with the current Tcl_Obj.
 *
 * Results:
 *	Returns the Object.
 *
 * Side effects:
 *	May convert Tcl_Obj to a jref internal rep.
 *
 *----------------------------------------------------------------------
 */

private static final native Object
getObject(
    long interpPtr,		// Pointer to Tcl_Interp.
    long objPtr)		// Pointer to Tcl_Obj.
throws
    TclException;		// If object does not identify a valid
				// Object in the specified interpreter.

/*
 *----------------------------------------------------------------------
 *
 * newObject --
 *
 *	Create a new jref Tcl_Obj to wrap a java.lang.Object.
 *
 * Results:
 *	The new Tcl_Obj handle.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static final native long
newObject(
    long interpPtr,		// Pointer to Tcl_Interp.
    Object obj);		// Object to wrap.

} // end ReflectObject
