/*
 * JtestCmd.java
 *
 * Copyright (c) 1997 Cornell University.
 * Copyright (c) 1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL
 * WARRANTIES.
 * 
 * SCCS: @(#) JtestCmd.java 1.10 97/10/14 12:59:38
 *
 */

package tcl.lang;

import java.util.*;

/**
 * This class implements the built-in "Jtest" command in Tcl. This
 * command is used mainly for debug purposes. E.g., verify whether the
 * refCount is maintained properly.
 */

class JtestCmd implements Command {
    static final private String validCmds[] = {
	"equal",
	"gc",
	"refcount",
	"type",
	"vars",
    };

    static final private int OPT_EQUAL 		= 0;
    static final private int OPT_GC		= 1;
    static final private int OPT_REFCOUNT	= 2;
    static final private int OPT_TYPE 		= 3;
    static final private int OPT_VARS 		= 4;

    public void cmdProc(Interp interp, TclObject argv[])
	    throws TclException {
	if (argv.length < 2) {
	    throw new TclNumArgsException(interp, 1, argv, 
		    "option ?arg arg ...?");
	}
	int opt = TclIndex.get(interp, argv[1], validCmds, "option", 0);

	switch (opt) {
	case OPT_EQUAL:
	    /*
	     * Returns if the two objects refer to the same Java object.
	     */
	    if (argv.length != 4) {
		throw new TclException(interp, "wrong # args: should be \"" +
			argv[0] + " equal object1 object2\"");
	    }

	    TclObject obj1 = argv[2];
	    TclObject obj2 = argv[3];

	    interp.setResult(TclBoolean.newInstance(obj1 == obj2));
	    break;

	case OPT_GC:
	    System.gc();
	    break;

	case OPT_REFCOUNT:
	    /*
	     * Returns the reference count of an object.
	     * E.g. info type $a
	     */
	    if (argv.length != 3) {
		throw new TclException(interp, "wrong # args: should be \"" +
			argv[0] + " type object\"");
	    }

	    TclObject o = argv[2];

	    /*
	     * The following script will return 2
	     *		set a foo
	     *		jtest refcount $a
	     * Two "owners" have preserveed to the object in $a
	     *		- the variable "a"
	     *		- the 3rd argument passed to the jtest command.
	     */	    
	    interp.setResult(TclInteger.newInstance(o.getRefCount()));
	    break;

	case OPT_TYPE:
	    /*
	     * Returns the Java class name of an object.
	     * E.g. info type $a
	     */
	    if (argv.length != 3) {
		throw new TclException(interp, "wrong # args: should be \"" +
			argv[0] + " type object\"");
	    }

	    interp.setResult(TclString.newInstance(
		    argv[2].getInternalRep().getClass().getName()));
	    break;

	case OPT_VARS:
	    Hashtable table = interp.varFrame.varTable;
	    for (Enumeration e1 = table.elements();
		    e1.hasMoreElements(); ) {
		Var v = (Var)e1.nextElement();
		System.out.print(v.hashKey + ": ");
		if ((v.flags & Var.UNDEFINED) != 0) {
		    System.out.print("undefined: ");
		}
		if (v.value == null) {
		    System.out.print("<null> ");
		} else if (v.value instanceof TclObject) {
		    System.out.print("<object> " + ((TclObject)v.value));
		} else if (v.value instanceof Hashtable) {
		    System.out.print("<table> ");
		} else if (v.value instanceof Var) {
		    System.out.print("<var> ");
		} else {
		    System.out.print("<error> ");
		}

		System.out.println();
	    }
	}	
    }
}

