/* 
 * javaCmd.c --
 *
 *	This file contains the Tcl command procedures for the
 *	TclJava package.
 *
 * Copyright (c) 1997 by Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) javaCmd.c 1.26 97/10/21 14:20:40
 */

#include "java.h"
#include "javaNative.h"
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>

extern void panic(char *, ...);

JavaInfo java;
static int initialized = 0;

/*
 * The following array contains the class names and jclass pointers for
 * all of the classes used by this module.  It is used to initialize
 * the java structure's jclass slots.
 */

static struct {
    jclass *addr;		/* Where to store jclass. */
    char *name;			/* Name of class to load. */
} classes[] = {
    { &java.InvocationTargetException,
      "java/lang/reflect/InvocationTargetException" },
    { &java.Object, "java/lang/Object" },
    { &java.Invoke, "tcl/lang/Invoke" },
    { &java.String, "java/lang/String" },
    { &java.Boolean, "java/lang/Boolean" },
    { &java.Double, "java/lang/Double" },
    { &java.Float, "java/lang/Float" },
    { &java.Number, "java/lang/Number" },
    { &java.Character, "java/lang/Character" },
    { &java.Field, "java/lang/reflect/Field" },
    { &java.Interp, "tcl/lang/Interp" },
    { &java.Command, "tcl/lang/Command" },
    { &java.TclObject, "tcl/lang/TclObject" },
    { &java.TclException, "tcl/lang/TclException" },
    { &java.CommandWithDispose, "tcl/lang/CommandWithDispose" },
    { &java.CObject, "tcl/lang/CObject" },
    { &java.Extension, "tcl/lang/Extension" },
    { &java.VarTrace, "tcl/lang/VarTrace" },
    { &java.Void, "java/lang/Void" },
    { NULL, NULL }
};

/*
 * The following array contains the information needed to load the jmethodID
 * pointers that this module uses.
 */

static struct {
    jmethodID *addr;		/* Where to store method id. */
    char *name;			/* Name of method to load. */
    jclass *class;		/* Where to find class id. */
    char *sig;			/* Signature of method. */
    int isStatic;		/* 1 if method is static. */
} methods[] = {
    { &java.toString, "toString", &java.Object,
      "()Ljava/lang/String;", 0 },
    { &java.callMethod, "callMethod", &java.Invoke,
      "(Ljava/lang/Object;[Ljava/lang/Object;[Ljava/lang/Object;)[Ljava/lang/Object;", 1 },
    { &java.get, "get", &java.Field, "(Ljava/lang/Object;)Ljava/lang/Object;",
      0 },
    { &java.booleanValue, "booleanValue", &java.Boolean, "()Z", 0 },
    { &java.intValue, "intValue", &java.Number, "()I", 0 },
    { &java.doubleValue, "doubleValue", &java.Number, "()D", 0 },
    { &java.getTargetException, "getTargetException",
      &java.InvocationTargetException, "()Ljava/lang/Throwable;", 0 },
    { &java.interpC, "<init>", &java.Interp, "(J)V", 0 },
    { &java.cmdProc, "cmdProc", &java.Command,
      "(Ltcl/lang/Interp;[Ltcl/lang/TclObject;)V", 0 },
    { &java.disposeCmd, "disposeCmd", &java.CommandWithDispose, "()V", 0 },
    { &java.newCObjectInstance, "newInstance", &java.CObject,
      "(J)Ltcl/lang/TclObject;", 1 },
    { &java.preserve, "preserve", &java.TclObject, "()V", 0},
    { &java.release, "release", &java.TclObject, "()V", 0},
    { &java.getInternalRep, "getInternalRep", &java.TclObject,
      "()Ltcl/lang/InternalRep;", 0},
    { &java.loadOnDemand, "loadOnDemand", &java.Extension,
      "(Ltcl/lang/Interp;Ljava/lang/String;Ljava/lang/String;)V", 1},
    { &java.traceProc, "traceProc", &java.VarTrace,
      "(Ltcl/lang/Interp;Ljava/lang/String;Ljava/lang/String;I)V", 0},
    { NULL, NULL, NULL, NULL, 0 }
};

/*
 * The following array contains signature information for fields used
 * by this module.
 */

static struct {
    jfieldID *addr;		/* Where to store method id. */
    char *name;			/* Name of method to load. */
    jclass *class;		/* Where to find class id. */
    char *sig;			/* Signature of method. */
} fields[] = {
    { &java.compCode, "compCode", &java.TclException, "I" },
    { &java.interpPtr, "interpPtr", &java.Interp, "J" },
    { &java.objPtr, "objPtr", &java.CObject, "J" },
    { NULL, NULL, NULL, NULL }
};

/*
 * The following array contains the association between Java classes
 * and the Tcl commands they implement.
 */

static struct {
    char *cmdName;		/* Name of Tcl command. */
    char *className;		/* Name of class that implements command. */
} commands[] = {
    { "java::call", "tcl.lang.JavaCallCmd" },
    { "java::field", "tcl.lang.JavaFieldCmd" },
    { "java::new", "tcl.lang.JavaNewCmd" },
    { "java::prop", "tcl.lang.JavaPropCmd" },
    { "java::info", "tcl.lang.JavaInfoCmd" },
    { "java::instanceof", "tcl.lang.JavaInstanceofCmd" },
    { "java::load", "tcl.lang.JavaLoadCmd" },
    { "java::null", "tcl.lang.JavaNullCmd" },
    { NULL, NULL }
};

/*
 * One JavaObj structure is allocated for every global reference to
 * a java object.
 */

typedef struct JavaObj {
    jobject id;			/* Global reference to java object. */
    int refCount;		/* Number of Tcl_Obj's that point to this
				 * object. */
    Tcl_HashTable *tablePtr;	/* Hash table that contains this object. */
    Tcl_Interp* interp;		/* Interpreter that owns the object. */
} JavaObj;

/*
 * Declarations of functions used only in this file.
 */

static void		FreeJavaObj(JavaObj *jPtr);
static int		InitJava(Tcl_Interp *interp);
static void		ConvertJavaObj(JNIEnv *env, Tcl_Interp *interp,
			    Tcl_HashTable *tablePtr, Tcl_Obj *objPtr,
			    jobject id);
static void		DupJavaCmdInternalRep(Tcl_Obj *srcPtr,
			    Tcl_Obj *dupPtr);
static void		DupJavaInternalRep(Tcl_Obj *srcPtr, Tcl_Obj *dupPtr);
static void		FreeJavaCmdInternalRep(Tcl_Obj *objPtr);
static void		FreeJavaInternalRep(Tcl_Obj *objPtr);
static int		JavaFieldCmd(ClientData clientData,
			    Tcl_Interp *interp, int objc,
			    Tcl_Obj *CONST objv[]);
static void		JavaInterpDeleted(ClientData clientData,
			    Tcl_Interp *interp);
static int		JavaObjectCmd(ClientData clientData,
			    Tcl_Interp *interp, int objc,
			    Tcl_Obj *CONST objv[]);
static jarray		NewJavaArray(JNIEnv *env, Tcl_Interp *interp,
			    Tcl_HashTable *objTable, int objc,
			    Tcl_Obj *CONST objv[]);
static int		SetJavaCmdFromAny(Tcl_Interp *interp, Tcl_Obj *objPtr);
static int		SetJavaFromAny(Tcl_Interp *interp, Tcl_Obj *objPtr);
static void		SetJavaObj(JNIEnv *env, Tcl_Interp *interp,
			    Tcl_HashTable *tablePtr, Tcl_Obj *objPtr,
			    jobject id);
static int		ToString(JNIEnv *env, Tcl_Obj *objPtr, jobject obj);
static void		UpdateStringOfJava(Tcl_Obj *objPtr);

/*
 * The structure below defines the Java Tcl object type by means of
 * procedures that can be invoked by generic object code.
 */

static Tcl_ObjType javaJRefType = {
    "jref",
    FreeJavaInternalRep,
    DupJavaInternalRep,
    UpdateStringOfJava,
    SetJavaFromAny
};

static Tcl_ObjType oldCmdType;
static Tcl_ObjType *cmdTypePtr = NULL;

/*
 *----------------------------------------------------------------------
 *
 * DllEntryPoint --
 *
 *	This wrapper function is used by Windows to invoke the
 *	initialization code for the DLL.  If we are compiling
 *	with Visual C++, this routine will be renamed to DllMain.
 *	routine.
 *
 * Results:
 *	Returns TRUE;
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

#ifdef __WIN32__
BOOL APIENTRY
DllEntryPoint(
    HINSTANCE hInst,		/* Library instance handle. */
    DWORD reason,		/* Reason this function is being called. */
    LPVOID reserved)		/* Not used. */
{
    return TRUE;
}
#endif

/*
 *----------------------------------------------------------------------
 *
 * Java_Init --
 *
 *	This procedure initializes the Java package in an unsafe interp.
 *	This is the initial entry point if this module is being loaded
 *	from Tcl.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

EXPORT(int,Java_Init)(
    Tcl_Interp *interp)
{
    JavaInterpInfo *infoPtr;
    jobject interpObj;
    jlong lvalue;
    int i;
    jstring cmdName, className;

    if (!initialized && (InitJava(interp) != TCL_OK)) {
	return TCL_ERROR;
    }

    infoPtr = (JavaInterpInfo *) ckalloc(sizeof(JavaInterpInfo));
    infoPtr->objTable = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(infoPtr->objTable, TCL_ONE_WORD_KEYS);

    /*
     * Allocate a new Interp instance to wrap this interpreter.
     */

    *(Tcl_Interp**)&lvalue = interp;
    interpObj = (*java.env)->NewObject(java.env, java.Interp,
	    java.interpC, lvalue);
    infoPtr->interpObj = (*java.env)->NewGlobalRef(java.env, interpObj);
    (*java.env)->DeleteLocalRef(java.env, interpObj);

    /*
     * Associate the interpreter data with the interp object.
     */

    Tcl_SetAssocData(interp, "java", JavaInterpDeleted, (ClientData) infoPtr);

    /*
     * Register the commands that are implemented in Java.
     */

    for (i = 0; commands[i].cmdName != NULL; i++) {
	cmdName = (*java.env)->NewStringUTF(java.env, commands[i].cmdName);
	className = (*java.env)->NewStringUTF(java.env, commands[i].className);
	(*java.env)->CallStaticVoidMethod(java.env, java.Extension,
		java.loadOnDemand, infoPtr->interpObj, cmdName, className);
	(*java.env)->DeleteLocalRef(java.env, cmdName);
	(*java.env)->DeleteLocalRef(java.env, className);
    }

    /*
     * Add a package variable that denotes the current patchlevel
     * of the java package.
     */

    Tcl_SetVar(interp, "::java::patchLevel", "1.0a1", TCL_GLOBAL_ONLY);
   
    Tcl_Eval(interp, "namespace eval ::java namespace export new call field");
    return Tcl_PkgProvide(interp, "java", "1.0");
}

/*
 *----------------------------------------------------------------------
 *
 * JavaInterpDeleted --
 *
 *	This routine is called when an interpreter that is using
 *	this module is deleted.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Releases all of the global object references in the object
 *	table, then destroys the table.
 *
 *----------------------------------------------------------------------
 */

static void
JavaInterpDeleted(
    ClientData clientData,	/* Pointer to Java info. */
    Tcl_Interp *interp)		/* Interpreter being deleted. */
{
    JavaInterpInfo *infoPtr = (JavaInterpInfo *)clientData;
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch search;
    JavaObj *objPtr;

    /*
     * Free all of the global object references and then mark the
     * JavaObj's as invalid.
     */

    for (entryPtr = Tcl_FirstHashEntry(infoPtr->objTable, &search);
	 entryPtr != NULL; entryPtr = Tcl_NextHashEntry(&search)) {
	objPtr = (JavaObj*)Tcl_GetHashValue(entryPtr);
	(*java.env)->DeleteGlobalRef(java.env, objPtr->id);
	objPtr->id = NULL;
	objPtr->tablePtr = NULL;
	objPtr->interp = NULL;
    }
    Tcl_DeleteHashTable(infoPtr->objTable);
    if (infoPtr->interpObj) {
	(*java.env)->DeleteGlobalRef(java.env, infoPtr->interpObj);
    }
    ckfree((char*)infoPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * InitJava --
 *
 *	Load and initialize the Java VM for use from Tcl. 
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
InitJava(
    Tcl_Interp *interp)
{
    JDK1_1InitArgs vm_args;
    jsize nVMs;
    int i;
    Tcl_Obj *resultPtr;
    jfieldID field;

    memset(&java, 0, sizeof(java));

    /*
     * Check to see if the current process already has a Java VM.  If so,
     * attach to it, otherwise create a new one.
     */

    if (JNI_GetCreatedJavaVMs(&java.vm, 1, &nVMs) < 0) {
	if (interp) {
	    Tcl_AppendResult(interp, "JNI_GetCreatedJavaVMs failed", NULL);
	}
	return TCL_ERROR;
    }

    if (nVMs == 0) {
	memset(&vm_args, 0, sizeof(vm_args));
	JNI_GetDefaultJavaVMInitArgs(&vm_args);
	
	vm_args.classpath = getenv("CLASSPATH");

	if (JNI_CreateJavaVM(&java.vm, &java.env, &vm_args) < 0) {
	    if (interp) {
		Tcl_AppendResult(interp, "JNI_CreateJavaVM failed", NULL);
	    }
	    return TCL_ERROR;
	}
    } else {
	(*java.vm)->AttachCurrentThread(java.vm, &java.env, NULL);
    }

    /*
     * Load the classes needed by this module.
     */

    for (i = 0; classes[i].addr != NULL; i++) {
	jclass classid;
	classid = (*java.env)->FindClass(java.env, classes[i].name);
	if (classid == NULL) {
	    if (interp) {
		Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			"could not find class ", classes[i].name,
			", check your CLASSPATH environment variable.", NULL);
	    }
	    goto error;
	} else {
	    *(classes[i].addr) = (jclass) (*java.env)->NewGlobalRef(java.env,
		    (jobject)classid);
	    (*java.env)->DeleteLocalRef(java.env, classid);
	}
    }

    /*
     * Load methods needed by this module.
     */

    for (i = 0; methods[i].addr != NULL; i++) {
	jmethodID id;
	if (methods[i].isStatic) {
	    id = (*java.env)->GetStaticMethodID(java.env, *(methods[i].class),
		    methods[i].name, methods[i].sig);
	} else {
	    id = (*java.env)->GetMethodID(java.env, *(methods[i].class),
		    methods[i].name, methods[i].sig);
	}
	if (id == NULL) {
	    if (interp) {
		resultPtr = Tcl_GetObjResult(interp);
		Tcl_AppendStringsToObj(resultPtr, "could not find method ",
			methods[i].name, " in ", NULL);
		ToString(java.env, resultPtr, *(methods[i].class));
	    }
	    goto error;
	}
	*(methods[i].addr) = id;
    }
    
    /*
     * Load fields needed by this module.
     */

    for (i = 0; fields[i].addr != NULL; i++) {
	field = (*java.env)->GetFieldID(java.env,
		*(fields[i].class), fields[i].name, fields[i].sig);
	if (field == NULL) {
	    if (interp) {
		Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			"could not find field ", fields[i].name,
			" in ", NULL);
		ToString(java.env, resultPtr, *(fields[i].class));
	    }
	    goto error;
	}
	*(fields[i].addr) = field;
    }

    /*
     * Get the Void.TYPE value.
     */

    field = (*java.env)->GetStaticFieldID(java.env, java.Void, "TYPE",
	    "Ljava/lang/Class;");
    java.voidTYPE = (*java.env)->GetStaticObjectField(java.env, java.Void,
	    field);
        
    /*
     * Register the Java object types.
     */

    Tcl_RegisterObjType(&javaJRefType);
    Tcl_RegisterObjType(&javaTclObjectType);
    
    /*
     * Interpose on the "cmdName" type to preserve 
     * java objects.
     */

    cmdTypePtr = Tcl_GetObjType("cmdName");
    oldCmdType = *cmdTypePtr;
    cmdTypePtr->freeIntRepProc = FreeJavaCmdInternalRep;
    cmdTypePtr->dupIntRepProc = DupJavaCmdInternalRep;
    cmdTypePtr->setFromAnyProc = SetJavaCmdFromAny;
    
    initialized = 1;
    return TCL_OK;

    error:
    (*java.env)->ExceptionClear(java.env);
    (*java.vm)->DestroyJavaVM(java.vm);
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * FreeJavaInternalRep --
 *
 *	Free the internal rep for a java object.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Decrements the refcount on a java object, and frees it if
 *	the last reference is gone.
 *
 *----------------------------------------------------------------------
 */

static void
FreeJavaInternalRep(
    Tcl_Obj *objPtr)
{
    JavaObj *jPtr = (JavaObj *)objPtr->internalRep.twoPtrValue.ptr2;

    FreeJavaObj(jPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * DupJavaInternalRep --
 *
 *	Initialize the internal representation of a java Tcl_Obj to a
 *	copy of the internal representation of an existing java object. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	"dupPtr"s internal rep is set to the java object corresponding to
 *	"srcPtr"s internal rep and the refcount on the java object
 *	in incremented.
 *
 *----------------------------------------------------------------------
 */

static void
DupJavaInternalRep(
    Tcl_Obj *srcPtr,
    Tcl_Obj *dupPtr)
{
    JavaObj *jPtr = (JavaObj *)srcPtr->internalRep.twoPtrValue.ptr2;
    jPtr->refCount++;
    dupPtr->typePtr = &javaJRefType;
    dupPtr->internalRep.twoPtrValue.ptr2 = (ClientData)jPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * UpdateStringOfJava --
 *
 *	Update the string representation for a java object.
 *	Note: This procedure does not invalidate an existing old string rep
 *	so storage will be lost if this has not already been done. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The object's string is set to a valid string that results from
 *	concatenating the string "java" with the object id's integer
 *	representation.
 *
 *----------------------------------------------------------------------
 */

static void
UpdateStringOfJava(
    Tcl_Obj *objPtr)
{
    JavaObj *jPtr = (JavaObj *)objPtr->internalRep.twoPtrValue.ptr2;
    char buffer[TCL_DOUBLE_SPACE+6];
    sprintf(buffer, "java0x%x", jPtr->id);
    objPtr->length = strlen(buffer);
    objPtr->bytes = ckalloc((unsigned) objPtr->length + 1);
    strcpy(objPtr->bytes, buffer);
}

/*
 *----------------------------------------------------------------------
 *
 * SetJavaFromAny --
 *
 *	Attempt to generate a java object ref internal form for the
 *	Tcl object "objPtr".
 *
 * Results:
 *	The return value is a standard object Tcl result. If an error occurs
 *	during conversion, an error message is left in the interpreter's
 *	result unless "interp" is NULL.
 *
 * Side effects:
 *	If no error occurs, an object reference is stored as "objPtr"s
 *	internal representation, and the object's refcount is increased.
 *
 *----------------------------------------------------------------------
 */

static int
SetJavaFromAny(
    Tcl_Interp *interp,
    Tcl_Obj *objPtr)
{
    jobject id;

    /*
     * If there is no interpreter context, generate a null reference.
     */

    if (!interp) {
	return TCL_ERROR;
    }

    if (JavaGetObject(interp, NULL, objPtr, &id) == TCL_ERROR) {
	Tcl_ResetResult(interp);
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"unknown java object \"", objPtr->bytes, "\"", NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * FreeJavaCmdInternalRep --
 *
 *	Free the internal rep for a java object.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Decrements the refcount on a java object, and frees it if
 *	the last reference is gone.
 *
 *----------------------------------------------------------------------
 */

static void
FreeJavaCmdInternalRep(
    Tcl_Obj *objPtr)
{
    JavaObj *jPtr = (JavaObj *)objPtr->internalRep.twoPtrValue.ptr2;

    if (jPtr) {
	FreeJavaObj(jPtr);
    }
    (oldCmdType.freeIntRepProc)(objPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * DupJavaCmdInternalRep --
 *
 *	Initialize the internal representation of a java Tcl_Obj to a
 *	copy of the internal representation of an existing java object. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	"dupPtr"s internal rep is set to the java object corresponding to
 *	"srcPtr"s internal rep and the refcount on the java object
 *	in incremented.
 *
 *----------------------------------------------------------------------
 */

static void
DupJavaCmdInternalRep(
    Tcl_Obj *srcPtr,
    Tcl_Obj *dupPtr)
{
    JavaObj *jPtr = (JavaObj *)srcPtr->internalRep.twoPtrValue.ptr2;
    if (jPtr) {
	jPtr->refCount++;
    }
    (oldCmdType.dupIntRepProc)(srcPtr, dupPtr);
    dupPtr->internalRep.twoPtrValue.ptr2
	= srcPtr->internalRep.twoPtrValue.ptr2;
}

/*
 *----------------------------------------------------------------------
 *
 * SetJavaCmdFromAny --
 *
 *	Attempt to generate a java object ref internal form for the
 *	Tcl object "objPtr".
 *
 * Results:
 *	The return value is a standard object Tcl result. If an error occurs
 *	during conversion, an error message is left in the interpreter's
 *	result unless "interp" is NULL.
 *
 * Side effects:
 *	If no error occurs, an object reference is stored as "objPtr"s
 *	internal representation, and the object's refcount is increased.
 *
 *----------------------------------------------------------------------
 */

static int
SetJavaCmdFromAny(
    Tcl_Interp *interp,
    Tcl_Obj *objPtr)
{
    int result;

    /*
     * Invoke the normal command type routine, but make sure
     * it doesn't free the java object.  Note that we have to
     * restore the ptr2 value after the conversion, since it gets
     * set to NULL by the setFromAnyProc.
     */

    if (objPtr->typePtr == &javaJRefType) {
	VOID *ptr2;
	if (objPtr->bytes == NULL) {
	    UpdateStringOfJava(objPtr);	
	}
	objPtr->typePtr = NULL;
	ptr2 = objPtr->internalRep.twoPtrValue.ptr2;
	result = (oldCmdType.setFromAnyProc)(interp, objPtr);
	objPtr->internalRep.twoPtrValue.ptr2 = ptr2;
    } else {
	result = (oldCmdType.setFromAnyProc)(interp, objPtr);

	/*
	 * Ensure that the ptr2 is null for non-java commands.
	 */

	if (result == TCL_OK) {
	    objPtr->internalRep.twoPtrValue.ptr2 = NULL;
	}
    }
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * JavaGetObject --
 *
 *	Get the java object represented by the given Tcl_Obj in the
 *	context given by the object hash table.
 *
 * Results:
 *	Returns a standard Tcl result and leaves the object in idPtr.
 *
 * Side effects:
 *	Converts the Tcl_Obj type if necessary.
 *
 *----------------------------------------------------------------------
 */

int
JavaGetObject(
    Tcl_Interp *interp,		/* Interp to use for error reporting, must
				 * not be NULL. */
    Tcl_HashTable *hashPtr,	/* Hash table containing JavaObj's. If null,
				 * gets table from assoc data of interp. */
    Tcl_Obj *objPtr,		/* Object to find. */
    jobject *idPtr)		/* Resulting java object. */
{
    Tcl_ObjType *oldTypePtr = objPtr->typePtr;
    JavaObj *jPtr = (JavaObj *) objPtr->internalRep.twoPtrValue.ptr2;
    int length;
    long id;
    char *str, *end;
    Tcl_HashEntry *entryPtr;
    
    /*
     * If value is already cached, perform sanity check and return.
     */

    if (((oldTypePtr == cmdTypePtr) || (oldTypePtr == &javaJRefType))
	    && jPtr && (jPtr->interp == interp)) {
	*idPtr = jPtr->id;
	return TCL_OK;
    }

    /*
     * Object needs to be looked up in hash table.
     */

    str = Tcl_GetStringFromObj(objPtr, &length);
    if ((length < 5) || (memcmp(str, "java", 4) != 0)) {
	return TCL_ERROR;
    }
    Tcl_SetErrno(0);
    id = strtoul(str+4, &end, 0);
    if ((end == str+4) || (Tcl_GetErrno() == ERANGE)
	    || (end != (str + length))) {
	return TCL_ERROR;
    }

    /*
     * Special case for the "null" object.
     */

    if (id == 0) {
	*idPtr = 0;
	return TCL_OK;
    }

    if (!hashPtr) {
	JavaInterpInfo *infoPtr = (JavaInterpInfo *) Tcl_GetAssocData(interp,
		"java", NULL);
	hashPtr = infoPtr->objTable;
    }
    entryPtr = Tcl_FindHashEntry(hashPtr, (char*) id);
    if (!entryPtr) {
	return TCL_ERROR;
    }

    jPtr = (JavaObj*) Tcl_GetHashValue(entryPtr);
    jPtr->refCount++;

    /*
     * The conversion to java object succeeded.  Free the old rep
     * before changing the type.
     */

    if ((oldTypePtr != NULL) &&	(oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(objPtr);
    }

    objPtr->internalRep.twoPtrValue.ptr2 = (VOID *) jPtr;
    objPtr->typePtr = &javaJRefType;
    *idPtr = jPtr->id;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * SetJavaObj --
 *
 *	Modify an object to be a Java object with the specified id.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Creates a new global reference if necessary.
 *
 *----------------------------------------------------------------------
 */

static void
SetJavaObj(
    JNIEnv *env,		/* Java environment pointer. */
    Tcl_Interp *interp,		/* Interp to associate object with. */
    Tcl_HashTable *tablePtr,	/* Hash table containing JavaObj's. */
    Tcl_Obj *objPtr,		/* Object to store into. */
    jobject id)			/* Java object reference to encapsulate. */
{
    Tcl_ObjType *oldTypePtr = objPtr->typePtr;
    Tcl_HashEntry *entryPtr;
    int new;
    JavaObj *jPtr;

    if (Tcl_IsShared(objPtr)) {
	panic("SetJavaObj called with shared object");
    }
    Tcl_InvalidateStringRep(objPtr);
    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(objPtr);
    }
    
    objPtr->typePtr = &javaJRefType;

    /*
     * If the entry already exists, then it's a duplicate reference and we
     * should simply increment the refcount.  Otherwise, it's a local
     * reference, so we need to create a new global reference and add it
     * to the hash table.
     */

    entryPtr = Tcl_FindHashEntry(tablePtr, (char*) id);
    if (entryPtr) {
	jPtr = (JavaObj*) Tcl_GetHashValue(entryPtr);
    } else {
	id = (*env)->NewGlobalRef(env, id);
	entryPtr = Tcl_CreateHashEntry(tablePtr, (char*) id, &new);
	if (new) {
	    jPtr = (JavaObj*)ckalloc(sizeof(JavaObj));
	    jPtr->refCount = 0;
	    jPtr->interp = interp;
	    jPtr->tablePtr = tablePtr;
	    jPtr->id = id;
	    objPtr->internalRep.twoPtrValue.ptr2 = (VOID*) jPtr;
	    UpdateStringOfJava(objPtr);
	    Tcl_CreateObjCommand(interp, objPtr->bytes, JavaObjectCmd,
		    (ClientData) jPtr, NULL);
	    Tcl_SetHashValue(entryPtr, (ClientData)jPtr);
	} else {
	    jPtr = (JavaObj*) Tcl_GetHashValue(entryPtr);
	}
    }
    jPtr->refCount++;
    objPtr->internalRep.twoPtrValue.ptr2 = (VOID*) jPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * FreeJavaObj --
 *
 *	This procedure releases a reference to a java object if the
 *	refcount is zero.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	May remove the object command and release the object reference.
 *
 *----------------------------------------------------------------------
 */

static void
FreeJavaObj(
    JavaObj *jPtr)		/* Object to free. */
{
    Tcl_HashEntry *entryPtr;
    jPtr->refCount--;
    if (jPtr->refCount <= 0) {
	if (jPtr->tablePtr) {
	    entryPtr = Tcl_FindHashEntry(jPtr->tablePtr, (char*)jPtr->id);
	    if (entryPtr) {
		Tcl_DeleteHashEntry(entryPtr);
	    }
	}
	if (jPtr->id) {
	    char buffer[TCL_DOUBLE_SPACE+6];
	    sprintf(buffer, "java0x%x", jPtr->id);
	    (*java.env)->DeleteGlobalRef(java.env, jPtr->id);
	    Tcl_DeleteCommand(jPtr->interp, buffer);
	}
	ckfree((char*)jPtr);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_ReflectObject_getObject --
 *
 *	Retrieve the Object associated with the current Tcl_Obj.
 *

 * Class:     tcl_lang_ReflectObject
 * Method:    getObject
 * Signature: (JJ)Ljava/lang/Object;
 *
 * Results:
 *	Returns the Object.
 *
 * Side effects:
 *	May convert Tcl_Obj to a jref internal rep.
 *
 *----------------------------------------------------------------------
 */

jobject JNICALL
Java_tcl_lang_ReflectObject_getObject(
    JNIEnv * env,		/* Java environment pointer. */
    jclass reflectClass,	/* Pointer to ReflectObject class. */
    jlong interpPtr,		/* Pointer to Tcl_Interp. */
    jlong objectPtr)		/* Pointer to Tcl_Obj. */
{
    Tcl_Interp *interp = *(Tcl_Interp **)&interpPtr;
    Tcl_Obj *objPtr = *(Tcl_Obj **)&objectPtr;
    jobject id;

    if (!objPtr) {
	jclass nullClass = (*env)->FindClass(env,
		"java/lang/NullPointerException");
	(*env)->ThrowNew(env, nullClass, "Invalid ReflectObject.");
	return 0;
    } else if (!interp) {
	jclass nullClass = (*env)->FindClass(env,
		"java/lang/NullPointerException");
	(*env)->ThrowNew(env, nullClass, "Invalid Interp.");
	return 0;
    }
	
    if (JavaGetObject(interp, NULL, objPtr, &id) == TCL_ERROR) {
	Tcl_ResetResult(interp);
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
		"unknown java object \"", objPtr->bytes, "\"", NULL);
	JavaThrowTclException(env, TCL_ERROR);
	return NULL;
    }
    return id;

}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_ReflectObject_newObject --
 *
 *	Create a new Tcl_Obj to wrap a given Java object.
 *
 * Results:
 *	Returns the newly allocated Tcl_Obj.
 *
 * Side effects:
 *	Registers a new object in the object table.
 *
 *----------------------------------------------------------------------
 */

jlong JNICALL
Java_tcl_lang_ReflectObject_newObject(
    JNIEnv * env,		/* Java environment pointer. */
    jclass reflectClass,	/* Pointer to ReflectObject class. */
    jlong interpPtr,		/* Pointer to Tcl_Interp. */
    jobject object)		/* Pointer to Object to wrap. */
{
    Tcl_Interp *interp = *(Tcl_Interp **)&interpPtr;
    Tcl_Obj *objPtr;
    JavaInterpInfo *infoPtr;
    jlong lvalue;

    if (!interp) {
	jclass nullClass = (*env)->FindClass(env,
		"java/lang/NullPointerException");
	(*env)->ThrowNew(env, nullClass, "Invalid Interp.");
	return 0;
    }

    /*
     * Create a new Tcl_Obj and convert it into a jref.
     */

    infoPtr = (JavaInterpInfo *) Tcl_GetAssocData(interp, "java", NULL);
    objPtr = Tcl_NewObj();
    SetJavaObj(env, interp, infoPtr->objTable, objPtr, object);
    *(Tcl_Obj**)&lvalue = objPtr;
    return lvalue;
}

/*
 *----------------------------------------------------------------------
 *
 * JavaObjectCmd --
 *
 *	This procedure implements the Tcl command that corresponds
 *	to a java object managed by this module.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

static int
JavaObjectCmd(
    ClientData clientData,	/* Pointer to JavaObj. */
    Tcl_Interp *interp,		/* Current interpreter. */
    int objc,			/* Number of arguments. */
    Tcl_Obj *CONST objv[])	/* Argument objects. */
{
    JavaObj *jPtr = (JavaObj *)clientData;
    jarray sigArray, argArray;
    Tcl_Obj **sigv, *objPtr;
    int sigc;
    jobject value, class, exception;
    jarray objects;

    /*
     * Add an object reference if this is a simple command.
     */

    if ((objc > 0) && (objv[0]->typePtr == cmdTypePtr)
	    && !(objv[0]->internalRep.twoPtrValue.ptr2)) {
	jPtr->refCount++;
	objv[0]->internalRep.twoPtrValue.ptr2 = (VOID *) jPtr;
    }
	

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "signature ?arg arg ...?");
        return TCL_ERROR;
    }

    /*
     * Extract the signature from the first argument.
     */

    if (Tcl_ListObjGetElements(interp, objv[1], &sigc, &sigv)
	    != TCL_OK) {
	return TCL_ERROR;
    } else if (sigc < 1) {
	Tcl_AppendToObj(Tcl_GetObjResult(interp),
		"bad signature: must specify a method name", -1);
	return TCL_ERROR;
    }	

    /*
     * Construct the signature array by converting the list
     * elements into java strings and adding them to an array of
     * string objects.
     */

    sigArray = NewJavaArray(java.env, interp, jPtr->tablePtr, sigc, sigv);

    /*
     * Now construct the argument array.
     */

    argArray = NewJavaArray(java.env, interp, jPtr->tablePtr, objc-2, objv+2);

    /*
     * Call Invoke::callMethod to invoke the method on the object.
     */

    objects = (*java.env)->CallStaticObjectMethod(java.env, java.Invoke,
	    java.callMethod, jPtr->id, sigArray, argArray);
    exception = (*java.env)->ExceptionOccurred(java.env);
    (*java.env)->DeleteLocalRef(java.env, sigArray);
    (*java.env)->DeleteLocalRef(java.env, argArray);

    /*
     * Must reset the interpeter result because the method call
     * might have modified the state of the interpreter.
     */

    Tcl_ResetResult(interp);

    if (exception) {
	JavaCatchException(java.env, interp, jPtr->tablePtr, exception);
	return TCL_ERROR;
    } 
	
    /*
     * Return the result of the method invocation, or reset the result
     * if the return type was "void".
     */

    value = (*java.env)->GetObjectArrayElement(java.env, objects, 0);
    class = (*java.env)->GetObjectArrayElement(java.env, objects, 1);

    (*java.env)->DeleteLocalRef(java.env, objects);

    if ((*java.env)->IsSameObject(java.env, java.voidTYPE, class)) {
	Tcl_ResetResult(interp);
    } else {
	objPtr = Tcl_GetObjResult(interp);
	ConvertJavaObj(java.env, interp, jPtr->tablePtr, objPtr, value);
    }

    (*java.env)->DeleteLocalRef(java.env, value);
    (*java.env)->DeleteLocalRef(java.env, class);

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ConvertJavaObj --
 *
 *	This function converts a java object to one of the Tcl object
 *	types.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Changes the given Tcl object to the value of the Java object,
 *	using the best internal rep available.
 *
 *----------------------------------------------------------------------
 */

static void
ConvertJavaObj(
    JNIEnv *env,		/* Java environment pointer. */
    Tcl_Interp *interp,		/* Interp to register objects in. */
    Tcl_HashTable *tablePtr,	/* Object table.  */
    Tcl_Obj *objPtr,		/* An initialized Tcl object for result. */
    jobject jobj)		/* A java object reference. */
{
    int noMatch = 0;

    /*
     * Select the best internal representation for the java object.  If
     * no representation works, construct a new JavaObj.
     */

    if (!jobj) {
	Tcl_SetStringObj(objPtr, "java0x0", -1);
    } else if ((*java.env)->IsInstanceOf(java.env, jobj, java.Boolean)) {
	jboolean value = (*java.env)->CallBooleanMethod(java.env, jobj,
		java.booleanValue);
	Tcl_SetBooleanObj(objPtr, value);
    } else if ((*java.env)->IsInstanceOf(java.env, jobj, java.Double)
	    || (*java.env)->IsInstanceOf(java.env, jobj, java.Float)) {
	jdouble value = (*java.env)->CallDoubleMethod(java.env, jobj,
		java.doubleValue);
	Tcl_SetDoubleObj(objPtr, (double)value);
    } else if ((*java.env)->IsInstanceOf(java.env, jobj, java.Number)) {
	jint value = (*java.env)->CallIntMethod(java.env, jobj, java.intValue);
	Tcl_SetLongObj(objPtr, (long)value);
    } else if ((*java.env)->IsInstanceOf(java.env, jobj, java.String)) {
	jsize length;
	const char *msg;
	msg = (*java.env)->GetStringUTFChars(java.env, jobj, NULL);
	if (msg) {
	    length = (*java.env)->GetStringUTFLength(java.env, jobj);
	    Tcl_SetStringObj(objPtr, (char *)msg, length);
	    (*java.env)->ReleaseStringUTFChars(java.env, jobj, msg);
	} else {
	    noMatch = 1;
	}
    } else if ((*java.env)->IsInstanceOf(java.env, jobj, java.Character)) {
	if (ToString(java.env, objPtr, jobj) != TCL_OK) {
	    noMatch = 1;
	}
    } else {
	noMatch = 1;
    }

    if (noMatch || (*java.env)->ExceptionOccurred(java.env)) {
	(*java.env)->ExceptionClear(java.env);
	SetJavaObj(java.env, interp, tablePtr, objPtr, jobj);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * NewJavaArray --
 *
 *	Constructs a new Java Object array from the given array of
 *	Tcl_Obj's.
 *
 * Results:
 *	Returns a new jarray.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static jarray
NewJavaArray(
    JNIEnv *env,		/* Java environment pointer. */
    Tcl_Interp *interp,		/* Interpreter to report errors in. */
    Tcl_HashTable *objTable,	/* Object table for the interpreter. */
    int objc,			/* Number of objects to put in array. */
    Tcl_Obj *CONST objv[])	/* Objects to put in array. */
{
    jobject id;
    int local, i;
    jarray array = (*env)->NewObjectArray(env, objc, java.Object, NULL);

    for (i = 0; i < objc; i++) {
	if (JavaGetObject(interp, objTable, objv[i], &id) == TCL_ERROR) {
	    id = (*env)->NewStringUTF(env,
		    Tcl_GetStringFromObj(objv[i], NULL));
	    local = 1;
	} else {
	    local = 0;
	}
	(*env)->SetObjectArrayElement(env, array, i, id);
	if (local) {
	    (*env)->DeleteLocalRef(env, id);
	}
    }
    return array;
}

/*
 *----------------------------------------------------------------------
 *
 * ToString --
 *
 *	Invoke the toString() method on an object
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Places the string in the specified object.
 *
 *----------------------------------------------------------------------
 */

int
ToString(
    JNIEnv *env,		/* Java environment pointer. */
    Tcl_Obj *objPtr,		/* Object that will hold the string. */
    jobject obj)		/* Object whose string should be retrieved. */
{
    jstring str;
    int length;
    const char * msg;

    str = (*env)->CallObjectMethod(env, obj, java.toString);
    if (!str) {
	return TCL_ERROR;
    }
    length = (*env)->GetStringUTFLength(env, str);
    msg = (*env)->GetStringUTFChars(env, str, NULL);
    Tcl_AppendToObj(objPtr, (char *)msg, length);
    (*env)->ReleaseStringUTFChars(env, str, msg);
    (*env)->DeleteLocalRef(env, str);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * JavaCatchException --
 *
 *	Converts an exception into a proper Tcl result.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Updates the interpreter result object and the
 *	errorCode variable.  Clears any pending Java exception and
 *	deallocates the exception object.
 *
 *----------------------------------------------------------------------
 */

void
JavaCatchException(
    JNIEnv *env,		/* Java VM environment. */
    Tcl_Interp *interp,
    Tcl_HashTable *tablePtr,
    jobject exception)		/* Exception or NULL. */
{
    Tcl_Obj *objPtr, *objv[2];

    (*env)->ExceptionClear(env);
    if ((*env)->IsInstanceOf(env, exception,
	    java.InvocationTargetException)) {
	jobject object;
	object = (*env)->CallObjectMethod(env, exception,
		java.getTargetException);
	(*env)->DeleteLocalRef(env, exception);
	exception = object;
    }
	
    objv[0] = Tcl_NewStringObj("JAVA", -1);
    objv[1] = Tcl_NewObj();
    SetJavaObj(env, interp, tablePtr, objv[1], exception);
    objv[2] = Tcl_GetObjResult(interp);
    ToString(env, objv[2], exception);
    (*env)->DeleteLocalRef(env, exception);

    objPtr = Tcl_NewListObj(3, objv);
    Tcl_SetObjErrorCode(interp, objPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * JavaThrowTclException --
 *
 *	Generate a TclException with the given result code.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Throws a new exception in the Java VM.
 *
 *----------------------------------------------------------------------
 */

void
JavaThrowTclException(
    JNIEnv *env,		/* Java environment pointer. */
    int result)			/* One of TCL_ERROR, etc. */
{
    jmethodID init = (*env)->GetMethodID(env, java.TclException, "<init>",
	    "(I)V");
    jobject exc = (*env)->NewObject(env, java.TclException, init, result);
    (*env)->Throw(env, exc);
    (*env)->DeleteLocalRef(env, exc);
}
