/* 
 * javaInterp.c --
 *
 *	This file contains the native method implementations for the
 *	tcl.lang.Interp class.
 *
 * Copyright (c) 1997 by Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) javaInterp.c 1.16 97/10/20 18:58:46
 */

#include "java.h"
#include "javaNative.h"

/*
 * Declaractions for functions used only in this file.
 */

static void		JavaCmdDeleteProc(ClientData clientData);
static int		JavaCmdProc(ClientData clientData, Tcl_Interp *interp,
			    int objc, Tcl_Obj *CONST objv[]);
static char *		JavaTraceProc(ClientData clientData,
			    Tcl_Interp *interp, char *name1, char *name2,
			    int flags);
static void		ThrowNullPointerException(JNIEnv *env);


/*
 *----------------------------------------------------------------------
 *
 * ThrowNullPointerException --
 *
 *	Generate a NullPointerException to indicate that a method was
 *	invoked on a dead Interp.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Throws a new exception in the Java VM.  Creates a local
 *	reference so this function can only be called from a native
 *	method implementation.
 *
 *----------------------------------------------------------------------
 */

static void
ThrowNullPointerException(
    JNIEnv *env)		/* Java environment pointer. */
{
    jclass nullClass = (*env)->FindClass(env,
	    "java/lang/NullPointerException");
    (*env)->ThrowNew(env, nullClass, "Invalid interpreter.");
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_create --
 *
 *	Create a new Tcl interpreter.
 *
 * Class:     tcl_lang_Interp
 * Method:    create
 * Signature: ()J
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

jlong JNICALL
Java_tcl_lang_Interp_create(
    JNIEnv *env,		/* Java environment. */
    jobject interpObj)		/* Handle to Interp object. */
{
    jlong lvalue;
    JavaInterpInfo *infoPtr;
    Tcl_Interp *interp = Tcl_CreateInterp();
    if (Java_Init(interp) != TCL_OK) {
	Tcl_DeleteInterp(interp);
	return 0;
    }
    infoPtr = (JavaInterpInfo *)Tcl_GetAssocData(interp, "java", NULL);
    infoPtr->interpObj = (*env)->NewGlobalRef(env, interpObj);
    *(Tcl_Interp**)&lvalue = interp;
    return lvalue;
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_doDispose --
 *
 *	Delete the interpeter associated with this Interp class.
 *
 * Class:     tcl_lang_Interp
 * Method:    doDispose
 * Signature: (J)V
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void JNICALL
Java_tcl_lang_Interp_doDispose(
    JNIEnv *env,		/* Java environment. */
    jclass interpClass,		/* Handle to Interp class. */
    jlong interpPtr)		/* Value of Interp.interpPtr. */
{
    Tcl_Interp *interp = *(Tcl_Interp **)&interpPtr;

    if (!interp) {
	ThrowNullPointerException(env);
	return;
    }
    Tcl_DeleteInterp(interp);
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_doEval --
 *
 *	Evaluate the given string.
 *
 * Class:     tcl_lang_Interp
 * Method:    doEval
 * Signature: (JLjava/lang/String;)I
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Whatever the eval does.
 *
 *----------------------------------------------------------------------
 */

jint JNICALL
Java_tcl_lang_Interp_doEval(
    JNIEnv *env,		/* Java environment. */
    jclass interpClass,		/* Handle to Interp class. */
    jlong interpPtr,		/* Value of Interp.interpPtr. */
    jstring string,		/* String to eval. */
    jint flags)			/* Evaluation flags. */
{
    Tcl_Interp *interp = *(Tcl_Interp **)&interpPtr;
    const char *utfStr;
    jsize length;
    Tcl_Obj *objPtr;
    int result;

    if (!interp || !string) {
	ThrowNullPointerException(env);
	return 0;
    }

    length = (*env)->GetStringUTFLength(env, string);
    utfStr = (*env)->GetStringUTFChars(env, string, NULL);

    objPtr = Tcl_NewStringObj((/*UNCONST*/ char*)utfStr, length);
    Tcl_IncrRefCount(objPtr);
    if (!flags) {
	result = Tcl_EvalObj(interp, objPtr);
    } else {
	result = Tcl_GlobalEvalObj(interp, objPtr);
    }
    Tcl_DecrRefCount(objPtr);
    (*env)->ReleaseStringUTFChars(env, string, utfStr);
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_doGetResult --
 *
 *	Returns the current interpreter result as a TclObject.
 *
 * Class:     tcl_lang_Interp
 * Method:    doGetResult
 * Signature: (J)Ljava/lang/TclObject;
 *
 * Results:
 *	Returns a newly allocated TclObject.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

jobject JNICALL
Java_tcl_lang_Interp_doGetResult(
    JNIEnv *env,		/* Java environment. */
    jclass interpClass,		/* Handle to Interp class. */
    jlong interpPtr)		/* Value of Interp.interpPtr. */
{
    Tcl_Interp *interp = *(Tcl_Interp **)&interpPtr;

    if (!interp) {
	ThrowNullPointerException(env);
	return 0;
    }

    return JavaGetTclObject(env, Tcl_GetObjResult(interp), NULL);
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_doSetResult --
 *
 *	Sets the current interpreter result.
 *
 * Class:     tcl_lang_Interp
 * Method:    doSetResult
 * Signature: (JLtcl/lang/TclObject;)V
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void JNICALL
Java_tcl_lang_Interp_doSetResult(
    JNIEnv *env,		/* Java environment. */
    jclass interpClass,		/* Handle to Interp class. */
    jlong interpPtr,		/* Value of Interp.interpPtr. */
    jobject result)		/* New result object. */
{
    Tcl_Interp *interp = *(Tcl_Interp **)&interpPtr;
    Tcl_Obj *resultPtr;

    if (!interp || !result) {
	ThrowNullPointerException(env);
	return;
    }

    resultPtr = JavaGetTclObj(env, result);
    Tcl_SetObjResult(interp, resultPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_doResetResult --
 *
 *	Clears the result of the given interpreter.
 *
 * Class:     tcl_lang_Interp
 * Method:    doResetResult
 * Signature: (J)V
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void JNICALL
Java_tcl_lang_Interp_doResetResult(
    JNIEnv *env,		/* Java environment. */
    jclass interpClass,		/* Handle to Interp class. */
    jlong interpPtr)
{
    Tcl_Interp *interp = *(Tcl_Interp **)&interpPtr;

    if (!interp) {
	ThrowNullPointerException(env);
	return;
    }

    Tcl_ResetResult(interp);
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_doSetVar2 --
 *
 *	Set a variable to the given string.
 *
 * Class:     tcl_lang_Interp
 * Method:    doSetVar2
 * Signature:  (JLjava/lang/String;Ljava/lang/String;Ltcl/lang/TclObject;I)Ltcl/lang/TclObject;
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	May create a new Tcl_Obj.
 *
 *----------------------------------------------------------------------
 */

jobject JNICALL
Java_tcl_lang_Interp_doSetVar2(
    JNIEnv *env,		/* Java environment. */
    jclass interpClass,		/* Handle to Interp class. */
    jlong interpPtr,		/* Value of Interp.interpPtr. */
    jstring part1Str,		/* If part2 is NULL, this is name of scalar
                                 * variable. Otherwise it is the name of
                                 * an array. */
    jstring part2Str,		/* Name of an element within an array, or
				 * NULL. */
    jobject value,		/* New value for variable. */
    jint flags)			/* Various flags that tell how to set value:
				 * any of GLOBAL_ONLY,
				 * NAMESPACE_ONLY, APPEND_VALUE,
				 * LIST_ELEMENT, LEAVE_ERR_MSG, or 
				 * PARSE_PART1. */
{
    Tcl_Interp *interp = *(Tcl_Interp **)&interpPtr;
    const char *part1, *part2;
    Tcl_Obj *part1Ptr, *part2Ptr, *valuePtr, *resultPtr;
    jsize length;

    /*
     * Throw an exception if any of the objects are null.
     */

    if (!interp || !part1Str || !value) {
	ThrowNullPointerException(env);
	return NULL;
    }

    /*
     * Get the Tcl_Obj that corresponds to the given TclObject.
     */

    valuePtr = JavaGetTclObj(env, value);

    length = part1Str ? (*env)->GetStringUTFLength(env, part1Str) : 0;
    part1 = (*env)->GetStringUTFChars(env, part1Str, NULL);
    part1Ptr = Tcl_NewStringObj((/*UNCONST*/ char *) part1, length);
    Tcl_IncrRefCount(part1Ptr);

    if (part2Str) {
	 length = (*env)->GetStringUTFLength(env, part2Str);
	 part2 = (*env)->GetStringUTFChars(env, part2Str, NULL);
	 part2Ptr = Tcl_NewStringObj((/*UNCONST*/ char *) part2, length);
	 Tcl_IncrRefCount(part2Ptr);
    } else {
	part2Ptr = NULL;
    }

    resultPtr = Tcl_ObjSetVar2(interp, part1Ptr, part2Ptr, valuePtr, flags);

    Tcl_DecrRefCount(part1Ptr);
    (*env)->ReleaseStringUTFChars(env, part1Str, part1);
    if (part2Str) {
	Tcl_DecrRefCount(part2Ptr);
	(*env)->ReleaseStringUTFChars(env, part2Str, part2);
    }

    /*
     * If the result is the same as the value, return the same TclObject.
     * Otherwise create a new TclObject for the return value.
     */

    if (!resultPtr) {
	JavaThrowTclException(env, TCL_ERROR);
	return NULL;
    } else if (resultPtr == valuePtr) {
	return value;
    } else {
	return JavaGetTclObject(env, resultPtr, NULL);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_doGetVar2 --
 *
 *	Get the value of the given variable.
 *
 * Class:     tcl_lang_Interp
 * Method:    doGetVar2
 * Signature: (JLjava/lang/String;Ljava/lang/String;I)Ltcl/lang/TclObject;
 *
 * Results:
 *	Returns the TclObject that corresponds to the value of the
 *	variable.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

jobject JNICALL
Java_tcl_lang_Interp_doGetVar2(
    JNIEnv *env,		/* Java environment. */
    jclass interpClass,		/* Handle to Interp class. */
    jlong interpPtr,		/* Value of Interp.interpPtr. */
    jstring part1Str,		/* If part2 is NULL, this is name of scalar
                                 * variable. Otherwise it is the name of
                                 * an array. */
    jstring part2Str,		/* Name of an element within an array, or
				 * NULL. */
    jint flags)			/* Various flags that tell how to get value:
				 * any of GLOBAL_ONLY,
				 * NAMESPACE_ONLY, LEAVE_ERR_MSG, or 
				 * PARSE_PART1. */
{
    Tcl_Interp *interp = *(Tcl_Interp **)&interpPtr;
    const char *part1, *part2;
    Tcl_Obj *part1Ptr, *part2Ptr, *valuePtr;
    jsize length;

    if (!interp || !part1Str) {
	ThrowNullPointerException(env);
	return NULL;
    }

    length = part1Str ? (*env)->GetStringUTFLength(env, part1Str) : 0;
    part1 = (*env)->GetStringUTFChars(env, part1Str, NULL);
    part1Ptr = Tcl_NewStringObj((/*UNCONST*/ char *) part1, length);
    Tcl_IncrRefCount(part1Ptr);

    if (part2Str) {
	 length = (*env)->GetStringUTFLength(env, part2Str);
	 part2 = (*env)->GetStringUTFChars(env, part2Str, NULL);
	 part2Ptr = Tcl_NewStringObj((/*UNCONST*/ char *) part2, length);
	 Tcl_IncrRefCount(part2Ptr);
    } else {
	part2Ptr = NULL;
    }

    valuePtr = Tcl_ObjGetVar2(interp, part1Ptr, part2Ptr, flags);

    Tcl_DecrRefCount(part1Ptr);
    (*env)->ReleaseStringUTFChars(env, part1Str, part1);
    if (part2Str) {
	Tcl_DecrRefCount(part2Ptr);
	(*env)->ReleaseStringUTFChars(env, part2Str, part2);
    }

    if (!valuePtr) {
	JavaThrowTclException(env, TCL_ERROR);
	return NULL;
    }
    return JavaGetTclObject(env, valuePtr, NULL);
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_doUnsetVar2 --
 *
 *	Unset the given variable.
 *
 * Class:     tcl_lang_Interp
 * Method:    doUnsetVar2
 * Signature: (JLjava/lang/String;Ljava/lang/String;I)I
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

jint JNICALL
Java_tcl_lang_Interp_doUnsetVar2(
    JNIEnv *env,		/* Java environment. */
    jclass interpClass,		/* Handle to Interp class. */
    jlong interpPtr,		/* Value of Interp.interpPtr. */
    jstring part1Str,		/* If part2 is NULL, this is name of scalar
                                 * variable. Otherwise it is the name of
                                 * an array. */
    jstring part2Str,		/* Name of an element within an array, or
				 * NULL. */
    jint flags)			/* Various flags that tell how to get value:
				 * any of GLOBAL_ONLY,
				 * NAMESPACE_ONLY, LEAVE_ERR_MSG, or 
				 * PARSE_PART1. */
{
    Tcl_Interp *interp = *(Tcl_Interp **)&interpPtr;
    const char *part1, *part2;
    int result;

    if (!interp || !part1Str) {
	ThrowNullPointerException(env);
	return TCL_ERROR;
    }

    part1 = (*env)->GetStringUTFChars(env, part1Str, NULL);
    part2 = (part2Str) ? (*env)->GetStringUTFChars(env, part2Str, NULL) : NULL;

    result = Tcl_UnsetVar2(interp, (/*UNCONST*/ char *) part1,
	    (/*UNCONST*/ char *) part2, flags);

    (*env)->ReleaseStringUTFChars(env, part1Str, part1);
    if (part2Str) {
	(*env)->ReleaseStringUTFChars(env, part2Str, part2);
    }
    return (jint)result;
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_traceVar --
 *
 *	Add a trace to a variable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void JNICALL
Java_tcl_lang_Interp_traceVar(
    JNIEnv *env,		/* Java environment. */
    jobject interpObj,		/* Interp object. */
    jstring part1Str,		/* Name of scalar variable or array. */
    jstring part2Str,		/* Name of element within array;  null means
				 * trace applies to scalar variable or array
				 * as-a-whole. */
    jobject trace,		/* Object to notify when specified ops are
				 * invoked upon varName. */
    jint flags)			/* OR-ed collection of bits, including any
				 * of TCL_TRACE_READS, TCL_TRACE_WRITES,
				 * TCL_TRACE_UNSETS, TCL_GLOBAL_ONLY,
				 * TCL_NAMESPACE_ONLY and
				 * TCL_PARSE_PART1. */
{
    int result;
    const char *part1, *part2;
    Tcl_Interp *interp = JavaGetInterp(env, interpObj);

    if (!interp || !part1Str) {
	ThrowNullPointerException(env);
	return;
    }
    
    /*
     * Get the variable name.
     */

    part1 = (*env)->GetStringUTFChars(env, part1Str, NULL);
    if (part2Str) {
	part2 = (*env)->GetStringUTFChars(env, part2Str, NULL);
    } else {
	part2 = NULL;
    }

    /*
     * Set the trace.
     */

    result = Tcl_TraceVar2(interp, (/*UNCONST*/ char *) part1,
	    (/*UNCONST*/ char *) part2, flags, JavaTraceProc,
	    (ClientData) trace);

    /*
     * Release the strings and throw an exception if necessary.
     */

    (*env)->ReleaseStringUTFChars(env, part1Str, part1);
    if (part2Str) {
	(*env)->ReleaseStringUTFChars(env, part2Str, part2);
    }
    if (result != TCL_OK) {
	JavaThrowTclException(env, result);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_untraceVar --
 *
 *	Remove a trace from a variable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void JNICALL
Java_tcl_lang_Interp_untraceVar(
    JNIEnv *env,		/* Java environment. */
    jobject interpObj,		/* Interp object. */
    jstring part1Str,		/* Name of scalar variable or array. */
    jstring part2Str,		/* Name of element within array;  null means
				 * trace applies to scalar variable or array
				 * as-a-whole. */
    jobject trace,		/* Object associated with trace. */
    jint flags)			/* OR-ed collection of bits describing current
				 * trace, including any of TCL_TRACE_READS,
				 * TCL_TRACE_WRITES, TCL_TRACE_UNSETS,
				 * TCL_GLOBAL_ONLY, TCL_NAMESPACE_ONLY and
				 * TCL_PARSE_PART1. */
{
    const char *part1, *part2;
    Tcl_Interp *interp = JavaGetInterp(env, interpObj);

    if (!interp || !part1Str) {
	ThrowNullPointerException(env);
	return;
    }
    
    /*
     * Get the variable name.
     */

    part1 = (*env)->GetStringUTFChars(env, part1Str, NULL);
    if (part2Str) {
	part2 = (*env)->GetStringUTFChars(env, part2Str, NULL);
    } else {
	part2 = NULL;
    }

    /*
     * Unset the trace.
     */

    Tcl_UntraceVar2(interp, (/*UNCONST*/ char *) part1,
	    (/*UNCONST*/ char *) part2, flags, JavaTraceProc,
	    (ClientData) trace);

    /*
     * Release the strings and throw an exception if necessary.
     */

    (*env)->ReleaseStringUTFChars(env, part1Str, part1);
    if (part2Str) {
	(*env)->ReleaseStringUTFChars(env, part2Str, part2);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * JavaTraceProc --
 *
 *	This routine is the wrapper for any trace procedure implemented
 *	in Java.  It converts the command arguments into TclObject values
 *	and then invokes the cmdProc method on the appropriate object.
 *
 * Results:
 *	If no error occurred, returns NULL.
 *
 * Side effects:
 *	Whatever the trace does.
 *
 *----------------------------------------------------------------------
 */

static char *
JavaTraceProc(
    ClientData clientData,	/* Object handle. */
    Tcl_Interp *interp,		/* Current interpreter. */
    char *name1,		/* Name of scalar or array variable. */
    char *name2,		/* Element name or NULL. */
    int flags)			/* Operation flags. */
{
    jobject trace = (jobject)clientData;
    JavaInterpInfo *infoPtr;
    jstring name1Str, name2Str;
    jobject exception;

    infoPtr = (JavaInterpInfo *) Tcl_GetAssocData(interp, "java", NULL);

    name1Str = (*java.env)->NewStringUTF(java.env, name1);
    name2Str = (name2 ? (*java.env)->NewStringUTF(java.env, name2) : NULL);

    /*
     * Invoke the command and check for an exception.
     */

    (*java.env)->CallVoidMethod(java.env, trace, java.traceProc,
	    infoPtr->interpObj, name1Str, name2Str, flags);
    exception = (*java.env)->ExceptionOccurred(java.env);
    (*java.env)->ExceptionClear(java.env);

    (*java.env)->DeleteLocalRef(java.env, name1Str);
    if (name2Str) {
	(*java.env)->DeleteLocalRef(java.env, name2Str);
    }

    if (exception) {
	/*
	 * For now we signal the general presence of an exception but
	 * nothing else since we can only return static strings.
	 */

        (*java.env)->DeleteLocalRef(java.env, exception);
	return "exception in Java trace procedure";
    }
    
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_doCreateCommand --
 *
 *	Create a new Tcl command that is implemented by a Java object.
 *
 * Class:     tcl_lang_Interp
 * Method:    doCreateCommand
 * Signature: (JLjava/lang/String;Ltcl/lang/Command;)V
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Creates a new JavaCommand object to wrap the Java object.
 *
 *----------------------------------------------------------------------
 */

void JNICALL
Java_tcl_lang_Interp_doCreateCommand(
    JNIEnv *env,		/* Java environment. */
    jclass interpClass,		/* Handle to Interp class. */
    jlong interpPtr,		/* Value of Interp.interpPtr. */
    jstring nameStr,		/* Name of command to create. */
    jobject cmd)		/* Object that implements the command. */
{
    Tcl_Interp *interp = *(Tcl_Interp **)&interpPtr;
    const char *name;

    if (!interp || !nameStr || !cmd) {
	ThrowNullPointerException(env);
	return;
    }

    name = (*env)->GetStringUTFChars(env, nameStr, NULL);
    cmd = (*env)->NewGlobalRef(env, cmd);

    Tcl_CreateObjCommand(interp, (/*UNCONST*/ char *) name, JavaCmdProc,
	    (ClientData) cmd, JavaCmdDeleteProc);

    (*env)->ReleaseStringUTFChars(env, nameStr, name);
}

/*
 *----------------------------------------------------------------------
 *
 * JavaCmdDeleteProc --
 *
 *	Invokes the dispose() method on the command object if the
 *	object implements the Disposable interfaces. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Frees the global reference for the command object.
 *
 *----------------------------------------------------------------------
 */

static void
JavaCmdDeleteProc(
    ClientData clientData)
{
    jobject cmd = (jobject)clientData;

    if ((*java.env)->IsInstanceOf(java.env, cmd, java.CommandWithDispose)) {
	(*java.env)->CallVoidMethod(java.env, cmd, java.disposeCmd);
    }
    (*java.env)->DeleteGlobalRef(java.env, (jobject)clientData);
}

/*
 *----------------------------------------------------------------------
 *
 * JavaCmdProc --
 *
 *	This routine is the wrapper for any Tcl command implemented in
 *	Java.  It converts the command arguments into TclObject values
 *	and then invokes the cmdProc method on the appropriate object.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	Whatever the command does.
 *
 *----------------------------------------------------------------------
 */

static int
JavaCmdProc(
    ClientData clientData,	/* Command object handle. */
    Tcl_Interp *interp,		/* Current interpreter. */
    int objc,			/* Number of arguments. */
    Tcl_Obj *CONST objv[])	/* Argument objects. */
{
    jobject cmd = (jobject)clientData;
    JavaInterpInfo *infoPtr;
    jarray args;
    jobject value, exception;
    int i;

    infoPtr = (JavaInterpInfo *) Tcl_GetAssocData(interp, "java", NULL);

    /*
     * Construct the argument array.
     */

    args = (*java.env)->NewObjectArray(java.env, objc, java.TclObject, NULL);
    for (i = 0; i < objc; i++) {
	int isLocal;

	value = JavaGetTclObject(java.env, objv[i], &isLocal);
	(*java.env)->SetObjectArrayElement(java.env, args, i, value);

	/*
	 * Delete a newly created local ref expliticly since it may have
	 * been created outside of the VM and so will never be garbage
	 * collected.
	 */

	if (isLocal) {
	    (*java.env)->DeleteLocalRef(java.env, value);
	}
    }

    /*
     * Invoke the command and check for an exception.
     */

    (*java.env)->CallVoidMethod(java.env, cmd, java.cmdProc,
	    infoPtr->interpObj, args);
    exception = (*java.env)->ExceptionOccurred(java.env);
    (*java.env)->ExceptionClear(java.env);
    (*java.env)->DeleteLocalRef(java.env, args);

    /*
     * If there is a TclException, extract the compCode field.  Otherwise
     * throw the exception upward as if it were a reflection invocation.
     */

    if (exception) {
	int result; 
	if ((*java.env)->IsInstanceOf(java.env, exception,
		java.TclException)) {
	    result = (int) (*java.env)->GetIntField(java.env, exception,
		    java.compCode);
	} else {
	    JavaCatchException(java.env, interp, infoPtr->objTable, exception);
	    result = TCL_ERROR;
	}
        (*java.env)->DeleteLocalRef(java.env, exception);
	return result;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_doDeleteCommand --
 *
 *	Delete the given command.
 *
 * Class:     tcl_lang_Interp
 * Method:    doDeleteCommand
 * Signature: (JLjava/lang/String;)I
 *
 * Results:
 *	Returns -1 if no command was found, else returns 0.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

jint JNICALL
Java_tcl_lang_Interp_doDeleteCommand(
    JNIEnv *env,		/* Java environment. */
    jclass interpClass,		/* Handle to Interp class. */
    jlong interpPtr,		/* Value of Interp.interpPtr. */
    jstring nameStr)		/* Name of command to create. */
{
    Tcl_Interp *interp = *(Tcl_Interp **)&interpPtr;
    const char *name;
    int result;

    if (!interp || !nameStr) {
	ThrowNullPointerException(env);
	return -1;
    }

    name = (*env)->GetStringUTFChars(env, nameStr, NULL);
    result = Tcl_DeleteCommand(interp, (/*UNCONST*/ char*) name);
    (*env)->ReleaseStringUTFChars(env, nameStr, name);
    return (jint) result;
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_commandComplete --
 *
 *	Determine if the String cmd is a complete command.
 *
 * Class:     tcl_lang_Interp
 * Method:    commandComplete
 * Signature: (JLjava/lang/String;)Z
 *
 * Results:
 *	Returns true if the command is complete
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

jboolean JNICALL
Java_tcl_lang_Interp_commandComplete(
    JNIEnv *env,		/* Java environment. */
    jclass interpClass,		/* Handle to Interp class. */
    jstring cmdStr)		/* Command string to test */
{
    const char *cmd;
    jboolean result;

    if (!cmdStr) {
	ThrowNullPointerException(env);
	return JNI_FALSE;
    }

    cmd = (*env)->GetStringUTFChars(env, cmdStr, NULL);
    result = (Tcl_CommandComplete((/*UNCONST*/ char*) cmd)
	    ? JNI_TRUE : JNI_FALSE);
    (*env)->ReleaseStringUTFChars(env, cmdStr, cmd);
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_setErrorCode
 *
 *	Set the errorCode variable to the given string.
 *
 * Class:     tcl_lang_Interp
 * Method:    setErrorCode
 * Signature: (Ltcl/lang/TclObject;)V
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void JNICALL
Java_tcl_lang_Interp_setErrorCode(
    JNIEnv *env,		/* Java environment. */
    jobject interpObj,		/* Handle to Interp object. */
    jobject code)		/* TclObject to use as errorCode value. */
{
    Tcl_Obj *errorObjPtr;
    Tcl_Interp *interp = JavaGetInterp(env, interpObj);

    if (!interp || !code) {
	ThrowNullPointerException(env);
	return;
    }

    errorObjPtr = JavaGetTclObj(env, code);
    Tcl_SetObjErrorCode(interp, errorObjPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Java_tcl_lang_Interp_addErrorInfo --
 *
 *	Add the given string to the errorInfo.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void JNICALL
Java_tcl_lang_Interp_addErrorInfo(
    JNIEnv *env,		/* Java environment. */
    jobject interpObj,		/* Handle to Interp object. */
    jstring string)		/* String to add to errorInfo. */
{
    const char *str;
    Tcl_Interp *interp = JavaGetInterp(env, interpObj);

    if (!interp || !string) {
	ThrowNullPointerException(env);
	return;
    }

    str = (*env)->GetStringUTFChars(env, string, NULL);
    Tcl_AddErrorInfo(interp, (/*UNCONST*/ char*)str);
    (*env)->ReleaseStringUTFChars(env, string, str);
}

/*
 *----------------------------------------------------------------------
 *
 * JavaGetInterp --
 *
 *	Retrieve the C interpreter pointer from a Java Interp object.
 *
 * Results:
 *	Returns the Tcl_Interp *, or NULL.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

Tcl_Interp *
JavaGetInterp(
    JNIEnv *env,		/* Java VM environment. */
    jobject interpObj)		/* Interp object. */
{
    jlong interpPtr;
    Tcl_Interp *interp;

    if (!interpObj) {
	return NULL;
    }

    /*
     * Get the Tcl_Interp * from the Interp object.
     */

    interpPtr = (*env)->GetLongField(env, interpObj, java.interpPtr);

    /*
     * Copy the pointer out of the jlong.  We have to do it this way since the
     * jlong may be a structure.
     */

    interp = *(Tcl_Interp**)&interpPtr;
    return interp;
}
